/*
 * Name: cifsbrowse.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-20
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cifsbrowse.c,v 1.26 2002/04/05 17:53:49 cs Exp $
 */

/* General Description:
This module implements the CIFS browsing backend module.
*/

#include "stdheaders.h"
#include "module.h"
#include "backend.h"
#include "hashtable.h"
#include "notification.h"
#include "unicode.h"
#include "main.h"
#include "errcodes.h"
#include "memalloc.h"
#include "process.h"
#include "logging.h"
#include "blockpool.h"
#include "cfgparse.h"
#include "util.h"
#include "lmb.h"
#include "raprpc.h"
#include "smb.h"
#include "server.h"
#include "user.h"
#include "share.h"
#include "pslayer.h"
#include "system.h"
#include "mainmodule.h"
#include "guimanager.h"
#include "cifsmod.h"
#include "ipc.h"
#include "frontend.h"
#include "foundationconst.h"
#include "../../uiLib/maininterface.h"
#include "mainconst.h"

/* ------------------------------------------------------------------------- */

/* type IDs are sorted by directory hierarchy level */
#define	FTYPE_METADOMAIN	0
#define	FTYPE_DOMAIN		1
#define	FTYPE_SERVER		2
#define	FTYPE_SHARE			3

#define	ENTIRE_NET_NAME	"Entire_Network"

/* ------------------------------------------------------------------------- */

typedef struct automounter{
	btu8		*mountpoint;
	btint		refCount;
	btbool		isMounted;
	btbool		mountIsRunning;
	btbool		deleteMountDir;
	btbool		usedExistingMount;
	procCond_t	*mountLock;
	bttime		umountTimeout;
}automounter_t;

typedef struct browseFile{
	btuni				*name;
	struct browseFile	*parent;
	struct browseShare	*share;
	btint				type;
	btu8				*domain;
	btbool				isInvalidated;
	btbool				childrenIsValid;
	bttime				childrenUpdatedAt;
	btint				childrenValidFor;	/* uid of last request */
	hashtable_t			*children;
	btint				refCount;
	btbool				flag;
	automounter_t		*automnt;
}browseFile_t;

typedef struct browseShare{
	btuni					*path;
	btint					refCount;
	browseFile_t			*root;
	notificationCenter_t	*openNotify;
	notificationCenter_t	*invalidateNotify;
/* configuration data: */
	btu32					cfgChg;
	btu8					*domain;
	bttime					cacheTime;
	bttime					umountTimeout;
	btu8					*mountWithFrontend;
	array_t					*lookupIgnorePatterns;	/* array of Unicode pattern strings */
	btu8					*automountPath;
	btbool					showAdminShares;
	btbool					hasEntireNetEntry;
	btbool					addExplicitServers;
	btbool					listMountedOnly;
	btbool					enableAutoUnmount;
}browseShare_t;


static hashtable_t	*allAutomounts = NULL;


/* ------------------------------------------------------------------------- */

typedef struct bfileCtx{
	PROC_VARSET;
	browseFile_t	*file;
	automounter_t	*am;
	btuni			*name;
	btint			uid;
	void			*user;
	server_t		*srv;
	btint			startIndex;
	btbool			forceLookup;
	btbool			(*callback)(void *ctx, btuni *name, btu32 index, btbool isDir);
}bfileCtx_t;

static blockpool_t *bfilePool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(bfilePool, block);
}

static browseFile_t	*bfileNew(browseFile_t *parent, btuni *name, btint type);
static void	bfileRetain(browseFile_t *file);
static void	bfileRelease(browseFile_t *file);

/* ------------------------------------------------------------------------- */
/* ---------------------------- vfs functions: ----------------------------- */
/* ------------------------------------------------------------------------- */

static btint	bshareCompare(void *obj, void *otherObj)
{
browseShare_t	*this = obj, *other = otherObj;

	return uniStrCmp(this->path, other->path, 0);
}

static btu32	bshareHashVal(void *obj)
{
browseShare_t	*this = obj;

	return uniStrHash(this->path, 0);
}

static collectionObjectI_t	bshareInterface = {
	bshareCompare,
	bshareHashVal,
	NULL,
	NULL
};

static hashtable_t	*allShares = NULL;

static void	bshareRetain(browseShare_t *share)
{
	share->refCount++;
}

static void	 bshareFree(browseShare_t *share)
{
	if(share != NULL){
		if(share->path != NULL)
			memFree(share->path);
		if(share->domain != NULL)
			memFree(share->domain);
		if(share->automountPath != NULL)
			memFree(share->automountPath);
		if(share->openNotify != NULL)
			ncRelease(share->openNotify);
		if(share->invalidateNotify != NULL)
			ncRelease(share->invalidateNotify);
		if(share->lookupIgnorePatterns != NULL)
			arrFree(share->lookupIgnorePatterns);
		memset(share, 0, sizeof(*share));
		memFree(share);
	}
}

static void	bshareRelease(browseShare_t *share)
{
	if(--share->refCount == 0){
		htabRemoveObjectEqualTo(allShares, share);
		bshareFree(share);
	}else if(share->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

static btint	bshareConfigPath(browseShare_t *share, struct cfgProperty **path, btint maxEntries)
{
cfgProperty_t	*p, *db, *cfg;
btint			i = 0;
btu8			entity[256];

	db = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
	cfg = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.cfgFile));
	db = cfgQueryWithName("browsers", 1, db);
	cfg = cfgQueryWithName("browsers", 1, cfg);
	convUniToLocalAscii(share->path, entity, ELEMENTS(entity));
	p = cfgQueryWithName(entity, 1, db);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(entity, 1, cfg);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(GUIMAN_DEFAULT_ENTRY, 1, db);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	p = cfgQueryWithName(GUIMAN_DEFAULT_ENTRY, 1, cfg);
	if(p != NULL && i < maxEntries)
		path[i++] = p;
	return i;
}

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	uniStringInterface = {
	NULL,
	(void *)memFree,
	NULL,
	NULL
};

static browseShare_t	*bshareLookup(btuni *path)
{
browseShare_t	*volatile share;
btuni			scratch[64];
btint			numEntries, count, i;
cfgProperty_t	*cfgPath[8], *arr;
btu8			*pattern;

	if(allShares == NULL)
		allShares = htabNew(&bshareInterface, 8, 75);
	share = htabFindObjectEqualTo(allShares, &path);
	if(share != NULL){
		bshareRetain(share);
		return share;
	}
EXC_TRY
	share = memCalloc(sizeof(browseShare_t));
	share->path = uniStrDup(path);
	numEntries = bshareConfigPath(share, cfgPath, ELEMENTS(cfgPath));
	share->root = bfileNew(NULL, NULL, FTYPE_DOMAIN);
	share->root->share = share;
	share->openNotify = ncNew(2);
	share->invalidateNotify = ncNew(2);
	share->refCount = 1;
	share->lookupIgnorePatterns = arrNew(&uniStringInterface, 8);
	arr = cfgQueryFromArray("lookupIgnorePatterns", numEntries, cfgPath);
	count = cfgQueryValueCount(arr);
	for(i=0;i<count;i++){
		pattern = cfgStringValue(cfgQueryAtIndex(i, arr), "");
		if(pattern == NULL || pattern[0] == 0)
			continue;
		convLocalAsciiToUni(pattern, scratch, ELEMENTS(scratch));
		arrAddObject(share->lookupIgnorePatterns, uniStrDup(scratch));
	}
EXC_CATCH
	bshareFree(share);
	EXC_RETHROW();
EXC_ENDTRY
	htabAddObject(allShares, share);
	return share;
}

/* ------------------------------------------------------------------------- */

static void	bshareUpdateConfig(browseShare_t *share)
{
btint			numEntries;
cfgProperty_t	*path[8];
btu8			*asciiDom, srvDomain[64];
btuni			domain[64];
btu8			automountPath[MAXPATHLEN], *s;

	if(CFG_DID_CHANGE(share->cfgChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		asciiDom = cfgStringValue(cfgQueryFromArray("domain", numEntries, path), "WORKGROUP");

		numEntries = bshareConfigPath(share, path, ELEMENTS(path));
		asciiDom = cfgStringValue(cfgQueryFromArray("domain", numEntries, path), asciiDom);
		convLocalAsciiToUni(asciiDom, domain, ELEMENTS(domain));
		uniStrToupper(domain);
		convUniToServerAscii(domain, srvDomain, ELEMENTS(srvDomain));
		if(share->domain != NULL){
			if(strcmp(share->domain, srvDomain) != 0){
				memFree(share->domain);
				share->domain = utilStrDup(srvDomain);
			}
		}else{
			share->domain = utilStrDup(srvDomain);
		}
		share->cacheTime = cfgIntValue(cfgQueryFromArray("cacheTime", numEntries, path), 10000);
		share->umountTimeout = cfgIntValue(cfgQueryFromArray("umountTimeout", numEntries, path), 60000 * 60);
		share->mountWithFrontend = cfgStringValue(cfgQueryFromArray("mountWithFrontend", numEntries, path), NULL);
		share->showAdminShares = cfgEnumValue(cfgQueryFromArray("showAdminShares", numEntries, path), cfgYesNoEnum, 0);
		share->hasEntireNetEntry = cfgEnumValue(cfgQueryFromArray("hasEntireNetEntry", numEntries, path), cfgYesNoEnum, 1);
		share->automountPath = cfgStringValue(cfgQueryFromArray("automountPath", numEntries, path), "/tmp");
		if((s = sysRealpath(share->automountPath, automountPath)) == NULL)
			share->automountPath = utilStrDup(share->automountPath);
		else
			share->automountPath = utilStrDup(s);
		share->addExplicitServers = cfgEnumValue(cfgQueryFromArray("addExplicitServers", numEntries, path), cfgYesNoEnum, 0);
		share->listMountedOnly = cfgEnumValue(cfgQueryFromArray("listMountedOnly", numEntries, path), cfgYesNoEnum, 0);
		share->enableAutoUnmount = cfgEnumValue(cfgQueryFromArray("enableAutoUnmount", numEntries, path), cfgYesNoEnum, 1);
	}
}

/* ------------------------------------------------------------------------- */

static void bshareConnect(browseShare_t *share, void *user, void *caller)
{
beVfsRes_t		res;
btint			i, numEntries;
cfgProperty_t	*cfgPath[8];
btu8			entity[256];

	/* here we have to check whether the given browser exists on config */
	numEntries = cifsModGlobalCfgPath(cfgPath, ELEMENTS(cfgPath));
	for(i=0;i<numEntries;i++)
		cfgPath[i] = cfgQueryWithName("browsers", 1, cfgPath[i]);
	convUniToLocalAscii(share->path, entity, ELEMENTS(entity));
	if(cfgQueryFromArray(entity, numEntries, cfgPath) == NULL){	/* not configured */
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCnobrowser;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	res.err.class = res.err.code = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static void bshareDisconnect(browseShare_t *share, void *user, void *caller)
{
beVfsRes_t	res;

	res.err.class = res.err.code = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static browseFile_t	*bshareRootFile(browseShare_t *share)
{
	bfileRetain(share->root);
	return share->root;
}

static void	bshareStatfs(browseShare_t *share, void *user, void *caller)
{
beVfsRes_t	res;

	res.err.class = res.err.code = 0;
	res.bytesPerAU = 1024;
	res.totalAU = 10;
	res.availableAU = 0;
	res.timeGranularity.seconds = 0;
	res.timeGranularity.nseconds = 100000000;
	res.features = BE_VFS_CAN_SYMLINK | BE_VFS_IS_CASE_PRESERVING;
	PROC_IMMEDIATE_RET(caller, &res);
}

static btuni	*bshareVfsPath(browseShare_t *share)
{
	return share->path;
}

static btbool	bshareIsCaseSensitive(browseShare_t *share)
{
	return 0;
}

static notificationCenter_t	*bshareOpenNotification(browseShare_t *share)
{
	return share->openNotify;
}

static notificationCenter_t	*bshareInvalidateNotification(browseShare_t *share)
{
	return share->invalidateNotify;
}

static void	buserAuthenticate(browseShare_t *share, btint uid, btint gid, void *caller)
{
beVfsRes_t	res;

	res.err.class = res.err.code = 0;
	res.user = (void *)uid;
	/* we don't want to authenticate the user now, but we need the UID later */
	PROC_IMMEDIATE_RET(caller, &res);
}

static void	buserRetain(void *share)
{
	/* dummy */
}

static void	buserRelease(void *share)
{
	/* dummy */
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static bttime	bshareUmountTimeout(browseShare_t *share)
{
	bshareUpdateConfig(share);
	return share->umountTimeout;
}

static btu8	*bshareMountWithFrontend(browseShare_t *share)
{
	bshareUpdateConfig(share);
	return share->mountWithFrontend;
}

static btbool	bshareShowAdminShares(browseShare_t	*share)
{
	bshareUpdateConfig(share);
	return share->showAdminShares;
}

static btbool	bshareEnableAutoUnmount(browseShare_t *share)
{
	bshareUpdateConfig(share);
	return share->enableAutoUnmount;
}

static btbool	bshareHasEntireNetEntry(browseShare_t *share)
{
	bshareUpdateConfig(share);
	return share->hasEntireNetEntry;
}

static btu8	*bshareAutomountPath(browseShare_t *share)
{
	bshareUpdateConfig(share);
	return share->automountPath;
}

static inline btu8 *bshareDomain(browseShare_t *share)
{
	bshareUpdateConfig(share);
	return share->domain;
}

/* ------------------------------------------------------------------------- */
/* ---------------------------- file functions: ---------------------------- */
/* ------------------------------------------------------------------------- */

static void	tryUmount(void *obj, void *dummy);

static btint	automntCompare(void *o1, void *o2)
{
automounter_t	*this = o1, *other = o2;

	return utilStrCmp(this->mountpoint, other->mountpoint, 1);
}

static btu32	automntHashVal(void *obj)
{
automounter_t	*this = obj;

	return utilStrHash(this->mountpoint, 1);
}

static collectionObjectI_t	automountInterface = {
	automntCompare,
	automntHashVal,
	NULL,
	NULL
};

static void	automntFree(automounter_t *am)
{
	if(am != NULL){
		if(am->isMounted){	/* can't free while mounted */
			DEBUG_TRAP();
		}
		elpCancelCallAt(tryUmount, am);
		if(am->mountpoint != NULL){
			htabRemoveObjectEqualTo(allAutomounts, am);
			memFree(am->mountpoint);
		}
		if(am->mountLock != NULL)
			procCondFree(am->mountLock);
		memset(am, 0, sizeof(*am));
		memFree(am);
	}
}

static void	automntRetain(automounter_t *am)
{
	am->refCount++;
}

static void	automntRelease(automounter_t *am)
{
	if(--am->refCount == 0){
		automntFree(am);
	}else if(am->refCount < 0){
		DEBUG_TRAP();
	}
}

static automounter_t	*automntLookup(btu8 *mountpoint)
{
automounter_t	*volatile am = NULL;

	if(allAutomounts == NULL)
		allAutomounts = htabNew(&automountInterface, 32, 50);
	if((am = htabFindObjectEqualTo(allAutomounts, &mountpoint)) != NULL){
		automntRetain(am);
		return am;
	}
EXC_TRY;
	am = memCalloc(sizeof(automounter_t));
	am->mountpoint = utilStrDup(mountpoint);
	am->mountLock = procCondNew();
	am->refCount = 1;
	htabAddObject(allAutomounts, am);
EXC_CATCH;
	automntFree(am);
	EXC_RETHROW();
EXC_ENDTRY;
	return am;
}

/* ------------------------------------------------------------------------- */

static btint	bfileCompare(void *o1, void *o2)
{
browseFile_t	*this = o1, *other = o2;

	return uniStrCmp(this->name, other->name, 0);
}

static btu32	bfileHashVal(void *obj)
{
browseFile_t	*this = obj;

	return uniStrHash(this->name, 0);
}

static collectionObjectI_t	bfileInterface = {
	bfileCompare,
	bfileHashVal,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

static void	bfileFree(browseFile_t *file)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "cifsbrowse: freeing file at 0x%08x\n", (btint)file);
	if(file != NULL){
		if(!file->isInvalidated){	/* we must invalidate before we free */
			file->isInvalidated = 1;
			if(file->children != NULL && htabObjectCount(file->children) > 0){
				DEBUG_TRAP();	/* if we have children we should be referenced */
			}
			ncPostNotification(file->share->invalidateNotify, file);
		}
		if(file->parent != NULL){
			htabRemoveObjectEqualTo(file->parent->children, file);
			bfileRelease(file->parent);
			file->parent = NULL;
		}
		if(file->name != NULL){
			memFree(file->name);
			file->name = NULL;
		}
		if(file->domain != NULL){
			memFree(file->domain);
			file->domain = NULL;
		}
		if(file->automnt != NULL){
			automntRelease(file->automnt);
			file->automnt = NULL;
		}
		memFree(file);
	}
}

/* ------------------------------------------------------------------------- */

static browseFile_t	*bfileNew(browseFile_t *parent, btuni *name, btint type)
{
browseFile_t	*volatile file = NULL;
btu8			domain[64];
btuni			uniDomain[64];

EXC_TRY
	file = memCalloc(sizeof(browseFile_t));
	file->name = uniStrDup(name);
	file->type = type;
	file->refCount = 1;
	if(type == FTYPE_DOMAIN){
		if(name != NULL){
			uniStrCpy(uniDomain, name, ELEMENTS(uniDomain));
			uniStrToupper(uniDomain);
			convUniToServerAscii(uniDomain, domain, sizeof(domain));
			file->domain = utilStrDup(domain);
		}
	}else if(type == FTYPE_METADOMAIN){
		file->domain = utilStrDup(parent->domain);
	}
	if(type != FTYPE_SHARE)
		file->children = htabNew(&bfileInterface, 8, 75);
	if(parent != NULL){
		file->parent = parent;
		bfileRetain(parent);
		file->share = parent->share;
		htabAddObject(parent->children, file);
	}
EXC_CATCH
	bfileFree(file);
	EXC_RETHROW();
EXC_ENDTRY;
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "cifsbrowse: new file ->%s<- at 0x%08x\n", name == NULL ? "" : (char *)uniToCstring(name), (btint)file);
	return file;
}

/* ------------------------------------------------------------------------- */

static void	bfileInvalidate(browseFile_t *file)
{
array_t	*arr;
btint	i, cnt;

	if(!file->isInvalidated){
		file->isInvalidated = 1;
		if(file->children != NULL && htabObjectCount(file->children) > 0){
			arr = htabAllObjects(file->children);
			cnt = arrObjectCount(arr);
			for(i=0;i<cnt;i++){
				bfileInvalidate(arrObjectAtIndex(arr, i));
			}
			arrFree(arr);
		}
	/* The frontend will release and thus match our retain from lookup upon
	 * the following notification */
		ncPostNotification(file->share->invalidateNotify, file);
	/* This release matches the implicit retain in bfileNew(): */
		bfileRelease(file);	
	}
}

/* ------------------------------------------------------------------------- */

static void	bfileRetain(browseFile_t *file)
{
	file->refCount++;
}

static void	bfileRelease(browseFile_t *file)
{
	if(--file->refCount == 0){
		bfileFree(file);
	}else if(file->refCount < 0){
		DEBUG_TRAP();
	}
}

static inline btu8	*bfileDomain(browseFile_t *file)
{
	if(file->domain != NULL)
		return file->domain;
	return bshareDomain(file->share);
}

/* ------------------------------------------------------------------------- */

static void	loadChildrenProc(void *obj, void *arg)
{
bfileCtx_t		*this = obj;
browseFile_t	*f, *file = this->file;
lmbRes_t		*lrval = arg;
rapRes_t		*rrval = arg;
srvRes_t		*srval = arg;
userRes_t		*urval = arg;
smbRequest_t	*smb = arg;
btint			i, j, cnt = 0, srvCount;
array_t			*arr = NULL, *cfgServers;
btuni			*p, name[64];
btu8			srvName[64], *cfgName;
beFileRes_t		res;
cfgProperty_t	*db;

	PROC_BEGIN;
retry:
	bshareUpdateConfig(file->share);
	if(file->type == FTYPE_SERVER){
		convUniToLocalAscii(file->name, srvName, sizeof(srvName));
		this->srv = srvWithName(srvName, 0);
		PROC_WAITFOR_X(1, srvConnect(this->srv, this->uid, this));
		if(srval->err.class != 0){
			res.err = srval->err;
			goto errorOccured;
		}
		/* don't set gid in the following call because it's ignored anyway */
		PROC_WAITFOR_X(2, userAuthenticate(this->srv, bfileDomain(file), this->uid, 0, this));
		if(urval->err.class != 0){
			res.err = urval->err;
			goto errorOccured;
		}
		this->user = urval->user;
	}else{
		PROC_WAITFOR_X(4, lmbForDomain(bfileDomain(file), this->forceLookup, this));
		if(lrval->err.class != 0){
			res.err = lrval->err;
			goto errorOccured;
		}
		this->srv = lrval->server;
		PROC_WAITFOR_X(5, srvConnect(this->srv, this->uid, this));
		this->user = srvNullSession(this->srv);
	}
	userRetain(this->user);
	if(shrConnectionStatus(srvIpcShare(this->srv)) != PSL_CONNECTED){
		PROC_WAITFOR_X(6, shrConnect(srvIpcShare(this->srv), this->user, this));
		if(shrConnectionStatus(srvIpcShare(this->srv)) != PSL_CONNECTED){
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "cifsbrowse: no IPC share\n");
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCconnlost;
			userRelease(this->user);
			goto errorOccured;
		}
	}
	PROC_WAITFOR_X(7, srvWantRequest(this->srv, this));
	if(file->type == FTYPE_SERVER){
		PROC_WAITFOR_X(8, rapEnumShares(smb, userSmbUid(this->user), shrTid(srvIpcShare(this->srv)), this));
	}else{
		PROC_WAITFOR_X(9, rapEnumServers(smb, bfileDomain(file), file->type == FTYPE_METADOMAIN, userSmbUid(this->user), shrTid(srvIpcShare(this->srv)), this));
	}
	userRelease(this->user);
	if(rrval->err.class != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "cifsbrowse: error in enum RAP: %s\n", errDescription(&rrval->err));
		res.err = rrval->err;
		goto errorOccured;
	}
	if(rrval->status != RAP_OK && rrval->status != RAP_ERR_MORE_DATA){
		if(file->type != FTYPE_SERVER && rrval->status == RAP_ERR_NOT_SUPPORTED && !this->forceLookup){
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "cifsbrowse: LMB changed\n");
			this->forceLookup = 1;
			if(this->srv != NULL)
				srvRelease(this->srv);
			this->srv = NULL;
			goto retry;
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "cifsbrowse: error status in enum RAP: %d\n", rrval->status);
		res.err.class = ERR_CIFS_SRVerror;
		res.err.code = ERRCLS_CIFS_SRV;
		goto errorOccured;
	}
	if(rrval->status == RAP_ERR_MORE_DATA && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "cifsbrowse: serverEnum result incomplete\n");
	arr = htabAllObjects(file->children);
	cnt = arrObjectCount(arr);
	for(i=0;i<cnt;i++){
		f = arrObjectAtIndex(arr, i);
		f->flag = 0;
	}
	p = name;
	for(i=0;i<rrval->numEntries;i++){
		if(file->type == FTYPE_SERVER && rrval->entries[i].type != RAP_SHARE_TYPE_DISKTREE)
			continue;		/* skip non-disk shares */
		for(j = 0; j < 15 && rrval->entries[i].name[j] != 0; j++);
		rrval->entries[i].name[j] = 0;	/* limit to 15 characters */
		convServerAsciiToUni(rrval->entries[i].name, name, ELEMENTS(name));
		/* now capitalize share- and server-names */
		for(j=0;name[j]!=0;j++){
			name[j] = j == 0 ? uniCharToupper(name[j]) : uniCharTolower(name[j]);
		}
		if(!bshareShowAdminShares(file->share)){
			j = uniStrLen(name);
			if(j > 0 && name[j-1] == '$')
				continue;	/* skip admin-shares */
		}
		f = htabFindObjectEqualTo(file->children, &p);
		if(f == NULL){
			f = bfileNew(file, name, file->type + 1);	/* type from the next level */
			htabAddObject(file->children, f);
		}
		f->flag = 1;
	}
ignoreError:
	if(file->type == FTYPE_DOMAIN && file->share->addExplicitServers){
		/* add explicitly configured servers */
		db = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
		db = cfgQueryWithName("servers", 1, db);
		cfgServers = cfgQueryList(db);
		srvCount = arrObjectCount(cfgServers);
		p = name;
		for(i=0;i<srvCount;i++){
			cfgName = arrObjectAtIndex(cfgServers, i);
			if(cfgName[0] == '.')
				continue;	/* ignore server names starting with a dot such as .default */
			convLocalAsciiToUni(cfgName, name, ELEMENTS(name));
			f = htabFindObjectEqualTo(file->children, &p);
			if(f == NULL){
				f = bfileNew(file, name, file->type + 1);	/* type from the next level */
				htabAddObject(file->children, f);
			}
			f->flag = 1;
		}
		arrFree(cfgServers);
	}
	if(file->parent == NULL && file->type == FTYPE_DOMAIN && bshareHasEntireNetEntry(file->share)){
		convServerAsciiToUni(ENTIRE_NET_NAME, name, ELEMENTS(name));
		f = htabFindObjectEqualTo(file->children, &p);
		if(f == NULL){
			f = bfileNew(file, name, FTYPE_METADOMAIN);
			htabAddObject(file->children, f);
		}
		f->flag = 1;
	}
	if(arr != NULL){
		for(i=0;i<cnt;i++){	/* invalidate those that were not in the reply */
			f = arrObjectAtIndex(arr, i);
			if(!f->flag)
				bfileInvalidate(f);
		}
		arrFree(arr);
	}
	file->childrenIsValid = 1;
	file->childrenUpdatedAt = elpNow();
	file->childrenValidFor = this->uid;
	if(this->srv != NULL)
		srvRelease(this->srv);
	res.err.code = res.err.class = 0;
	PROC_RETURN(&res);
errorOccured:
	if(file->type == FTYPE_DOMAIN && file->share->addExplicitServers){
		/* ignore errors if we have explicit servers to add */
		goto ignoreError;
	}
	file->childrenIsValid = 0;
	arr = htabAllObjects(file->children);
	cnt = arrObjectCount(arr);
	for(i=0;i<cnt;i++){
		f = arrObjectAtIndex(arr, i);
		bfileInvalidate(f);
	}
	arrFree(arr);
	if(this->srv != NULL)
		srvRelease(this->srv);
	PROC_RETURN(&res);
	PROC_END;
}

static void	loadChildren(browseFile_t *file, btint uid, void *caller)
{
bfileCtx_t	*ctx;
beFileRes_t	res;

	if(file->type == FTYPE_SHARE){
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadpath;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, loadChildrenProc, caller, bfilePool, poolFreeBlock);
	ctx->file = file;
	ctx->uid = uid;
	PROC_START(ctx, NULL);
}


/* ------------------------------------------------------------------------- */

static btbool	childrenIsValid(browseFile_t *f, btint uid)
{
bttime	age;

	if(!f->childrenIsValid)
		return 0;
	age = elpNow() - f->childrenUpdatedAt;
	if(age < 0 || age > f->share->cacheTime){
		f->childrenIsValid = 0;
		return 0;
	}
	return f->childrenValidFor == uid;
}

/* ------------------------------------------------------------------------- */

static void	bfileLookupProc(void *obj, void *arg)
{
bfileCtx_t		*this = obj;
browseFile_t	*f;
beFileRes_t		res, *rval = arg;

	PROC_BEGIN;
	if(!childrenIsValid(this->file, this->uid)){
		PROC_WAITFOR_X(1, loadChildren(this->file, this->uid, this));
		if(rval->err.class != 0){
			PROC_RETURN(rval);
		}
	}
	f = htabFindObjectEqualTo(this->file->children, &this->name);
	if(f != NULL){
		bfileRetain(f);
		res.file = f;
		res.err.class = res.err.code = 0;
	}else{
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadpath;
	}
	PROC_RETURN(&res);
	PROC_END;
}

static void	bfileLookup(browseFile_t *parent, btuni *name, void *user, void *caller)
{
beFileRes_t	res;
bfileCtx_t	*ctx;
btint		i, count;

	if(parent->children == NULL){
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadpath;
		PROC_IMMEDIATE_RET(caller, &res);
	}else{
		count = arrObjectCount(parent->share->lookupIgnorePatterns);
		for(i=0;i<count;i++){
			if(uniWildcardMatch(name, arrObjectAtIndex(parent->share->lookupIgnorePatterns, i))){
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "cifsbrowse: ignoring lookup of %s due to pattern match\n", uniToCstring(name));
				res.err.class = ERRCLS_UNIX;
				res.err.code = ENOENT;
				PROC_IMMEDIATE_RET(caller, &res);
			}
		}
		PROC_INIT(ctx, bfileLookupProc, caller, bfilePool, poolFreeBlock);
		ctx->file = parent;
		ctx->name = name;
		ctx->uid = (btint)user;
		PROC_START(ctx, NULL);
	}
}

/* ------------------------------------------------------------------------- */

static void	vfsTimeFromTime64(struct beFileTime *vt, bti64 tm)
{
    tm -= FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL * 1000LL * 10000LL;	/* convert to 1970-based time */
    vt->seconds = tm / 10000000;
    vt->nseconds = (tm % 10000000) * 100;
}

static void	bfileGetAttributes(browseFile_t *file, void *user, void *caller)
{
beFileRes_t		res;
beFileAttr_t	attr;

	memset(&attr, 0, sizeof(attr));
	res.attr = &attr;
	res.err.class = res.err.code = 0;
	if(file->children == NULL){	/* leafs are automount links */
		attr.type = BE_FILETYPE_LNK;
		attr.unixMode = 0777;
		if(file->automnt == NULL || !file->automnt->isMounted){
			attr.unixMode |= 01000;	/* set sticky bit to indicate automounter */
		}
	}else{
		attr.type = BE_FILETYPE_DIR;
		attr.unixMode = 0555;
	}
	vfsTimeFromTime64(&attr.mtime, elpNow64());
	attr.atime = attr.ctime = attr.mtime;
	attr.numLinks = 0;	/* 0 if not known */
	attr.uid = (btint)user;
	attr.gid = 0;
	attr.size = 1024;
	attr.allocatedSize = 1024;
	PROC_IMMEDIATE_RET(caller, &res);
}

/* ------------------------------------------------------------------------- */

static btbool	bfileIsMounted(browseFile_t *file)
{
btint	i, cnt;
array_t	*arr;

	if(file->type == FTYPE_SHARE){
		if(file->automnt == NULL)
			return 0;
		return file->automnt->isMounted;
	}
	if(file->type == FTYPE_SERVER){
		if(file->children == NULL || !file->childrenIsValid)
			return 0;
		arr = htabAllObjects(file->children);
		cnt = arrObjectCount(arr);
		for(i=0;i<cnt;i++){
			if(bfileIsMounted(arrObjectAtIndex(arr, i)))
				break;
		}
		arrFree(arr);
		return i < cnt;	/* something is mounted if loop exited with break */
	}
	return 1;	/* other types default to "mounted" */
}

static btint	nextMountedChild(array_t *children, btint index)
{
btint	i, cnt;

	cnt = arrObjectCount(children);
	for(i=index;i<cnt;i++){
		if(bfileIsMounted(arrObjectAtIndex(children, i)))
			break;
	}
	return i;
}

/* ------------------------------------------------------------------------- */

static void	bfileReaddirProc(void *obj, void *arg)
{
bfileCtx_t		*this = obj;
array_t			*arr;
btint			cnt, i, next;
browseFile_t	*f;
beFileRes_t		res, *rval = arg;

	PROC_BEGIN;
	if(!childrenIsValid(this->file, this->uid)){
		PROC_WAITFOR_X(1, loadChildren(this->file, this->uid, this));
		if(rval->err.class != 0){
			PROC_RETURN(rval);
		}
	}
	bshareUpdateConfig(this->file->share);
	arr = htabAllObjects(this->file->children);
	cnt = arrObjectCount(arr);
	if(this->file->share->listMountedOnly){
		this->startIndex = nextMountedChild(arr, this->startIndex);
	}
	for(i=this->startIndex;i<cnt;i=next){
		f = arrObjectAtIndex(arr, i);
		if(this->file->share->listMountedOnly){
			next = nextMountedChild(arr, i + 1);
		}else{
			next = i + 1;
		}
		if(!this->callback(this->proc.caller, f->name, next, f->children != NULL))
			break;
	}
	arrFree(arr);
	res.eof = i >= cnt;
	res.err.class = res.err.code = 0;
	PROC_RETURN(&res);
	PROC_END;
}

static void	bfileReaddir(browseFile_t *file, void *user, btint startIndex, btbool (*resultCallback)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir), void *caller)
{
beFileRes_t	res;
bfileCtx_t	*ctx;

	if(file->children == NULL){
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadaccess;
		PROC_IMMEDIATE_RET(caller, &res);
	}else{
		PROC_INIT(ctx, bfileReaddirProc, caller, bfilePool, poolFreeBlock);
		ctx->file = file;
		ctx->uid = (btint)user;
		ctx->startIndex = startIndex;
		ctx->callback = resultCallback;
		PROC_START(ctx, NULL);
	}
}

/* ------------------------------------------------------------------------- */

static btuni	*bfileName(browseFile_t *file)
{
	return file->name;
}

static browseFile_t	*bfileParent(browseFile_t *file)
{
	return file->parent;
}

/* ------------------------------------------------------------------------- */

static void	shareDidUmount(void *obj, void *arg);

static void	useExistingMount(browseFile_t *file)
{
	file->automnt->deleteMountDir = 1;	/* always delete mountpoint dir afterwards */
	file->automnt->isMounted = 1;
	automntRetain(file->automnt);	/* avoid that we are freed before we unmounted */
	ncAddObserver(mainModUmountNotification(), shareDidUmount, file->automnt);
}

static btint	createMountpoint(browseFile_t *file, btint uid)
{
#ifndef HAVE_DISKARB
/* We do not need to create the mount point directory if we use diskarb
 * because diskarb.c will create it in the pre mount notification anyway.
 * We MUST avoid access to the automount directory from the main thread/task
 * because this could deadlock us with the autodiskmount daemon.
 */
	if(mkdir(file->automnt->mountpoint, 0777) == 0){
		chown(file->automnt->mountpoint, uid, -1);
	}
#endif
	useExistingMount(file);
	return 0;
}

static void	deleteMountpoint(automounter_t *am)
{
#ifndef HAVE_DISKARB
	/* diskarb (autodiskmount) should have done this for us */
	rmdir(am->mountpoint);	/* always delete mountpoint dir and ignore errors */
#endif
	if(am->deleteMountDir){
		am->deleteMountDir = 0;
	}
	if(am->isMounted){
		am->isMounted = 0;
		ncRemoveObserver(mainModUmountNotification(), shareDidUmount, am);
		automntRelease(am);	/* free now, if object not valid any more */
	}
}

static void	shareDidUmount(void *obj, void *arg)
{
automounter_t	*automnt = obj;
btu8			*mpPath = arg;

	if(automnt->mountpoint != NULL && mpPath != NULL){
		if(utilStrCmp(automnt->mountpoint, mpPath, 1) == 0){
			deleteMountpoint(automnt);
		}
	}
}

static void	tryUmountProc(void *obj, void *arg)
{
bfileCtx_t		*this = obj;
automounter_t	*am = this->am;
btError			*rval = arg;

	PROC_BEGIN;
	automntRetain(am);
	while(am->mountIsRunning){
		PROC_WAITFOR_X(1, procCondWait(am->mountLock, this));
	}
	if(!am->isMounted){
		goto doReturn;
	}
	am->mountIsRunning = 1;
	PROC_WAITFOR_X(2, mainModUmount(0, am->mountpoint, 0, this));
	if(rval->class == 0 || (rval->class == ERRCLS_FRAMEWORK && rval->code == ERR_FRWnosuchmnt)){
		deleteMountpoint(am);
	}else{	/* retry umount later... */
		elpCallAtCancelPrev(elpNow() + am->umountTimeout, tryUmount, am);
	}
	am->mountIsRunning = 0;
	procCondBroadcast(am->mountLock, NULL);
doReturn:
	automntRelease(am);
	PROC_END;
}

static void	tryUmount(void *obj, void *dummy)
{
automounter_t	*am = obj;
bfileCtx_t		*ctx;

	PROC_INIT(ctx, tryUmountProc, NULL, bfilePool, poolFreeBlock);
	ctx->am = am;
	PROC_START(ctx, NULL);
}

static void	bfileReadlinkProc(void *obj, void *arg)
{
bfileCtx_t		*this = obj;
browseFile_t	*file = this->file;
btError			*mrval = arg;
btuni			uncPath[256];
btint			i;
beFileRes_t		res;

	PROC_BEGIN;
	if(file->automnt == NULL){
		btu8 *s, mountPath[MAXPATHLEN + 1], name[64];

		strcpy(mountPath, bshareAutomountPath(file->share));
		strncat(mountPath, "/", ELEMENTS(mountPath));
		mountPath[ELEMENTS(mountPath)-1] = 0;
		i = strlen(mountPath);
		convUniToLocalAscii(file->parent->name, name, sizeof(name));
		for(s=name; *s != 0 && i < ELEMENTS(mountPath)-1; ){
			if(*s != '/')
				mountPath[i++] = *s;
			s++;
		}
		if(i < ELEMENTS(mountPath)-1)
			mountPath[i++] = '.';
		convUniToLocalAscii(file->name, name, sizeof(name));
		for(s=name; *s != 0 && i < ELEMENTS(mountPath)-1; ){
			if(*s != '/')
				mountPath[i++] = *s;
			s++;
		}
		mountPath[i] = 0;
		file->automnt = automntLookup(mountPath);
	}

	while(file->automnt->mountIsRunning){
		if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
			logPrintf(LOGM_RARE, "cifsbrowse: waiting for already running mount on %s to complete\n", file->automnt->mountpoint);
		PROC_WAITFOR_X(1, procCondWait(file->automnt->mountLock, this));
	}
	if(file->automnt->isMounted){	/* already mounted */
		res.err.class = res.err.code = 0;	/* nothing to be done */
		goto doReturn;
	}
	if(!mainModMountExists(file->automnt->mountpoint)){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "cifsbrowse: decided to mount on %s\n", file->automnt->mountpoint);
		file->automnt->mountIsRunning = 1;
		i = 0;
		uncPath[i++] = '\\';
		uncPath[i++] = '\\';
		uniStrCpy(uncPath + 2, file->parent->name, ELEMENTS(uncPath) - 2);
		i = uniStrLen(uncPath);
		if(i < 254)
			uncPath[i++] = '\\';
		uniStrCpy(uncPath + i, file->name, ELEMENTS(uncPath) - i);
		if(createMountpoint(file, this->uid) < 0){
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCnomountdir;
			goto doReturn;
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "cifsbrowse: ready creating mount point %s, going to mount\n", file->automnt->mountpoint);
		PROC_WAITFOR_X(4, mainModMount(this->uid, bshareMountWithFrontend(file->share), file->automnt->mountpoint, "cifsFile", uncPath, 0, 0, 1, this));
		file->automnt->usedExistingMount = 0;
		res.err = *mrval;
		if(mrval->class != 0){
			deleteMountpoint(file->automnt);
		}
	}else{
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "cifsbrowse: using exissting mount %s\n", file->automnt->mountpoint);
		useExistingMount(file);
		file->automnt->usedExistingMount = 1;
	}
doReturn:
	if(file->automnt != NULL){
		btuni	linkPath[MAXPATHLEN];

		if(file->automnt->mountIsRunning){
			file->automnt->mountIsRunning = 0;
			procCondBroadcast(file->automnt->mountLock, NULL);
		}
		if(!file->automnt->usedExistingMount){
			file->automnt->umountTimeout = bshareUmountTimeout(file->share);
			if(bshareEnableAutoUnmount(file->share))
				elpCallAtCancelPrev(elpNow() + file->automnt->umountTimeout, tryUmount, file->automnt);
		}
		convLocalAsciiToUni(file->automnt->mountpoint, linkPath, ELEMENTS(linkPath));
		res.path = linkPath;
	}
	PROC_RETURN(&res);
	PROC_END;
}

static void	bfileReadlink(browseFile_t *file, void *user, void *caller)
{
bfileCtx_t	*ctx;
beFileRes_t	res;

	if(file->children != NULL){		/* not a leaf */
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadaccess;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, bfileReadlinkProc, caller, bfilePool, poolFreeBlock);
	ctx->file = file;
	ctx->uid = (btint)user;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */
/* ----------------------------- configuration ----------------------------- */
/* ------------------------------------------------------------------------- */

static void	marshalError(ipcMsg_t *reply, btint class, btint code)
{
btError	err;

	ipcAddInt(reply, class);
	ipcAddInt(reply, code);
	err.class = class;
	err.code = code;
	ipcAddAscii(reply, errDescription(&err));
}

static array_t	*frontendModuleNames;

#define	PERM_NORMAL	MAIN_GUI_PERM_AUTHORIZED
#define	PERM_EXPERT	(MAIN_GUI_PERM_AUTHORIZED | MAIN_GUI_PERM_EXPERT)

static guiTypedesNumeric_t	numdesUmountTimeout = {
/*	div		mpy		inc		min		max			unit	*/
	1000,	1,		1000,	20000,	2147483647,	"seconds"
};

#define	ALLOW_NULL	0
#define	FLG_0		0

static guiEntry_t	browserCfgEntries[] = {
	{	MAIN_GUI_ENTRY_STRING,	"Browser\'s Domain",
		NULL,					"cifsBrowseConfigDomain",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"domain",
	},
	{	MAIN_GUI_ENTRY_FILE,	"Automount Path",
		NULL,					"cifsBrowseConfigAutomountPath",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"automountPath",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"Auto-Unmount Timeout",
		&numdesUmountTimeout,	"cifsBrowseConfigUmountTimeout",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"umountTimeout",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Enable Auto-Unmount",
		NULL,					"cifsBrowseConfigEnableAutoUnmount",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"enableAutoUnmount",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Show List of Domains",
		NULL,					"cifsBrowseConfigShowMeta",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"hasEntireNetEntry",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Show Admin Shares",
		NULL,					"cifsBrowseConfigShowAdmin",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"showAdminShares",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Add explicitly configured Servers",
		NULL,					"cifsBrowseAddExplicitServers",
		PERM_NORMAL,			FLG_0, ALLOW_NULL,	"addExplicitServers",
	},
	{	MAIN_GUI_ENTRY_BOOL,	"List mounted Servers only",
		NULL,					"cifsBrowseListMountedOnly",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"listMountedOnly",
	},
	{	MAIN_GUI_ENTRY_STRING,	"Mount With Frontend",
		NULL,					"cifsBrowseConfigFrontend",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"mountWithFrontend",
	},
};

static guiButton_t	buttonEntries[] = {
	{	"Defaults",			MAIN_GUI_METHOD_GET,
		1,	/* parameter */	0,	/* flags */
		"cifsBrowseConfigButtonDefaults",
	},
	{	"Revert",			MAIN_GUI_METHOD_GET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"cifsBrowseConfigButtonRevert",
	},
	{	"OK",				MAIN_GUI_METHOD_SET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"cifsBrowseConfigButtonOk",
	},
};

static guiClass_t	browserCfgDescription = {
	MAIN_GUI_CLASS_HIERARCHYCFG,
	"CIFS\nBrowsers",
	NULL,
	NULL
};

static guiObject_t	*browsersDsource;

/* ------------------------------------------------------------------------- */

static void	getValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	**cfgPathes[] = {NULL, NULL};
btu8	*hPath[] = {CIFS_CONFIG_SECTION, "browsers", NULL};
btu8	**hPathes[2]; /* = {hPath, NULL}; */

	hPathes[0] = hPath;
	hPathes[1] = NULL;
	/* NULL terminate data section of frontendModuleNames array: */
	arrAddObject(frontendModuleNames, NULL);
	arrRemoveLastObject(frontendModuleNames);
	cfgPathes[0] = (btu8 **)arrContentArray(frontendModuleNames);
	guimanGetValues(cfgPathes, hPathes, browserCfgDescription.entries, rq, reply);
	PROC_IMMEDIATE_RET(caller, NULL);
}

static void	setValues(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8	*path[] = {CIFS_CONFIG_SECTION, "browsers", NULL};

	guimanSetValues(1, path, browserCfgDescription.entries, rq, reply, uid);
	mainModGuiUpdate(browsersDsource);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static btint	stringCompare(void *s1, void *s2){
	return utilStrCmp(s1, s2, 0);
}

static btu32	stringHashVal(void *s){
	return utilStrHash(s, 0);
}

static collectionObjectI_t	browserNameInterface = {
	stringCompare, stringHashVal, NULL, NULL
};

static void	listBrowsers(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btbool			listAll;
btint			cnt, i, j;
cfgProperty_t	*p;
array_t			*all;
hashtable_t		*browsers;

	listAll = ipcGetInt(rq);
	browsers = htabNew(&browserNameInterface, 16, 75);
	for(j = (listAll ? 0 : 1);j < 2; j++){
		p = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(j ? gdata.dbFile : gdata.cfgFile));
		p = cfgQueryWithName("browsers", 1, p);
		all = cfgQueryList(p);
		cnt = arrObjectCount(all);
		for(i=0;i<cnt;i++)
			htabAddObject(browsers, arrObjectAtIndex(all, i));
		arrFree(all);
	}
	marshalError(reply, 0, 0);
	all = htabAllObjects(browsers);
	htabFree(browsers);
	arrSort(all);
	cnt = arrObjectCount(all);
	ipcAddInt(reply, cnt);		/* number of entries */
	for(i=0;i<cnt;i++)
		ipcAddAscii(reply, arrObjectAtIndex(all, i));
	arrFree(all);
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	addBrowser(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8			*name;
cfgProperty_t	*sect, *p;

	if(guimanAllowAccess(MAIN_GUI_PERM_AUTHORIZED, uid)){
		name = ipcGetAscii(rq);
		sect = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
		if(sect == NULL){
			sect = cfgBuildDictProperty();
			cfgSetProperty(CIFS_CONFIG_SECTION, cfgFilePlist(gdata.dbFile), sect);
		}
		p = cfgQueryWithName("browsers", 1, sect);
		if(p == NULL){
			p = cfgBuildDictProperty();
			cfgSetProperty("browsers", sect, p);
		}
		sect = p;
		p = cfgQueryWithName(name, 1, sect);
		if(p == NULL){
			p = cfgBuildDictProperty();
			cfgSetProperty(name, sect, p);
		}
		marshalError(reply, 0, 0);
		cfgUnparse(gdata.dbFile);
		mainModGuiUpdate(browsersDsource);
	}else{
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnotauth);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static void	removeBrowser(struct ipcSrvClient *client, bti32 uid, bti32 gid, void *dsource, ipcMsg_t *rq, ipcMsg_t *reply, void *caller)
{
btu8			*name;
cfgProperty_t	*p;

	if(guimanAllowAccess(MAIN_GUI_PERM_AUTHORIZED, uid)){
		name = ipcGetAscii(rq);
		p = cfgQueryWithName(CIFS_CONFIG_SECTION, 1, cfgFilePlist(gdata.dbFile));
		p = cfgQueryWithName("browsers", 1, p);
		if(p != NULL){
			cfgSetProperty(name, p, NULL);
		}
		marshalError(reply, 0, 0);
		cfgUnparse(gdata.dbFile);
		mainModGuiUpdate(browsersDsource);
	}else{
		marshalError(reply, ERRCLS_FRAMEWORK, ERR_FRWnotauth);
	}
	PROC_IMMEDIATE_RET(caller, NULL);
}

/* ------------------------------------------------------------------------- */

static guiDsourceMethod_t	browserCfgMethods[] = {
	getValues,
	setValues,
	listBrowsers,
	addBrowser,
	removeBrowser,
};

/* ------------------------------------------------------------------------- */

static void	moduleAdded(void *dummy, void *module)
{
module_t	*mod = module;
array_t		*arr, *(*feCfg)(void);
btint		i;

	if(mod->moduleType == MOD_TYPE_FRONTEND){
		arrAddObject(frontendModuleNames, mod->name);
		feCfg = mod->implementationOf(FE_CFGENTRIES_FOR_BACKEND);
		if(feCfg != NULL){
			arr = feCfg();
			for(i=0;i<arrObjectCount(arr);i++){
				arrAddObject(browserCfgDescription.entries, arrObjectAtIndex(arr, i));
			}
			arrFree(arr);
		}
	}
}

/* ------------------------------------------------------------------------- */
/* --------------------------- module functions: --------------------------- */
/* ------------------------------------------------------------------------- */

static struct{
	btu8	*name;
	void	*function;
} functionTable[] = {
	{BE_CONFIG_PATH,			bshareConfigPath},
	{BE_VFS_FOR_PATH,			bshareLookup},
	{BE_VFS_RETAIN,				bshareRetain},
	{BE_VFS_RELEASE,			bshareRelease},
	{BE_VFS_CONNECT,			bshareConnect},
	{BE_VFS_DISCONNECT,			bshareDisconnect},
	{BE_VFS_ROOT_FILE,			bshareRootFile},
	{BE_VFS_STATFS,				bshareStatfs},
	{BE_VFS_PATH,				bshareVfsPath},
	{BE_VFS_CASE_SENSITIVE,		bshareIsCaseSensitive},
	{BE_VFS_OPEN_NOTIFICATION,	bshareOpenNotification},
	{BE_VFS_INVALIDATE_NOTIFICATION,	bshareInvalidateNotification},
	{BE_USER_AUTHENTICATE,		buserAuthenticate},
	{BE_USER_RETAIN,			buserRetain},
	{BE_USER_RELEASE,			buserRelease},
	{BE_FILE_LOOKUP,			bfileLookup},
	{BE_FILE_RETAIN,			bfileRetain},
	{BE_FILE_RELEASE,			bfileRelease},
	{BE_FILE_GETATTR,			bfileGetAttributes},
	{BE_FILE_READDIR,			bfileReaddir},
	{BE_FILE_NAME,				bfileName},
	{BE_FILE_PARENT,			bfileParent},
	{BE_FILE_READLINK,			bfileReadlink},
};

static void	*getCifsBrowseImp(btu8 *functionName)
{
btint	i;

	for(i=0;i<sizeof(functionTable)/sizeof(functionTable[0]);i++){
		if(strcmp(functionName, functionTable[i].name) == 0)
			return functionTable[i].function;
	}
	return NULL;
}

void	cifsBrowse_init(void)
{
module_t	*mod, *m;
array_t		*modules;
btint		i;

    if(bfilePool == NULL)
        bfilePool = poolNew(sizeof(bfileCtx_t), sizeof(bfileCtx_t));
	mod = memCalloc(sizeof(*mod));
	mod->name = "cifsBrowse";
	mod->moduleType = MOD_TYPE_BACKEND;
	mod->implementationOf = getCifsBrowseImp;
	mod->interfaceVersion = 0;
	mod->implementationVersion = MAINCONST_VERSION;
	mod->betaVersion = MAINCONST_BETA_VERSION;
	mod->sortPrecedence = cfgIntValue(cfgQueryWithName("sortPrecedence", 1, cfgQueryWithName(mod->name, 1, cfgFilePlist(gdata.cfgFile))), 1000);
	modRegister(mod);
	frontendModuleNames = arrNew(NULL, 8);
	ncAddObserver(modRegisterNotification(), moduleAdded, NULL);

	browserCfgDescription.entries = arrNew(NULL, 16);
	browserCfgDescription.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(browserCfgEntries);i++){
		arrAddObject(browserCfgDescription.entries, &browserCfgEntries[i]);
	}
	modules = modAllModules();
	for(i=0;i<arrObjectCount(modules);i++){
		m = arrObjectAtIndex(modules, i);
		moduleAdded(NULL, m);
	}
	arrFree(modules);
	for(i=0;i<ELEMENTS(buttonEntries);i++){
		arrAddObject(browserCfgDescription.buttons, &buttonEntries[i]);
	}
	browsersDsource = mainModRegisterDataSource(&browsersDsource, &browserCfgDescription, browserCfgMethods, ELEMENTS(browserCfgMethods));
	mainModAdvertizeConfig(browsersDsource, 30000);
}

/* ------------------------------------------------------------------------- */

