/*
 * Name: dgsocket.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 2001-04-23
 * Tabsize: 4
 * Copyright: (c) 2001 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: dgsocket.c,v 1.6 2001/06/16 19:37:20 cs Exp $
 */


#include "stdheaders.h"
#include "dgsocket.h"
#include "system.h"
#include "blockpool.h"
#include "hashtable.h"
#include "process.h"
#include "cfgparse.h"
#include "evntloop.h"
#include "main.h"
#include "logging.h"
#include "errcodes.h"

/* ------------------------------------------------------------------------- */

typedef struct dgSocketCtx{
	PROC_VARSET;
	dgSocket_t	*sock;
	btu32		remoteAddr;
	btint		remotePort;
	btu8		*data;
	btint		len;
	btint		i;
	btError		err;
}dgSocketCtx_t;

static blockpool_t *dgSocketPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(dgSocketPool, block);
}

/* ------------------------------------------------------------------------- */

typedef struct networkInfo{
	btu32	netBroadcast;
	btu32	netNetmask;
	btu32	netIfAddr;
}networkInfo_t;

static networkInfo_t	netArray[DG_SOCKET_MAX_INTERFACES];
static btint			numNetworks;
static btbool			bindUdpExplicitly;	/* from cifs config section */

static void	updateNetworks(void)
{
sysNetIf_t		ifArray[DG_SOCKET_MAX_INTERFACES];
btint			numInterfaces = -1;
cfgProperty_t	*cfg;
btint			i, j, cnt, len;
btu8			*s, *slash, buffer[32];
bttime			age, now;
static btu32	cfgChg = 0;
static btbool	ifaceConfigured = 0;
static bttime	lastAskKernel = 0;

	if(CFG_DID_CHANGE(cfgChg)){
		lastAskKernel = 0;
		cfg = cfgQueryWithName("main", 1, cfgFilePlist(gdata.cfgFile));
		bindUdpExplicitly = cfgEnumValue(cfgQueryWithName("bindUdpExplicitly", 1, cfg), cfgYesNoEnum, 0);
		cfg = cfgQueryWithName("networkInterfaces", 1, cfg);
		cnt = cfgQueryValueCount(cfg);
		ifaceConfigured = cnt > 0;
		if(ifaceConfigured){
			if(cnt > ELEMENTS(ifArray))
				cnt = ELEMENTS(ifArray);
			for(numInterfaces=i=0;i<cnt;i++){
				s = cfgStringValue(cfgQueryAtIndex(i, cfg), "");
				slash = strchr(s, '/');
				if(slash != NULL){
					len = slash - s;
					if(len >= sizeof(buffer))
						len = sizeof(buffer) - 1;
					memcpy(buffer, s, len);
					buffer[len] = 0;
					ifArray[numInterfaces].ifAddr = inet_addr(buffer);
					len = atoi(slash + 1);
					ifArray[numInterfaces].ifNetmask = 0xffffffff << (32 - len);
					ifArray[numInterfaces].ifBroadcast = ifArray[numInterfaces].ifAddr | ~ifArray[numInterfaces].ifNetmask;
					numInterfaces++;
				}
			}
		}
	}
	if(!ifaceConfigured){
		now = elpNow();
		age = now - lastAskKernel;
		if(lastAskKernel == 0 || age < 0 || age > 10000){
			lastAskKernel = now;
			numInterfaces = sysNetworkInterfaces(ifArray, ELEMENTS(ifArray));
			if(numInterfaces == 0){
				/* if we failed to get kernel interfaces: use generic broadcast */
				ifArray[0].ifAddr = 0;
				ifArray[0].ifNetmask = 0;
				ifArray[0].ifBroadcast = 0xffffffff;
				numInterfaces = 1;
			}
		}
	}
	if(numInterfaces >= 0){
		for(i=cnt=0;i<numInterfaces;i++){
			for(j=0;j<cnt;j++){
				if(ifArray[i].ifBroadcast == netArray[j].netBroadcast)
					break;
			}
			if(j >= cnt){	/* net not already listed */
				netArray[cnt].netBroadcast = ifArray[i].ifBroadcast;
				netArray[cnt].netNetmask = ifArray[i].ifNetmask;
				netArray[cnt++].netIfAddr = ifArray[i].ifAddr;
			}
		}
		numNetworks = cnt;
	}
}

/* ------------------------------------------------------------------------- */

static btint	dgSocketCompare(void *o1, void *o2)
{
dgSocket_t	*s1 = o1, *s2 = o2;

	return s1->localPort - s2->localPort;
}

static btu32	dgSocketHash(void *o)
{
dgSocket_t	*s = o;

	return s->localPort;
}

static collectionObjectI_t	allSocketsInterface = {
	dgSocketCompare,
	dgSocketHash,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

dgSocket_t	*dgSocketForPort(btint localPort)
{
static hashtable_t	*allObjects = NULL;
dgSocket_t			*sock;

    if(dgSocketPool == NULL)
        dgSocketPool = poolNew(sizeof(dgSocketCtx_t), sizeof(dgSocketCtx_t));
	if(allObjects == NULL){
		allObjects = htabNew(&allSocketsInterface, 6, 50);
	}
	sock = htabFindObjectEqualTo(allObjects, &localPort);
	if(sock == NULL){
		sock = memCalloc(sizeof(*sock));
		sock->localPort = localPort;
		sock->msgReceivedSignal = procCondNew();
		htabAddObject(allObjects, sock);
	}
	dgSocketRetain(sock);
	return sock;
}

/* ------------------------------------------------------------------------- */

static void	sockRxEvent(void *obj, void *arg)
{	
dgSocket_t			*sock = obj;
elpFile_t			*info = arg;
dgSocketRx_t		rxData;
struct sockaddr_in	remoteAddr;
btint				slen;

	rxData.unixError = 0;
	slen = sizeof(remoteAddr);
	rxData.len = recvfrom(info->fd, rxData.data, sizeof(rxData.data), 0, (struct sockaddr *)&remoteAddr, &slen);
	if(rxData.len < 0){
		rxData.unixError = errno;
	}else{
		rxData.remoteIp = ntohl(remoteAddr.sin_addr.s_addr);
		rxData.remotePort = ntohs(remoteAddr.sin_port);
	}
	procCondBroadcast(sock->msgReceivedSignal, &rxData);
}

/* ------------------------------------------------------------------------- */

btint	dgSocketRetain(dgSocket_t *sock)
{
btint				cnt, i, x, error = 0;
struct sockaddr_in	localAddr;

	if(sock->refCount++ == 0){
		updateNetworks();
		sock->numSocks = 0;
		memset(&localAddr, 0, sizeof(localAddr));
		localAddr.sin_family = AF_INET;
		localAddr.sin_port = htons(sock->localPort);
		/* if there is only one network, don't bind explicitly */
		if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
			logPrintf(LOGM_NBNS, "dgsocket: numNetworks = %d bindUdpExplicitly = %d\n", numNetworks, bindUdpExplicitly);
		cnt = bindUdpExplicitly ? numNetworks : 0;
		for(i=-1;i<cnt;i++){
			if((sock->socks[sock->numSocks].fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
				error = errno;
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "dgsocket: socket(): [%d] %s\n", error, sysStrerror(error));
				goto errorOccured;
			}
			if(bindUdpExplicitly){
				x = 1;
				x = setsockopt(sock->socks[sock->numSocks].fd, SOL_SOCKET, SO_REUSEADDR, (void *)&x, sizeof(x));
				if(x < 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "dgsocket: error setting socket reusable: [%d] %s\n", errno, sysStrerror(errno));
			}
			localAddr.sin_addr.s_addr = i < 0 ? 0 : htonl(netArray[i].netIfAddr);
			x = bind(sock->socks[sock->numSocks].fd, (struct sockaddr *)&localAddr, sizeof(localAddr));
			if(i < 0){
				if(x != 0 && LOG_MODE_IS_ACTIVE(LOGM_NBNS))
					logPrintf(LOGM_NBNS, "dgsocket: ignoring error binding to port %d on 0x%08x: [%d] %s\n", ntohs(localAddr.sin_port), ntohl(localAddr.sin_addr.s_addr), errno, sysStrerror(errno));
				if(x != 0 && cnt > 0){	/* we want to bind other interfaces to same random port */
					localAddr.sin_port = 0;
					x = bind(sock->socks[sock->numSocks].fd, (struct sockaddr *)&localAddr, sizeof(localAddr));
					if(x != 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
						logPrintf(LOGM_ERROR, "dgsocket: error binding to address: 0x%08x [%d] %s\n", ntohl(localAddr.sin_addr.s_addr), errno, sysStrerror(errno));
					x = sizeof(localAddr);
					getsockname(sock->socks[sock->numSocks].fd, (struct sockaddr *)&localAddr, (void *)&x);
					if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
						logPrintf(LOGM_NBNS, "dgsocket: will bind to local port %d\n", ntohs(localAddr.sin_port));
				}
			}else{
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "dgsocket: error binding to port %d on 0x%08x: [%d] %s\n", ntohs(localAddr.sin_port), ntohl(localAddr.sin_addr.s_addr), errno, sysStrerror(errno));
				close(sock->socks[sock->numSocks].fd);
				continue;	/* simply skip this interface/address */
			}
#ifdef SO_BROADCAST
			x = 1;	/* switch on broadcast permission */
			if(setsockopt(sock->socks[sock->numSocks].fd, SOL_SOCKET, SO_BROADCAST, (void *)&x, sizeof(x)) < 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "dgsocket: Error setting broadcast option on socket: [%d] %s\n", errno, sysStrerror(errno));
#endif
			if(sysSetNonblock(sock->socks[sock->numSocks].fd) != 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "dgsocket: error setting socket nonblock: [%d] %s\n", errno, sysStrerror(errno));
			elpRegisterForFd(sock->socks[sock->numSocks].fd, ELP_FDMODE_READ, sockRxEvent, sock);
			sock->socks[sock->numSocks].netmask = i < 0 ? 0 : netArray[i].netNetmask;
			sock->socks[sock->numSocks].localIp = i < 0 ? 0 : netArray[i].netIfAddr;
			sock->numSocks++;
		}
	}
	return 0;
errorOccured:
	for(i=0;i<sock->numSocks;i++){
		elpCancelForFd(sock->socks[i].fd, -1);
		close(sock->socks[i].fd);
	}
	sock->numSocks = 0;
	return error;
}

/* ------------------------------------------------------------------------- */

void	dgSocketRelease(dgSocket_t *sock)
{
btint	i;
	if(--sock->refCount == 0){
		for(i=0;i<sock->numSocks;i++){
			elpCancelForFd(sock->socks[i].fd, -1);
			close(sock->socks[i].fd);
		}
		sock->numSocks = 0;
		if(procCondWaiting(sock->msgReceivedSignal)){
			logPrintf(LOGM_ERROR, "dgsocket: ref count going to zero while process waiting!\n");
			DEBUG_TRAP();
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	dgSocketSendProc(void *obj, void *arg)
{
dgSocketCtx_t		*this = obj;
dgSocket_t			*sock = this->sock;
struct sockaddr_in	remoteAddr;
btError				err;
btint				i;

	PROC_BEGIN;
	memset(&remoteAddr, 0, sizeof(remoteAddr));
	remoteAddr.sin_family = AF_INET;
	remoteAddr.sin_port = htons(this->remotePort);
	remoteAddr.sin_addr.s_addr = htonl(this->remoteAddr);
	for(i=1;i<sock->numSocks;i++){	/* skip generic socket since it matches anything */
		if(((sock->socks[i].localIp ^ this->remoteAddr) & sock->socks[i].netmask) == 0)
			break;
	}
	this->i = i < sock->numSocks ? sock->socks[i].fd : sock->socks[0].fd;	/* default: generic socket */
	if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
		logPrintf(LOGM_NBNS, "dgsocket: sending to 0x%08x:%d over interface 0x%08x\n", this->remoteAddr, this->remotePort, sock->socks[i < sock->numSocks ? i : 0].localIp);
retry:
	if(sendto(this->i, this->data, this->len, 0, (struct sockaddr *)&remoteAddr, sizeof(remoteAddr)) < 0){
		if(errno == EWOULDBLOCK){
			if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
				logPrintf(LOGM_NBNS, "dgsocket: sendto() would block, waiting...\n");
			PROC_WAITFOR_X(1, elpRegisterForFd(this->i, ELP_FDMODE_WRITE, dgSocketSendProc, this));
			elpCancelForFd(this->i, ELP_FDMODE_WRITE);
			goto retry;
		}
		err.class = ERRCLS_UNIX;
		err.code = errno;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "dgsocket: sendto(%d): [%d] %s\n", this->remotePort, errno, sysStrerror(errno));
		goto errorOccured;
	}
	memset(&err, 0, sizeof(err));
errorOccured:
	PROC_RETURN(&err);
	PROC_END;
}

void	dgSocketSend(void *caller, dgSocket_t *sock, btu32 ipAddr, btint remotePort, btu8 *data, btint len)
{
dgSocketCtx_t	*ctx;

	PROC_INIT(ctx, dgSocketSendProc, caller, dgSocketPool, poolFreeBlock);
	ctx->sock = sock;
	ctx->remoteAddr = ipAddr;
	ctx->remotePort = remotePort;
	ctx->data = data;
	ctx->len = len;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	dgSocketBroadcastProc(void *obj, void *arg)
{
dgSocketCtx_t		*this = obj;
dgSocket_t			*sock = this->sock;
btError				*err = arg;

	PROC_BEGIN;
	for(this->i = 0; this->i < numNetworks; this->i++){
		PROC_WAITFOR_X(1, dgSocketSend(this, sock, netArray[this->i].netBroadcast, this->remotePort, this->data, this->len));
		if(err->class != 0)
			this->err = *err;	/* only return last error */
	}
	PROC_RETURN(&this->err);
	PROC_END;
}

void	dgSocketBroadcast(void *caller, dgSocket_t *sock, btint remotePort, btu8 *data, btint len)
{
dgSocketCtx_t	*ctx;

	PROC_INIT(ctx, dgSocketBroadcastProc, caller, dgSocketPool, poolFreeBlock);
	ctx->sock = sock;
	ctx->remotePort = remotePort;
	ctx->data = data;
	ctx->len = len;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */
