/*
 * Name: dircache.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-03
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: dircache.h,v 1.3 1999/07/26 20:06:17 cs Exp $
 */

#ifndef __dircache_H_INCLUDED__
#define __dircache_H_INCLUDED__

#include "basictypes.h"
#include "cache.h"
#include "lru.h"
#include "file.h"

/*
General Description:
This class implements the directory cache. It is based on the functions
fileReadDir() and fileCloseReaddir() from file.c, to which it is statically
bound. It's a subclass of cacheSuper (see file cache.h).
*/

#define	DIRCACHE_SIZE	128

/* ------------------------------------------------------------------------- */

typedef btbool	(*dirCallback_t)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir);

typedef struct dirRecord{
    btuni	*name;
    btbool	isDir;
}dirRecord_t;

typedef struct dirCache{
	cacheSuper_t		general;
    bttime				lastUpdated;

    btu8				resumeKey[24];	/* needed for core directory search */
    btsize				resumeKeySize;	/* size of resume key */
    btbool				noMoreEntries;
	btbool				dirWasModified;

    btint				headIndex;		/* where next read entry goes */
    btint				tailIndex;		/* oldest entry */
    btint				headPtr;
    btint				tailPtr;
    dirRecord_t			records[DIRCACHE_SIZE];
}dirCache_t;

/* ------------------------------------------------------------------------- */
dirCache_t	*dirCacheNew(void);
/* This function creates a new directory cache object
 */
void	dirCacheFree(void *cache);
/* This function frees a directory cache object
 */
void	dirCacheRead(dirCache_t *cache, struct cifsUser *user, btint startIndex, dirCallback_t resCallback, void *caller);
/* This function reads entries through the cache, starting at index
 * 'startIndex'. All entries are returned through calls to 'resCallback', as
 * long as resCallback() returns 1. dirCacheRead() returns when resCallback()
 * returns 0 or if no more entries are available.
 * Callback: beFileRes_t: err, eof
 */
void	dirCacheFileAdded(dirCache_t *cache, btuni *file);
/* This function tells the cache that a file has been created in the
 * directory. The directory cache can decide how to handle this event. It
 * turned out to be unwise to declare the cache invalid upon this event.
 * Parameters: 'cache' is a pointer to the dirCache_t object and 'file' is
 *     the name of the file which has been added.
 */
void	dirCacheFileRemoved(dirCache_t *cache, btuni *file);
/* This function tells the cache that a file has been removed from the
 * directory. The cache can decide how to handle this event. It's probably
 * best to scan the cache for occurences of the file and purge them from
 * the cache.
 * Parameters: 'cache' is a pointer to the dirCache_t object and 'file' is
 *     the name of the file which has been removed.
 */
void	dirCacheDirWasModified(dirCache_t *cache);
/* This function is called when other modules think that the underlying
 * directory has been modified, e.g. because the modification date of the
 * directory has changed. The cache should invalidate to avoid the following
 * problem: Sharity currently works through NFS2. It updates the cache only
 * if the directory modification date has changed. If the directory changes,
 * the kernel first reads the directory date and then the directory contents.
 * If the cache still caches the old contents, the new contents will never
 * be read because the kernel associates the old contents with the new
 * modification date.
 * There's an other problem, though: This function should NOT invalidate
 * the cache, because this will render all NFS2 cookies invalid if files
 * have been added or removed.
 * As a workaround [hack], this function currently does the following: If the
 * next request starts at index 0, the cache is invalidated. If it starts at
 * an other index, it is not invalidated. If the kernel re-reads the directory
 * because it may have changed, it will start at index 0 and the cache is
 * invalidated as it should. If it does not re-read the contents but starts
 * at index 0 incidentally, there's no problem because there's no cookie
 * context built up so far. If it continues an old read, the cache is NOT
 * invalidated.
 */ 
/* ------------------------------------------------------------------------- */

#endif /* __dircache_H_INCLUDED__ */
