/*
 * Name: file.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-02
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: file.h,v 1.3 1999/07/22 17:53:57 cs Exp $
 */

#ifndef __file_H_INCLUDED__
#define __file_H_INCLUDED__

#include "basictypes.h"
#include "lru.h"

/*
General Description:
This module implements the "file" class. This is the most central class of
the project as it represents data files, links, devices and directories.
All file requests from the client originate in "nfs.c" and are delegated
into this module. "file.c", on the other hand, uses the SMB RPCs defined
in "cifs.c" to implement the file operations. A file object can't live by
it's own. It must be part of a share (see "share.h"), which, in turn, must
be part of a server (see "server.h"). The "part of" relation is implemented
by container data types.
*/

/* ------------------------------------------------------------------------- */

/* open modes: */
#define	FILE_CLOSED		BE_FILE_CLOSED
#define	FILE_OPEN_READ	BE_FILE_OPEN_READ
#define	FILE_OPEN_WRITE	BE_FILE_OPEN_WRITE
#define	FILE_OPEN_RW	BE_FILE_OPEN_RW

/* operations: */
#define	FILE_OP_NONE	BE_FILE_OP_NONE
#define	FILE_OP_READ	BE_FILE_OP_READ
#define	FILE_OP_WRITE	BE_FILE_OP_WRITE

/* ------------------------------------------------------------------------- */

struct share;
struct dcache;
struct dirCache;
struct cifsUser;
struct dirRecord;
struct beFileSetAttr;

typedef btbool	(*fileReaddirCallback_t)(void *ref, struct dirRecord *record);

/* ------------------------------------------------------------------------- */

typedef struct cifsFile{
    lruObject_t			lookupLru;		/* lru-pointers for pre-looked up files */
    bti32				refCount;		/* number of references (not by collections) */
    struct cifsFile		*parent;		/* parent directory, NULL for root */
	struct hashtable	*children;		/* if directory: list of children, NULL otherwise */
	struct share		*share;			/* share this file belongs to */
	btuni				*name;			/* name within directory */
	void				*attributes;	/* polymorphic, depends on underlying FS */
	btbool				isInvalid;		/* file has been invalidated */

	btu8				openMode;		/* one of FILE_CLOSED, FILE_OPEN_READ, FILE_OPEN_RW */
	btbool				doAutoclose;	/* whether we are responsible for closing */
	btu8				oplockLevel;	/* oplock level owned by the file */
	btu16				openedBy;		/* if the node is open: CIFS user that opened it */
	btu32				fid;			/* file-ID, if openMode != FILE_CLOSED */
	bttime				lastWrite;		/* time of last write to the file, if != 0 */
	btint				openLock;		/* if set, open mode must not be touched */
	bti32				needOpen;		/* counts processes that need file open, advisory */
	struct procCond		*openSignal;	/* fires if openLock is released */

#if 0	/* not currently implemented: */
	btError				writebackErr;	/* accumulates errors that are not directly returned */
#endif

	union{
		struct dcache		*data;
		struct dirCache		*dir;
		void				*any;
	}					cache;			/* pointer to the cache, if there is one */
}cifsFile_t;

/* ------------------------------------------------------------------------- */

typedef struct filePrivRes{
	btError	err;
/* for readdir: */
	btsize			resumeKeySize;
	btu8			*resumeKey;
/* for setOpenMode: */
    btu8			*data;
    btu32			dataLen;
	btbool			didRW;
}filePrivRes_t;
/* used by fileReadDirUncached, fileSetOpenMode, fileForceOpenWrite, fileRWUncached */

/* ------------------------------------------------------------------------- */

cifsFile_t	*fileNew(struct share *share, btuni *name);
/* Allocates and initializes a new file structure, but does not add it
 * to any tables.
 * Returns: the newly allocated file object.
 */
void	fileFree(cifsFile_t *file);
/* Frees all resources used by the file representation 'file'.
 * Parameters: 'file' is a pointer to the file representation to be freed.
 */
void	fileIsDirectory(cifsFile_t *file);
/* This function is called from the attribute representation if it turns
 * out that the file is a directory. It allocates the cache hashtable
 * holding all allocated children.
 */
/* inline void	fileRetain(cifsFile_t *file);*/
/* Increments the retain count.
 */
/* inline void	fileRelease(cifsFile_t *file);*/
/* Decrements the file's retain count. If it reaches zero, the file is
 * free'd.
 */
void	fileRegisterChild(cifsFile_t *file, cifsFile_t *child);
/* Adds 'child' to the list of children in 'this', sets the 'parent' entry
 * of 'child' to 'this'.
 */
void	fileUnregisterChild(cifsFile_t *file, cifsFile_t *child, btbool bidirectional);
/* Removes 'child' from the list of children in 'file'.
 */
void	fileInvalidate(cifsFile_t *file);
/* Remove the file from all tables except the parent/child relations, make
 * all further operations fail for this file. This also invalidates all
 * children of this file.
 */
/* inline void	fileSetCache(cifsFile_t *file, void *cache); */
/* Sets the 'cache' instance variable, does nothing more
*/
/* inline void	*fileCache(cifsFile_t *file); */
/* Returns the 'cache' instance variable.
 */
btuni	*fileBuildPath(cifsFile_t *file, void *append, btbool isUnicode, btuni *buffer, btsize bufSize);
/* This function builds the absolute path of a given file. It is able to
 * append an arbitrary string to the path, which may be given in local ASCII
 * representation or in Unicode. The result is written within 'buffer', which
 * is'bufSize' Unicode characters long. A pointer to the beginning of the
 * path is returned. This pointer is within 'buffer', but is not the start of
 * 'buffer'.
 */
/* ------------------------------------------------------------------------- */
void	fileLookupChildWithName(cifsFile_t *file, btuni *name, struct cifsUser *user, void *caller);
/* Allocates (if necessary) and fully initializes the child with name 'name'
 * and returns a pointer to its structure.
 * Callback: err, file
 */
void	fileGetAttributes(cifsFile_t *file, struct cifsUser *user, void *caller);
/* Returns the attributes of the file and loads them if necessary.
 * Callback: err, attr
 */
void	fileSetAttributes(cifsFile_t *file, struct cifsUser *user, struct beFileSetAttr *setAttr, void *caller);
/* Sets the attributes of the file and synchronizes them with the server.
 * Callback: err
 */
void	fileReadLink(cifsFile_t *file, struct cifsUser *user, void *caller);
/* Reads content of a file of type "symbolic link".
 * Callback: err, path
 */
/* ------------------------------------------------------------------------- */
void	fileReadDirUncached(cifsFile_t *file, struct cifsUser *user, btu8 resumeKey[24], btsize resumeKeySize, btsize maxCount, fileReaddirCallback_t resCb, void *caller);
/* Reads entries from the directory 'file' and passes the data to
 * 'resCb'. If resumeKeySize == 0, the search starts at the beginning.
 * If the returned resume key size is 0, there are no more entries.
 * The resumeKey array must be valid until this function returns.
 * Callback: filePrivRes_t: err, resumeKey, resumeKeySize
 */
void	fileCloseReaddir(cifsFile_t *file, btu16 cifsUid, btu8 resumeKey[24], btsize resumeKeySize, void *caller);
/* This function must be called if fileReadDirUncached() is not called
 * repeatedly until it returns a resumeKeySize of 0.
 * Callback: err
 */
void	fileReadDir(cifsFile_t *file, struct cifsUser *user, btu32 startIndex, btbool (*resCb)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir), void *caller);
/* Abstraction around fileReadDirUncached(), gives random access to the
 * directory and caches the data. It tries to open the directory as file
 * and set oplocks as an indication for modifications to the directory.
 * Callback: err, eof
 */
void	fileCreateUnix(cifsFile_t *file, btuni *name, btbool isDir, btbool isLink, struct cifsUser *user, struct beFileSetAttr *attr, void *caller);
/* Creates a file, directory or faked softlink with name 'name' in the
 * directory 'this'. The initial attributes are set to 'attr' and files are
 * opened for read/write.
 * Callback: err, file (except if directory was created)
 */
void	fileRemoveChild(cifsFile_t *file, btuni *name, btbool isDir, struct cifsUser *user, void *caller);
/* Removes the file or directory with name 'name' from directory 'this' and
 * invalidates it.
 * Callback: err
 */
void	fileRename(cifsFile_t *fromDir, btuni *fromName, cifsFile_t *toDir, btuni *toName, struct cifsUser *user, void *caller);
/* Renames the file or directory with name 'fromName' in directory 'fromDir'
 * to 'toName' in directory 'toDir', invalidating any existing cached
 * structures.
 * Callback: err
 */
void	fileHardLink(cifsFile_t *sourceFile, cifsFile_t *newDir, btuni *newName, struct cifsUser *user, void *caller);
/* Makes a new directory entry for file 'sourceFile' in the directory
 * 'newDir' under the name 'newName'.
 * Callback:
 */
void	fileSymbolicLink(cifsFile_t *newDir, btuni *newName, btuni *path, struct cifsUser *user, struct beFileSetAttr *attr, void *caller);
/* creates a file of type "symbolic link" in 'newDir' under the name
 * 'newName', pointing to the path 'path'.
 * Callback:
 */
/* ------------------------------------------------------------------------- */
void	fileSetOpenMode(cifsFile_t *file, btu16 cifsOpener, btu8 desiredMode, btu8 relevantMask, btbool autoclose, int additionalOperation, btu8 *src, btoffset offset, btsize byteCount, btbool writeThrough, btbool noLocking, void *caller);
/* Opens or closes a file. The bits of the open mode that are set in
 * 'relevantMask' are set to the values given in 'desiredMode' and
 * data is read or written (depending on 'additionalOperation') if the
 * open can be batched with read or write.
 * Callback: err, didRW, (dataLen if read or written), (data if read)
 */
void	fileForceOpenWrite(cifsFile_t *file, struct cifsUser *user, btint additionalOperation, btu8 *src, btoffset offset, btsize byteCount, btbool writeThrough, void *caller);
/* Like fileSetOpenMode(), but the file is always opened for write, even if
 * it is write protected.
 * Callback: err, didRW, (dataLen if read or written), (data if read)
 */
/* inline int		fileOpenMode(cifsFile_t *file);*/
/* returns the current open mode of the file.
 */
void	fileOplockBreak(cifsFile_t *file, btint mode);
/* This function is called when the server breaks an OPLOCK held by the
 * file.
 */
void	fileRWUncached(cifsFile_t *file, int operation, btbool writeThrough, btu8 *src, btoffset offset, btsize len, struct cifsUser *user, void *caller);
/* Reads or writes to/from the given file. Opens or re-opens the file if
 * necessary.
 * In read mode, 'writeThrough' indicates that a write will follow and the
 * file should be opened for write if it is closed.
 * Callback: err, dataLen, (data if read)
 */
void	fileRW(cifsFile_t *file, int operation, btbool writeThrough, btu8 *buffer, btoffset offset, btsize len, struct cifsUser *user, void *caller);
/* Reads or writes the given file, satisfies the request from the cache, if
 * possible and opens the file for the desired mode if necessary. Read data
 * is copied into the 'buffer' and not returned in the callback.
 * Callback: err, dataLen
 */
void	fileTruncate(cifsFile_t *file, btoffset newSize, struct cifsUser *user, void *caller);
/* Truncates (or enlarges) the file 'this' to the given size, sets the
 * open-mode to R/W if necessary.
 * Callback: err
 */
void	fileConnectionLost(cifsFile_t *file);
/* This function is called by the transport layers to inform the file that
 * the connection to the server was lost. The file is expected to set its
 * internal state appropriately.
 */
/* ------------------------------------------------------------------------- */
/* ---------------------------- inline methods ----------------------------- */
/* ------------------------------------------------------------------------- */

static inline void	fileRetain(cifsFile_t *file)
{
	file->refCount++;
}

static inline void	fileRelease(cifsFile_t *file)
{
	file->refCount--;
	if(file->refCount == 0){
		fileFree(file);
	}else if(file->refCount < 0){
		DEBUG_TRAP();
	}
}

static inline void	fileSetCache(cifsFile_t *file, void *cache)
{
	file->cache.any = cache;
}

static inline void	*fileCache(cifsFile_t *file)
{
	return file->cache.any;
}

static inline int	fileOpenMode(cifsFile_t *file)
{
	return file->openMode;
}

static inline struct share	*fileShare(cifsFile_t *file)
{
	return file->share;
}

static inline btuni	*fileName(cifsFile_t *file)
{
	return file->name;
}

static inline cifsFile_t	*fileParent(cifsFile_t *file)
{
	if(file->parent != NULL)
		return file->parent;
	return file;
}

static inline btu32	fileFid(cifsFile_t *file)
{
	return file->fid;
}

/* use defines to avoid including all the headers: */
#define	fileMaxDirCacheAge(file)	(srvDirCacheTime(shrServer((((cifsFile_t *)(file))->share))))

#define	fileMaxDataCacheAge(file)	(srvCacheValidDelay(shrServer(((cifsFile_t *)(file))->share), file->oplockLevel))

/* ------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */

#endif /* __file_H_INCLUDED__ */
