/*
 * Name: lmb.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-01-03
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: lmb.c,v 1.9 2002/04/05 17:53:49 cs Exp $
 */

#include "stdheaders.h"
#include "hashtable.h"
#include "process.h"
#include "memalloc.h"
#include "evntloop.h"
#include "server.h"
#include "nbnsclient.h"
#include "nbdgs.h"
#include "blockpool.h"
#include "util.h"
#include "unicode.h"
#include "lmb.h"
#include "cfgparse.h"
#include "main.h"
#include "cifsmod.h"
#include "errcodes.h"
#include "logging.h"

/* ------------------------------------------------------------------------- */

typedef struct lmb{
	btu8		*domain;
	btu8		*srvName;
	bttime		createdAt;
	btint		refCount;
}lmb_t;

/* ------------------------------------------------------------------------- */

typedef struct{
	PROC_VARSET;
	btu8		domain[64];
	server_t	*srv;
	btint		retryCount;
	btu32		ip;
	btu8		*name;
	btbool		forceDetection;
}lmbCtx_t;

static blockpool_t *lmbPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(lmbPool, block);
}

/* ------------------------------------------------------------------------- */

static hashtable_t	*lmbsByDomain = NULL;
static hashtable_t	*lmbServers = NULL;

/* ------------------------------------------------------------------------- */

static void	renewLmbEntry(void *obj, void *arg)
{
lmb_t	*lmb = obj;

	lmbForDomain(lmb->domain, 1, NULL);	/* do new lookup in background */
}

static void	lmbFree(lmb_t *lmb)
{
	if(lmb != NULL){
		elpCancelCallAt(renewLmbEntry, lmb);
		if(lmb->srvName != NULL){
			lmbRemoveServer(lmbLookupServer(lmb->srvName));
			memFree(lmb->srvName);
		}
		if(lmb->domain != NULL)
			memFree(lmb->domain);
		memset(lmb, 0, sizeof(*lmb));
		memFree(lmb);
	}
}

static lmb_t	*lmbNew(btu8 *volatile domain, btu8 *volatile srvName, bttime maxAge)
{
lmb_t	*volatile lmb;
bttime	callTime;

EXC_TRY
	lmb = memCalloc(sizeof(*lmb));
	lmb->refCount = 0;
	lmb->domain = utilStrDup(domain);
	lmb->srvName = utilStrDup(srvName);
	lmb->createdAt = elpNow();
	if(maxAge > 60000){
		callTime = elpNow() + maxAge - 20000;	/* renew 20 seconds before expiry */
		elpCallAtCancelPrev(callTime, renewLmbEntry, lmb);
	}
EXC_CATCH
	lmbFree(lmb);
	EXC_RETHROW();
EXC_ENDTRY
	return lmb;
}

static void	lmbRetain(void *obj)
{
lmb_t	*lmb = obj;

	lmb->refCount++;
}

static void	lmbRelease(void *obj)
{
lmb_t	*lmb = obj;

	lmb->refCount--;
    if(lmb->refCount == 0){
        lmbFree(lmb);
    }else if(lmb->refCount < 0){
		DEBUG_TRAP();
	}
}

static btint	lmbCompare(void *obj1, void *obj2)
{
lmb_t	*this = obj1, *other = obj2;

	return utilStrCmp(this->domain, other->domain, 0);
}

static btu32	lmbHashVal(void *obj)
{
lmb_t	*lmb = obj;

	return utilStrHash(lmb->domain, 0);
}

/* ------------------------------------------------------------------------- */

static btint	srvCompare(void *obj1, void *obj2)
{
	return utilStrCmp(srvName(obj1), srvName(obj2), 0);
}
 
static btu32	srvHashVal(void *obj)
{
	return utilStrHash(srvName(obj), 0);
}

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	lmbsByDomainInterface = {
	lmbCompare,
	lmbHashVal,
	lmbRetain,
	lmbRelease
};

static collectionObjectI_t	lmbServersInterface = {
	srvCompare,
	srvHashVal,
	srvRetain,
	srvRelease
};

/* ------------------------------------------------------------------------- */

static void	reverseLookupProc(void *obj, void *arg)
{
lmbCtx_t	*this = obj;
nbnsRes_t	*rval = arg;

	PROC_BEGIN;
	PROC_WAITFOR(nbnsNodeStatus(this->ip, this->name, this->retryCount, this));
	if(rval->err.class == 0)
		goto foundName;
	if(this->name[0] != '*' && rval->err.class == ERRCLS_CIFS_LOCAL && rval->err.code == ERR_CIFS_LOClookup){
		PROC_WAITFOR(nbnsNodeStatus(this->ip, "*", 0, this));
		if(rval->err.class == 0)
			goto foundName;
	}
	if(this->forceDetection && rval->err.class == ERRCLS_CIFS_LOCAL && rval->err.code == ERR_CIFS_LOCtimeout){
		/* try broadcasting node status request if unicast failed */
		PROC_WAITFOR(nbnsNodeStatus(0xffffffff, this->name, this->retryCount, this));
	}
foundName:
	PROC_RETURN(rval);
	PROC_END
}

static void	reverseLookup(btu32 ip, btu8 *knownName, btint knownType, btbool allowBroadcast, void *caller)
{
lmbCtx_t	*ctx;

	PROC_INIT(ctx, reverseLookupProc, caller, lmbPool, poolFreeBlock);
	ctx->ip = ip;
	ctx->name = knownName;
	ctx->retryCount = knownType;			/* misuse varable */
	ctx->forceDetection = allowBroadcast;	/* misuse variable */
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	lmbForDomainProc(void *obj, void *arg)
{
lmbCtx_t			*this = obj;
lmb_t				*lmb;
bttime				now, age;
btu8				*srvName, *domainp;
btbool				capitalizeServer = 0;
nbnsRes_t			*rval = arg;
srvRes_t			*srval = arg;
lmbRes_t			res;
btint				i, numEntries;
cfgProperty_t		*path[8];
static btu32		cfgChg = 0;
static bttime		maxAge;
static bttime		electionHoldoff;
static btint		maxRetries;
static btbool		lmbDetectRunning = 0;
static procCond_t	*lmbDetectSignal = NULL;
static bttime		lastLmbElection = 0;

	PROC_BEGIN;
	if(this->domain[0] == 0){	/* no domain given */
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCtimeout;
		PROC_RETURN(&res);
	}
	if(lmbDetectSignal == NULL)
		lmbDetectSignal = procCondNew();
	if(CFG_DID_CHANGE(cfgChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		maxAge = cfgIntValue(cfgQueryFromArray("lmbCacheTime", numEntries, path), 300000);
		electionHoldoff = cfgIntValue(cfgQueryFromArray("lmbElectionHoldoff", numEntries, path), 300000);
		maxRetries = cfgIntValue(cfgQueryFromArray("lmbLookupRetries", numEntries, path), 7);
	}
	srvName = NULL;
	if(!this->forceDetection){
		domainp = this->domain;
		if(lmbsByDomain != NULL && (lmb = htabFindObjectEqualTo(lmbsByDomain, &domainp)) != NULL){
			now = elpNow();
			age = now - lmb->createdAt;
			if(age >= 0 && age < maxAge){
				srvName = lmb->srvName;
				goto foundLMB;
			}else{
				htabRemoveObjectEqualTo(lmbsByDomain, lmb);	/* is outdated */
			}
		}
	}
	while(lmbDetectRunning){
		PROC_WAITFOR_X(9, procCondWait(lmbDetectSignal, this));
	}
	lmbDetectRunning = 1;
retry:
	/* first method: try broadcasting for domain name */
	PROC_WAITFOR_X(1, nbnsResolve(this->domain, 0x1d, 0xffffffff, 1, this));
	if(rval->err.class == 0){
		PROC_WAITFOR_X(2, reverseLookup(rval->ipAddress, this->domain, 0x1d, 1, this));
		if(rval->err.class == 0){
			srvName = rval->name;
			capitalizeServer = 1;
			if(LOG_MODE_IS_ACTIVE(LOGM_NBNS))
				logPrintf(LOGM_NBNS, "lmb: found LMB = %s\n", srvName);
			goto foundLMB;
		}
	}
	/* second method: initiate LMB election, wait and retry */
	now = elpNow();
	age = now - lastLmbElection;
	if(lastLmbElection == 0 || age > electionHoldoff || age < 0){
		lastLmbElection = now;
		PROC_WAITFOR_X(4, nbdgsSendRequestBrowserElection(this, this->domain));
/* We are aware of the fact that the browser election will take more than 10
 * seconds and that we will probably use the domain master browse server for
 * the next 5 minutes (until the LMB times out). By using the Domain Browser
 * we avoid the annoying 20 second delay for the user. This should happen only
 * in rare cases anyway.
 */
	}
	/* third method: try to find domain master browser */
	if(this->retryCount < 2){	/* if it failed at least once, don't try again */
		PROC_WAITFOR_X(6, nbnsResolve(this->domain, 0x1b, nbnsWinsIp(), 0, this));
		if(rval->err.class == 0){
			PROC_WAITFOR_X(7, reverseLookup(rval->ipAddress, this->domain, 0x1b, 0, this));
			if(rval->err.class == 0){
				srvName = rval->name;
				capitalizeServer = 1;
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "lmb: failed to find LMB, using Domain Master Browser %s\n", srvName);
				goto foundLMB;
			}
		}
	}
	if(++this->retryCount < maxRetries)
		goto retry;
	res.err.class = ERRCLS_CIFS_LOCAL;
	res.err.code = ERR_CIFS_LOCtimeout;
	goto doReturn;
foundLMB:
	if(capitalizeServer){
		for(i = 0; i < 15 && srvName[i] != 0; i++){
			if(i == 0)
				srvName[i] = asciiCharToupper(srvName[i]);
			else
				srvName[i] = asciiCharTolower(srvName[i]);
		}
		srvName[i] = 0;
	}
	this->srv = srvWithName(srvName, 1);
	lmbAddServer(this->srv, this->domain, maxAge);
	PROC_WAITFOR_X(8, srvConnect(this->srv, 0, this));
	if(srval->err.class != 0){
		lmbRemoveServer(this->srv);
		srvRelease(this->srv);
		if(++this->retryCount < maxRetries)
			goto retry;
	}
	res.err = srval->err;
	res.server = this->srv;
doReturn:
	lmbDetectRunning = 0;
	procCondSignal(lmbDetectSignal, NULL);
	PROC_RETURN(&res);
	PROC_END;
}

void	lmbForDomain(btu8 *domain, btbool forceDetection, void *caller)
{
lmbCtx_t	*ctx;

	if(lmbPool == NULL)
		lmbPool = poolNew(sizeof(lmbCtx_t), sizeof(lmbCtx_t));
	PROC_INIT(ctx, lmbForDomainProc, caller, lmbPool, poolFreeBlock);
	if(strlen(domain) >= sizeof(ctx->domain))
		domain[sizeof(ctx->domain) - 1] = 0;	/* emergency brake */
	strcpy(ctx->domain, domain);
	ctx->forceDetection = forceDetection;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	lmbAddServer(struct server *srv, btu8 *domain, bttime maxAge)
{

	if(lmbsByDomain == NULL)
		lmbsByDomain = htabNew(&lmbsByDomainInterface, 4, 75);
	if(lmbServers == NULL)
		lmbServers = htabNew(&lmbServersInterface, 4, 75);
	htabAddObject(lmbsByDomain, lmbNew(domain, srvName(srv), maxAge));
	htabAddObject(lmbServers, srv);
}

/* ------------------------------------------------------------------------- */

void	lmbRemoveServer(struct server *srv)
{
	if(lmbServers != NULL && srv != NULL)
		htabRemoveObjectEqualTo(lmbServers, srv);
}

/* ------------------------------------------------------------------------- */

struct server	*lmbLookupServer(btu8 *srvName)
{
	if(lmbServers == NULL)
		return NULL;
	return htabFindObjectEqualTo(lmbServers, &srvName);
}

/* ------------------------------------------------------------------------- */

struct array	*lmbAllServers(void)
{
	if(lmbServers == NULL){
		return arrNew(&lmbServersInterface, 1);
	}else{
		return htabAllObjects(lmbServers);
	}
}

/* ------------------------------------------------------------------------- */

