/*
 * Name: message.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-21
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: message.c,v 1.1.1.1 1999/04/12 13:56:35 cs Exp $
 */


#include "stdheaders.h"
#include "logging.h"
#include "memalloc.h"
#include "message.h"

/* ------------------------------------------------------------------------- */

#define	MAX_USAGE_COUNT	500

/* ------------------------------------------------------------------------- */

fifo_t	*messageStore = NULL;

static void	msgFree(nbmsg_t *this);

/* ------------------------------------------------------------------------- */

nbmsg_t	*msgNew(btsize reserved)
{
nbmsg_t	*volatile this;

	if(messageStore == NULL)
		messageStore = fifoNew();
	if((this = fifoGetObject(messageStore)) != NULL){
		msgClear(this, reserved);
	}else{
	EXC_TRY
		this = memCalloc(sizeof(nbmsg_t));
		this->reserved = reserved;
		this->size = reserved;
		this->readIndex = this->reserved;
		this->allocated = reserved + 256;
		this->data = memAlloc(this->allocated);
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "message: allocating new object 0x%08x\n", (int)this);
	EXC_CATCH
		msgFree(this);
		EXC_RETHROW();
	EXC_ENDTRY
	}
	this->refCount++;
	this->usageCount++;
	return this;
}

/* ------------------------------------------------------------------------- */

static void	msgFree(nbmsg_t *this)
{
	if(this != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "message: freeing object 0x%08x\n", (int)this);
		if(this->data != NULL)
			memFree(this->data);
		memFree(this);
	}
}

void	msgIncreaseSize(nbmsg_t *this, btsize size)
{
btsize	newSize;

	newSize = this->allocated;
	if(newSize < 16)
		newSize = 16;
	while(newSize < size)
		newSize *= 2;
	this->data = memRealloc(this->data, newSize);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "message: resizing from %d to %d\n", this->allocated, newSize);
	this->allocated = newSize;
}

void	msgRetain(void *obj)
{
nbmsg_t	*this = obj;

	this->refCount++;
}

void	msgRelease(void *obj)
{
nbmsg_t	*this = obj;

	this->refCount--;
	if(this->refCount == 0){
		if(this->usageCount > MAX_USAGE_COUNT){
			msgFree(this);
		}else{
			fifoAddObject(messageStore, this);
		}
	}else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

void	msgAddSpace(nbmsg_t *this, btsize bytes)
{
	msgEnsureSpace(this, bytes);
	memset(&this->data[this->size], 0, bytes);
	this->size += bytes;
}

void	msgAddByte(nbmsg_t *this, btu8 data)
{
	msgEnsureSpace(this, 1);
	this->data[this->size++] = data;
}

void	msgAddBlock(nbmsg_t *this, btu8 *data, btsize len)
{
	msgEnsureSpace(this, len);
	memcpy(&this->data[this->size], data, len);
	this->size += len;
}

void	msgSetByte(nbmsg_t *this, btsize index, btu8 data)
{
	if(index < 0)
		return;
	index += this->reserved;
	if(index > this->size)
		return;
	this->data[index] = data;
}

static void	memcpyBackwards(btu8 *dest, btu8 *src, btsize len)
{
	dest += len;
	src += len;
	while(len--){
		*--dest = *--src;
	}
}

void	msgInsertBlock(nbmsg_t *this, btu8 *data, btsize len)
{
btsize	distance, start;

	if(len > this->reserved){
		distance = len - this->reserved;
		msgEnsureSpace(this, distance);
		memcpyBackwards(&this->data[this->reserved] + distance, &this->data[this->reserved], this->size - this->reserved);
		this->size += distance;
		this->reserved += distance;
		if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
			logPrintf(LOGM_RARE, "message: moving data block for insert by %d bytes\n", distance);
	}
	start = this->reserved - len;
	memcpy(&this->data[start], data, len);
	this->reserved = start;
}


/* ------------------------------------------------------------------------- */

