/*
 * Name: pslsocket.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-05-21
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: pslsocket.c,v 1.6 2001/04/05 13:20:28 cs Exp $
 */

#include "stdheaders.h"
#include "pslsocket.h"
#include "exception.h"
#include "memalloc.h"
#include "evntloop.h"
#include "logging.h"
#include "system.h"
#include "process.h"
#include "message.h"
#include "blockpool.h"
#include "util.h"
#include "errcodes.h"
#include "server.h"

/* ------------------------------------------------------------------------- */

typedef struct pslCtx{
    PROC_VARSET;
    pslSocket_t	*psl;
	bttime		timeout;
}pslCtx_t;

static blockpool_t	*pslPool = NULL;

static void	freePslBlock(void *block)
{
    poolReturnBlock(pslPool, block);
}

/* ------------------------------------------------------------------------- */

static void	connectionLost(pslSocket_t *psl);

/* if we can't fulfill a send request within a given amount of time, there
 * must be something wrong and we'd better assume that the connection is
 * broken. The higher levels should try to reconnect anyway.
 */
static void	wantSendTimeout(void *obj, void *arg)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "pslsocket: send timed out, disconnecting\n");
	connectionLost(obj);
}

/* ------------------------------------------------------------------------- */

static void	pslSocketFree(pslSocket_t *psl)
{
	if(psl != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "pslsocket: freeing object at 0x%08x\n", (int)psl);
		elpCancelCallAt(wantSendTimeout, psl);
		if(psl->fd >= 0)
			elpCancelForFd(psl->fd, -1);
		if(psl->connectSignal != NULL){
			procCondFree(psl->connectSignal);
			psl->connectSignal = NULL;
		}
		memFree(psl);
	}
}

/* ------------------------------------------------------------------------- */

pslSocket_t	*pslSocketNew(btu32 destIp, int port, struct server *server)
{
pslSocket_t	*volatile psl = NULL;

    if(pslPool == NULL)
        pslPool = poolNew(sizeof(pslCtx_t), sizeof(pslCtx_t));
EXC_TRY
	psl = memCalloc(sizeof(pslSocket_t));
	psl->connectSignal = procCondNew();
	psl->cstatus = PSL_NOTCONNECTED;
	psl->destIp = destIp;
	psl->port = port;
	psl->fd = -1;
	psl->refCount = 1;
	psl->server = server;
EXC_CATCH
	pslSocketFree(psl);
	EXC_RETHROW();
EXC_ENDTRY
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "pslsocket: allocated new object at 0x%08x\n", (int)psl);
	return psl;
}

/* ------------------------------------------------------------------------- */

static inline void	pslSocketRetain(void *obj)
{
pslSocket_t *psl = obj;

	psl->refCount++;
}

static inline void	pslSocketRelease(void *obj)
{
pslSocket_t *psl = obj;

	psl->refCount--;
    if(psl->refCount == 0){
        pslSocketFree(psl);
    }else if(psl->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

static void	pslSocketSetHigherLayer(void *obj, void *nextObj, struct psDownInterface *nextI)
{
pslSocket_t	*psl = obj;

	psl->higherI = nextI;
	psl->higherObj = nextObj;
}

/* ------------------------------------------------------------------------- */

static void	connectionLost(pslSocket_t *psl)
{
	if(psl->fd >= 0)
		elpCancelForFd(psl->fd, -1);
	psl->wantSendCount = 0;
	close(psl->fd);
	psl->fd = -1;
    psl->cstatus = PSL_NOTCONNECTED;
	if(psl->higherI != NULL)
		psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
	procCondBroadcast(psl->connectSignal, NULL);
}

/* ------------------------------------------------------------------------- */

static void	sockReadable(void *obj, void *arg)
{
pslSocket_t	*psl = obj;
btu8		buffer[8192];
btsize		bytesRead;

	pslSocketRetain(psl);
	do{
		bytesRead = read(psl->fd, (char *)buffer, sizeof(buffer));
		if(bytesRead == 0){	/* socket closed by peer */
			if(LOG_MODE_IS_ACTIVE(LOGM_SMB_CONNECT))
				logPrintf(LOGM_SMB_CONNECT, "pslsocket: connection closed by server\n");
			connectionLost(psl);
		}else if(bytesRead < 0){
			if(errno == EWOULDBLOCK || errno == EAGAIN || errno == EINTR)
				break;
			if(LOG_MODE_IS_ACTIVE(LOGM_SMB_CONNECT)){
				if(errno == ECONNRESET)
					logPrintf(LOGM_SMB_CONNECT, "pslsocket: connection reset by server\n");
				else
					logPrintf(LOGM_SMB_CONNECT, "pslsocket: read error: [%d] %s\n", errno, sysStrerror(errno));
			}
			connectionLost(psl);
		}else{
			psl->bytesRead += bytesRead;
			if(psl->higherI != NULL)
				psl->higherI->rxMessage(psl->higherObj, buffer, bytesRead);
		}
	}while(bytesRead == sizeof(buffer));
	pslSocketRelease(psl);
}

/* ------------------------------------------------------------------------- */

static inline void	txReady(pslSocket_t *psl, btint errcode)
{
	msgRelease(psl->currentMsg);
	psl->currentMsg = NULL;
	if(psl->higherI != NULL)
		psl->higherI->txReady(psl->higherObj, errcode);
	if(psl->wantSendCount > 0){
		psl->wantSendCount--;
		if(psl->wantSendCount == 0 && psl->fd >= 0){
			elpCancelCallAt(wantSendTimeout, psl);
			elpCancelForFd(psl->fd, ELP_FDMODE_WRITE);
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	sockWritable(void *obj, void *arg)
{
pslSocket_t	*psl = obj;
btu8		*data;
btsize		len, written;

	pslSocketRetain(psl);
	/* retrigger timeout on each chunk sent: */
	elpCallAtCancelPrev(elpNow() + srvRequestTimeout(psl->server), wantSendTimeout, psl);
	do{
		if(psl->currentMsg == NULL){
			if(psl->higherI != NULL)
				psl->currentMsg = psl->higherI->getTxMessage(psl->higherObj);
			if(psl->currentMsg == NULL){
				psl->wantSendCount = 0;
				if(psl->fd >= 0){
					elpCancelCallAt(wantSendTimeout, psl);
					elpCancelForFd(psl->fd, ELP_FDMODE_WRITE);
				}
				break;
			}
			msgResetRead(psl->currentMsg);
		}
		msgPeekData(psl->currentMsg, &data, &len);
		written = write(psl->fd, data, len);
		if(written <= 0){
			if(errno != EINTR && errno != EAGAIN && errno != EWOULDBLOCK){
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "pslsocket: write error: [%d] %s\n", errno, sysStrerror(errno));
				txReady(psl, written == 0 ? EPIPE : errno);
				connectionLost(psl);
			}
		}else{
			psl->bytesWritten += written;
			if(written >= len){		/* ready? */
				txReady(psl, 0);
			}else{
				msgReadData(psl->currentMsg, &data, written);
			}
		}
	}while(psl->wantSendCount > 0 && written >= len);
	pslSocketRelease(psl);
}

/* ------------------------------------------------------------------------- */

static void	pslConnectProc(void *obj, void *arg)
{
pslCtx_t			*this = obj;
pslSocket_t			*psl = this->psl;
btError				res;
int					*rval = arg;
struct sockaddr_in	sockAddr;

	PROC_BEGIN;
	pslSocketRetain(psl);
    while(psl->cstatus == PSL_CONNECTING || psl->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(psl->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(psl->cstatus == PSL_CONNECTED){	/* if someone connected meanwhile */
		pslSocketRelease(psl);
		res.class = res.code = 0;
		PROC_RETURN(&res);
	}
    psl->cstatus = PSL_CONNECTING;
	if(psl->fd >= 0)
		elpCancelForFd(psl->fd, -1);
	psl->fd = -1;
	if((psl->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
		res.class = ERRCLS_UNIX;
		res.code = errno;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "pslsocket: error creating socket: [%d] %s\n", errno, sysStrerror(errno));
		goto errorOccured;
	}
	if(sysSetNonblock(psl->fd) < 0){
		res.class = ERRCLS_UNIX;
		res.code = errno;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "pslsocket: error set nonblock: [%d] %s\n", errno, sysStrerror(errno));
		goto errorOccured;
	}
	memset(&sockAddr, 0, sizeof(sockAddr));
	sockAddr.sin_family = AF_INET;
	sockAddr.sin_port = htons(psl->port);
	sockAddr.sin_addr.s_addr = htonl(psl->destIp);
	psl->bytesWritten = psl->bytesRead = 0;
	if(connect(psl->fd, (struct sockaddr *)&sockAddr, sizeof(sockAddr)) != 0){
		if(errno == EINPROGRESS){
			elpCallAtCancelPrev(elpNow() + this->timeout, pslConnectProc, this);	/* setup timeout */
			PROC_WAITFOR_X(2, elpRegisterForFd(psl->fd, ELP_FDMODE_WRITE, pslConnectProc, this));
			elpCancelCallAt(pslConnectProc, this);
			elpCancelForFd(psl->fd, ELP_FDMODE_WRITE);
			if(rval == NULL){	/* connect timeout */
				res.class = ERRCLS_CIFS_LOCAL;
				res.code = ERR_CIFS_LOCtimeout;
				goto errorOccured;
			}
#if 0
/* don't poll because man-page does not explicitly say that writing
 * 0 bytes is OK for polling. If a connect error occured, we'll get
 * it with our next read or write.
 */
			if(write(psl->fd, &res, 0) != 0){	/* poll for a connect error */
				res.class = ERRCLS_UNIX;
				res.code = errno;
				goto errorOccured;
			}
#endif
		}else{
			res.class = ERRCLS_UNIX;
			res.code = errno;
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "pslsocket: error connecting: [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
	}
	elpRegisterForFd(psl->fd, ELP_FDMODE_READ, sockReadable, psl);
    psl->cstatus = PSL_CONNECTED;
	psl->wantSendCount = 0;	/* there can't be pending writes just after connect */
	res.code = res.class = 0;
	goto doReturn;
errorOccured:
    psl->cstatus = PSL_NOTCONNECTED;
	if(psl->fd >= 0){
		elpCancelForFd(psl->fd, -1);
		close(psl->fd);
		psl->fd = -1;
	}
doReturn:
    PROC_RETURN_AND_CLEANUP(&res,
		if(psl->higherI != NULL)
			psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
		procCondBroadcast(psl->connectSignal, NULL);
		pslSocketRelease(psl);
	);
    PROC_END;
}

/* ------------------------------------------------------------------------- */

static void	pslDisconnectProc(void *obj, void *arg)
{
pslCtx_t	*this = obj;
pslSocket_t	*psl = this->psl;
btError		res;

	PROC_BEGIN;
	pslSocketRetain(psl);
    while(psl->cstatus == PSL_CONNECTING || psl->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(psl->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(psl->cstatus == PSL_NOTCONNECTED){	/* if someone disconnected meanwhile */
		pslSocketRelease(psl);
		res.class = res.code = 0;
		PROC_RETURN(&res);
	}
    psl->cstatus = PSL_DISCONNECTING;
	elpCancelForFd(psl->fd, -1);
	psl->wantSendCount = 0;
	close(psl->fd);
	psl->fd = -1;
    psl->cstatus = PSL_NOTCONNECTED;
	res.class = res.code = 0;
    PROC_RETURN_AND_CLEANUP(&res,
		if(psl->higherI != NULL)
			psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
		procCondBroadcast(psl->connectSignal, NULL);
		pslSocketRelease(psl);
	);
    PROC_END;
}

/* ------------------------------------------------------------------------- */

static void	pslSocketSetConnectStatus(void *obj, btint wantStatus, bttime timeout, void *caller)
{
pslSocket_t	*psl = obj;
pslCtx_t	*ctx;
btError		res;

	if(wantStatus == psl->cstatus){
		res.class = res.code = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}else{
		PROC_INIT(ctx, wantStatus == PSL_CONNECTED ? pslConnectProc : pslDisconnectProc, caller, pslPool, freePslBlock);
		ctx->psl = psl;
		ctx->timeout = timeout;
		PROC_START(ctx, NULL);
	}
}

/* ------------------------------------------------------------------------- */

static btint	pslSocketConnectStatus(void *obj)
{
pslSocket_t	*psl = obj;

	return psl->cstatus;
}

/* ------------------------------------------------------------------------- */

static void	pslSocketWantSend(void *obj)
{
pslSocket_t	*psl = obj;

	if(psl->cstatus == PSL_CONNECTED){
		if(psl->wantSendCount == 0){
			elpCallAtCancelPrev(elpNow() + srvRequestTimeout(psl->server), wantSendTimeout, psl);
			elpRegisterForFd(psl->fd, ELP_FDMODE_WRITE, sockWritable, psl);
		}
		psl->wantSendCount++;
	}
}

/* ------------------------------------------------------------------------- */

static void	pslSocketMaxInserts(void *obj, btsize *preInsert, btsize *postInsert)
{
	*preInsert = 0;
	*postInsert = 0;
}

/* ------------------------------------------------------------------------- */

psUpInterface_t	pslSocketUpInterface = {
	pslSocketSetHigherLayer,
	pslSocketRelease,
	pslSocketSetConnectStatus,
	pslSocketConnectStatus,
	pslSocketWantSend,
	pslSocketMaxInserts
};
