/*
 * Name: user.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-28
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: user.h,v 1.7 2001/05/06 15:21:24 cs Exp $
 */

#ifndef __user_H_INCLUDED__
#define __user_H_INCLUDED__

#include "basictypes.h"
#include "array.h"	/* for collectionObjectI_t */

/*
General Description:
This module implements the class describing a user (local _and_ remote).
It also implements the procedure of user login at the remote server.
User objects are stored at the server where the user is logged in.
*/

/* ------------------------------------------------------------------------- */

struct server;
struct share;

typedef struct cifsUser{
    btu32				unixUid;
    btu32				unixGid;
    struct server		*server;
    btuni				*smbUsername;
	btu8				*domain;
    bti32				refCount;
	btuni				*plainPasswd;
	btu8				pwHashPreNt[16];
	btu8				pwHashNt[16];
	struct hashtable	*sharePasswordCache;
    btu16				smbUid;				/* UID negotiated for this user */
	btu8				allowGuest;			/* whether guest connects are OK */
    btbool				isGuest;
	btbool				isInvalidated;		/* should be freed when possible */
    int					cstatus;			/* connection status */
    struct procCond		*connectSignal;		/* only one connect at a time */
}cifsUser_t;

typedef struct userRes{
	btError		err;
	cifsUser_t	*user;
	btuni		*passwd;
	struct ipcSrvClient	*client;
	btbool		userIsNew;
	btbool		allowUnencrypted;
}userRes_t;

extern collectionObjectI_t	userByUnixUidInterface;

cifsUser_t	*userNew(btu32 unixUid, btu32 unixGid, btuni *smbUsername, btuni *plainPasswd, btu8 *domain, btu8 passHashNt[16], btu8 passHashPreNt[16], btbool allowUnencrypted);
void	userFree(cifsUser_t *this);
void	userConnectionLost(cifsUser_t *this);
void	userConnect(cifsUser_t *this, void *caller);
void	userDisconnect(cifsUser_t *this, void *caller);
void	userAuthenticate(struct server *server, btu8 *domain, btint unixUid, btint unixGid, void *caller);
/* returns userRes_t with res->user containing a retain-neutral user object. Caller should not release unless the user object should be freed. */
btbool	userUnixUidsAreEqual(cifsUser_t *this, cifsUser_t *other);
btu32	userUnixUidHashval(cifsUser_t *this);
void	userRetain(cifsUser_t *this);
void	userRelease(cifsUser_t *this);
void	userInvalidate(cifsUser_t *user);

void	userGetSharePasswordForShare(cifsUser_t *user, struct share *share, void *caller);
/* This function finds the share password from a particular user. If there is
 * no share password stored in the user object, it asks the user.
 * Callback: userRes_t with err, passwd, client and allowUnencrypted set.
 *     'passwd' and 'client' are released/freed after the callback.
 */
void	userDeleteSharePasswordForShare(cifsUser_t *user, struct share *share);
/* This function deletes a stored share password from the user object. The
 * next time userGetSharePasswordForShare() is called, it will ask the user
 * instead of returning a stored password.
 */

/* ------------------------------------------------------------------------- */

static inline btu16	userSmbUid(cifsUser_t *this)
{
    return this->smbUid;
}

static inline btu32	userUnixUid(cifsUser_t *this)
{
    return this->unixUid;
}

static inline btu32	userUnixGid(cifsUser_t *this)
{
	return this->unixGid;
}

static inline int	userConnectionStatus(cifsUser_t *this)
{
    return this->cstatus;
}

static inline btuni	*userCifsName(cifsUser_t *this)
{
	return this->smbUsername;
}

static inline btu8	*userDomain(cifsUser_t *this)
{
	return this->domain;
}

/* ------------------------------------------------------------------------- */

#endif	/* __user_H_INCLUDED__ */
