/*
 * Name: dummyfsmod.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-04
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: dummyfsmod.c,v 1.5 2001/01/18 19:29:49 cs Exp $
 */

/* General Description:
This module implements the minimum backend module for demonstration purposes.
It displays some files of size 0 which can not be read or written.
*/

#include "stdheaders.h"
#include "module.h"
#include "backend.h"
#include "hashtable.h"
#include "notification.h"
#include "unicode.h"
#include "main.h"
#include "errcodes.h"
#include "memalloc.h"
#include "process.h"
#include "logging.h"
#include "mainconst.h"

#define	NUM_CHILDREN	5

struct cfgProperty;

/* ------------------------------------------------------------------------- */

typedef struct dummyFile{
	struct dummyFile	*parent;
	btuni				*name;
	struct dummyFile	*children[NUM_CHILDREN];
}dummyFile_t;

typedef struct dummyShare{
	btuni					*path;
	btint					refCount;
	dummyFile_t				*root;
	notificationCenter_t	*openNotify;
	notificationCenter_t	*invalidateNotify;
}dummyShare_t;

/* ------------------------------------------------------------------------- */

static dummyFile_t	*createTree(btint depth, dummyFile_t *parent)
{
dummyFile_t		*root;
btu8			name[256];
btuni			uniName[256];
static btint	filecnt = 0;
btint			i;

	root = memCalloc(sizeof(dummyFile_t));
	root->parent = parent == NULL ? root : parent;
	sprintf(name, "file-%d", filecnt++);
	convLocalAsciiToUni(name, uniName, 256);
	root->name = uniStrDup(uniName);
	if(depth != 0){
		for(i=0;i<NUM_CHILDREN;i++){
			root->children[i] = createTree(depth - 1, root);
		}
	}
	return root;
}

static void	freeTree(dummyFile_t *tree)
{
btint			i;
	if(tree != NULL){
		for(i=0;i<NUM_CHILDREN;i++){
			freeTree(tree->children[i]);
		}
		memFree(tree->name);
		memFree(tree);
	}
}

/* ------------------------------------------------------------------------- */
/* ---------------------------- vfs functions: ----------------------------- */
/* ------------------------------------------------------------------------- */

static btint	shareCompare(void *obj, void *otherObj)
{
dummyShare_t	*this = obj, *other = otherObj;

	return uniStrCmp(this->path, other->path, 0);
}

static btu32	shareHashVal(void *obj)
{
dummyShare_t	*this = obj;

	return uniStrHash(this->path, 0);
}

static collectionObjectI_t	shareInterface = {
	shareCompare,
	shareHashVal,
	NULL,
	NULL
};

static hashtable_t	*allShares = NULL;

static void	shareRetain(dummyShare_t *share)
{
	share->refCount++;
}

static void	shareRelease(dummyShare_t *share)
{
	if(--share->refCount == 0){
		htabRemoveObjectEqualTo(allShares, share);
		freeTree(share->root);
		memFree(share->path);
		ncRelease(share->openNotify);
		ncRelease(share->invalidateNotify);
		memFree(share);
	}else if(share->refCount < 0){
		DEBUG_TRAP();
	}
}

static dummyShare_t	*shareLookup(btuni *path)
{
dummyShare_t		*share;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareLookup(path=%s)\n", uniToCstring(path));
	if(allShares == NULL)
		allShares = htabNew(&shareInterface, 8, 75);
	share = htabFindObjectEqualTo(allShares, &path);
	if(share != NULL){
		shareRetain(share);
		return share;
	}
	share = memCalloc(sizeof(dummyShare_t));
	share->root = createTree(3, NULL);
	share->openNotify = ncNew(2);
	share->invalidateNotify = ncNew(2);
	share->refCount = 1;
	share->path = uniStrDup(path);
	htabAddObject(allShares, share);
	return share;
}

static void shareConnect(dummyShare_t *share, void *user, void *caller)
{
beVfsRes_t	res;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareConnect(share=%s)\n", uniToCstring(share->path));
	res.err.class = res.err.code = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static void shareDisconnect(dummyShare_t *share, void *user, void *caller)
{
beVfsRes_t	res;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareDisconnect(share=%s)\n", uniToCstring(share->path));
	res.err.class = res.err.code = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static dummyFile_t	*shareRootFile(dummyShare_t *share)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
		logPrintf(LOGM_DEBUG, "dummyFs: shareRootFile(share=%s)\n", uniToCstring(share->path));
		logPrintf(LOGM_DEBUG, "dummyFs: returning file=%s\n", uniToCstring(share->root->name));
	}
	return share->root;
}

static void	shareStatfs(dummyShare_t *share, void *user, void *caller)
{
beVfsRes_t	res;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareStatfs(share=%s)\n", uniToCstring(share->path));
	res.err.class = res.err.code = 0;
	res.bytesPerAU = 32768;
	res.totalAU = 10;
	res.availableAU = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static btint	shareConfigPath(dummyShare_t *share, struct cfgProperty **path, btint maxEntries)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareConfigPath(share=%s)\n", uniToCstring(share->path));
	return 0;
}

static btuni	*shareVfsPath(dummyShare_t *share)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareVfsPath(share=%s)\n", uniToCstring(share->path));
	return share->path;
}

static btbool	shareIsCaseSensitive(dummyShare_t *share)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareIsCaseSensitive(share=%s)\n", uniToCstring(share->path));
	return 1;
}

static notificationCenter_t	*shareOpenNotification(dummyShare_t *share)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareOpenNotification(share=%s)\n", uniToCstring(share->path));
	return share->openNotify;
}

static notificationCenter_t	*shareInvalidateNotification(dummyShare_t *share)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: shareInvalidateNotification(share=%s)\n", uniToCstring(share->path));
	return share->invalidateNotify;
}

static void	userAuthenticate(dummyShare_t *share, btint uid, btint gid, void *caller)
{
beVfsRes_t	res;

	res.err.class = res.err.code = 0;
	res.user = (void *)uid;
	/* we don't need any type of complex authentication token */
	PROC_IMMEDIATE_RET(caller, &res);
}

static void	userRetain(void *share)
{
	/* dummy */
}

static void	userRelease(void *share)
{
	/* dummy */
}

/* ------------------------------------------------------------------------- */
/* ---------------------------- file functions: ---------------------------- */
/* ------------------------------------------------------------------------- */

static void	fileLookup(dummyFile_t *parent, btuni *name, void *user, void *caller)
{
btint		i;
beFileRes_t	res;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
		logPrintf(LOGM_DEBUG, "dummyFs: fileLookup(parent=%s)\n", uniToCstring(parent->name));
		logPrintf(LOGM_DEBUG, "dummyFs: lookup name=%s\n", uniToCstring(name));
	}
	if(parent->children[0] == NULL){
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadaccess;
	}else{
		for(i=0;i<NUM_CHILDREN;i++){
			if(uniStrCmp(parent->children[i]->name, name, 1) == 0){
				break;
			}
		}
		if(i < NUM_CHILDREN){
			res.err.class = res.err.code = 0;
			res.file = parent->children[i];
		}else{
			res.err.class = ERRCLS_CIFS_DOS;
			res.err.code = ERR_CIFS_DOSbadpath;
		}
	}
	PROC_IMMEDIATE_RET(caller, &res);
}

static void	fileRetain(dummyFile_t *file)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: fileRetain(%s)\n", uniToCstring(file->name));
	/* no ref-counting! */
}

static void	fileRelease(dummyFile_t *file)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: fileRelease(%s)\n", uniToCstring(file->name));
	/* no ref-counting! */
}

static void	fileGetAttributes(dummyFile_t *file, void *user, void *caller)
{
beFileRes_t		res;
beFileAttr_t	attr;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: fileGetAttributes(%s)\n", uniToCstring(file->name));
	memset(&attr, 0, sizeof(attr));
	res.attr = &attr;
	res.err.class = res.err.code = 0;
	attr.type = file->children[0] == NULL ? BE_FILETYPE_REG : BE_FILETYPE_DIR;
	attr.unixMode = file->children[0] == NULL ? 0666 : 0777;
	attr.numLinks = 0;	/* 0 if not known */
	attr.uid = (btint)user;
	attr.gid = 0;
	attr.size = 0;
	attr.allocatedSize = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static void	fileReaddir(dummyFile_t *file, void *user, btint startIndex, btbool (*resultCallback)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir), void *caller)
{
btint		i;
beFileRes_t	res;

	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: fileReaddir(%s)\n", uniToCstring(file->name));
	if(file->children[0] == NULL){
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadaccess;
	}else{
		for(i=startIndex;i<NUM_CHILDREN;i++){
			if(!resultCallback(caller, file->children[i]->name, i+1, file->children[i]->children[0] != NULL))
				break;
		}
		res.eof = i >= NUM_CHILDREN;
		res.err.class = res.err.code = 0;
	}
	PROC_IMMEDIATE_RET(caller, &res);
}

static btuni	*fileName(dummyFile_t *file)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: fileName(%s)\n", uniToCstring(file->name));
	return file->name;
}

static dummyFile_t	*fileParent(dummyFile_t *file)
{
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "dummyFs: fileParent(%s)\n", uniToCstring(file->name));
	return file->parent;
}

/* ------------------------------------------------------------------------- */
/* --------------------------- module functions: --------------------------- */
/* ------------------------------------------------------------------------- */

static struct{
	btu8	*name;
	void	*function;
} functionTable[] = {
	{BE_CONFIG_PATH,			shareConfigPath},
	{BE_VFS_FOR_PATH,			shareLookup},
	{BE_VFS_RETAIN,				shareRetain},
	{BE_VFS_RELEASE,			shareRelease},
	{BE_VFS_CONNECT,			shareConnect},
	{BE_VFS_DISCONNECT,			shareDisconnect},
	{BE_VFS_ROOT_FILE,			shareRootFile},
	{BE_VFS_STATFS,				shareStatfs},
	{BE_VFS_PATH,				shareVfsPath},
	{BE_VFS_CASE_SENSITIVE,		shareIsCaseSensitive},
	{BE_VFS_OPEN_NOTIFICATION,	shareOpenNotification},
	{BE_VFS_INVALIDATE_NOTIFICATION,	shareInvalidateNotification},
	{BE_USER_AUTHENTICATE,		userAuthenticate},
	{BE_USER_RETAIN,			userRetain},
	{BE_USER_RELEASE,			userRelease},
	{BE_FILE_LOOKUP,			fileLookup},
	{BE_FILE_RETAIN,			fileRetain},
	{BE_FILE_RELEASE,			fileRelease},
	{BE_FILE_GETATTR,			fileGetAttributes},
	{BE_FILE_READDIR,			fileReaddir},
	{BE_FILE_NAME,				fileName},
	{BE_FILE_PARENT,			fileParent},
};

static void	*getImp(btu8 *functionName)
{
btint	i;

	for(i=0;i<sizeof(functionTable)/sizeof(functionTable[0]);i++){
		if(strcmp(functionName, functionTable[i].name) == 0)
			return functionTable[i].function;
	}
	return NULL;
}

void	dummyFs_init(void)
{
module_t	*mod;

	mod = memCalloc(sizeof(*mod));
	mod->name = "dummyFs";
	mod->moduleType = MOD_TYPE_BACKEND;
	mod->implementationOf = getImp;
	mod->interfaceVersion = 0;
	mod->implementationVersion = MAINCONST_VERSION;
	mod->betaVersion = MAINCONST_BETA_VERSION;
	modRegister(mod);
}

/* ------------------------------------------------------------------------- */

