/*
 * Name: array.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-24
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: array.c,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */


#include "stdheaders.h"
#include "exception.h"
#include "memalloc.h"
#include "array.h"
#include "logging.h"

/* ------------------------------------------------------------------------- */

static btint	dummyCompare(void *o1, void *o2)
{
btu8	*this = o1, *other = o2;

	return this - other;
}

static btu32	dummyHashVal(void *obj)
{
btint	x = (btint)obj;

	return x ^ (x >> 6) ^ (x >> 12);
}

static collectionObjectI_t	dummyInterface = {
	dummyCompare,
	dummyHashVal,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

array_t	*arrNew(const collectionObjectI_t *contentObjectInterface, btu32 volatile initialCapacity)
{
array_t	*volatile this = NULL;

EXC_TRY
	if(initialCapacity <= 0)
		initialCapacity = 1;
    this = memCalloc(sizeof(array_t));
    this->tableSize = initialCapacity;
    this->table = memAlloc(this->tableSize * sizeof(void *));
	if(contentObjectInterface != NULL)
    	this->objectInterface = *contentObjectInterface;
	else
    	this->objectInterface = dummyInterface;
EXC_CATCH
	arrFree(this);
	EXC_RETHROW();
EXC_ENDTRY
    return this;
}

/* ------------------------------------------------------------------------- */

void	arrFree(array_t *this)
{
int		i;
    
	if(this != NULL){
		if(this->table != NULL){
			if(this->objectInterface.notifyRemovedFrom != NULL){
				for(i=0;i<this->objectCount;i++){
					this->objectInterface.notifyRemovedFrom(this->table[i]);
				}
			}
			memFree(this->table);
		}
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

void	arrAddObject(array_t *this, void *obj)
{
btu32	newSize;
    
    if(this->objectCount >= this->tableSize){
        newSize = this->tableSize * 2;
        this->table = memRealloc(this->table, newSize * sizeof(void *));
        this->tableSize = newSize;
        if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE)){
            logPrintf(LOGM_RESOURCE, "array: increasing to size %d\n", newSize);
        }
    }
    this->table[this->objectCount++] = obj;
    if(this->objectInterface.notifyAddedTo != NULL)
        this->objectInterface.notifyAddedTo(obj);
}

/* ------------------------------------------------------------------------- */

void	arrRemoveObjectAtIndex(array_t *this, btsize index)
{
btsize	i;

	if(index < 0 || index >= this->objectCount)
		return;
	if(this->objectInterface.notifyRemovedFrom != NULL){
		this->objectInterface.notifyRemovedFrom(this->table[index]);
	}
	this->objectCount--;
	for(i=index;i<this->objectCount;i++)
		this->table[i] = this->table[i+1];
}

/* ------------------------------------------------------------------------- */

void	arrRemoveAllObjects(array_t *this)
{
btu32	i;

	if(this->objectInterface.notifyRemovedFrom != NULL){
		for(i=0;i<this->objectCount;i++)
			this->objectInterface.notifyRemovedFrom(this->table[i]);
	}
	this->objectCount = 0;
}

/* ------------------------------------------------------------------------- */

btbool	arrRemoveObjectEqualTo(array_t *this, void *obj)
{
btsize	i;
btbool	anyRemoved = 0;

	for(i=0;i<this->objectCount;){
		if(this->objectInterface.compare(this->table[i], obj) == 0){
			arrRemoveObjectAtIndex(this, i);
			anyRemoved = 1;
		}else
			i++;
	}
	return anyRemoved;
}

/* ------------------------------------------------------------------------- */

void	*arrFindObjectEqualTo(array_t *this, void *obj)
{
btsize	i;

	for(i=0;i<this->objectCount;i++){
		if(this->objectInterface.compare(this->table[i], obj) == 0)
			return this->table[i];
	}
	return NULL;
}

/* ------------------------------------------------------------------------- */

btsize	arrIndexOfObjectEqualTo(array_t *this, void *obj)
{
btsize	i;

	for(i=0;i<this->objectCount;i++){
		if(this->objectInterface.compare(this->table[i], obj) == 0)
			return i;
	}
	return -1;
}

/* ------------------------------------------------------------------------- */

static btint	(*sortCompareFunction)(void *obj, void *otherObj);

static btint sortCompare(const void *o1, const void *o2)
{
void	**obj1 = (void **)o1, **obj2 = (void **)o2;

	return sortCompareFunction(*obj1, *obj2);
}

void	arrSort(array_t *arr)
{
	sortCompareFunction = arr->objectInterface.compare;
	qsort(arr->table, arr->objectCount, sizeof(void *), sortCompare);
}

/* ------------------------------------------------------------------------- */

