/*
 * Name: cfgparse.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-10
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cfgparse.c,v 1.2 2001/01/08 15:53:49 cs Exp $
 */


#include "stdheaders.h"
#include "logging.h"
#include "memalloc.h"
#include "array.h"
#include "hashtable.h"
#include "scanner.h"
#include "cfgparse.h"
#include "util.h"
#include "system.h"
#include "notification.h"

/* ------------------------------------------------------------------------- */

EXC_DEFINE(excParseError);

btu8	*cfgYesNoEnum[3] = { "no", "yes", NULL };

static notificationCenter_t	*changeNotification = NULL;

btu32	cfgGlobalGenerationValue = 1;

/* ------------------------------------------------------------------------- */

static void	cfgInvalidateCache(void)
{
	cfgGlobalGenerationValue++;
	if(changeNotification != NULL){
		ncPostNotification(changeNotification, NULL);
	}
}

/* ------------------------------------------------------------------------- */

struct notificationCenter	*cfgChangedNotification(void)
{
	if(changeNotification == NULL)
		changeNotification = ncNew(8);
	return changeNotification;
}

/* ------------------------------------------------------------------------- */

static void	cfgPropertyFree(void *obj)
{
cfgProperty_t	*this = obj;

	if(this != NULL){
		if(this->name != NULL)
			memFree(this->name);
		switch(this->contentType){
		case CFG_PROPTYPE_STRING:
			if(this->content.string != NULL)
				memFree(this->content.string);
			break;
		case CFG_PROPTYPE_LIST:
			if(this->content.list != NULL)
				arrFree(this->content.list);
			break;
		case CFG_PROPTYPE_DICT:
			if(this->content.dict != NULL)
				htabFree(this->content.dict);
			break;
		}
		memFree(this);
	}
}

static btu32	cfgPropertyHash(void *obj)
{
cfgProperty_t	*this = obj;

	if(this->name != NULL)
		return utilStrHash(this->name, 0);
	else
		return 0;
}

static btint	cfgPropertyCompare(void *obj, void *otherObj)
{
cfgProperty_t *this = obj, *other = otherObj;

	if(this->name != NULL && other->name != NULL){
		return utilStrCmp(this->name, other->name, 0);
	}
	return 1;
}

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	propertyCollection = {
	cfgPropertyCompare,
	cfgPropertyHash,
	NULL,
	cfgPropertyFree
};

static collectionObjectI_t	dummyInterface = {
	NULL,
	NULL,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static btbool	cfgParseName(scanner_t *stream, btu8 **name)
{
int		token;

	token = scannerGetToken(stream, name);
	switch(token){
	case SCAN_TOKEN_ALNUM:
	case SCAN_TOKEN_STRING:
	case SCAN_TOKEN_CHAR:
		return 1;
	default:
		scannerUngetToken(stream, token);
		return 0;
	}
}

static btbool	cfgParseOperator(scanner_t *stream, btu8 op)
{
int		token;
btu8	*value;

	if((token = scannerGetToken(stream, &value)) == SCAN_TOKEN_OPERATOR){
		if(value[0] == op && value[1] == 0)
			return 1;
	}
	scannerUngetToken(stream, token);
	return 0;
}

static btbool	cfgParseOptionalOperator(scanner_t *stream, btu8 op)
{
int		token;
btu8	*value;

	if((token = scannerGetToken(stream, &value)) == SCAN_TOKEN_OPERATOR){
		if(value[0] == op && value[1] == 0)
			return 1;
	}
	scannerUngetToken(stream, token);
	return 1;
}

static btbool	cfgParseProperty(scanner_t *stream, btu8 *file, void **contentp, int *type);

static btbool	cfgParseDictEntry(scanner_t *stream, btu8 *file, cfgProperty_t **entry)
{
btu8			*name;
cfgProperty_t	*volatile p = NULL;

	if(!cfgParseName(stream, &name)){
		return 0;
	}
EXC_TRY
	p = memCalloc(sizeof(cfgProperty_t));
	p->line = scannerGetLineNumber(stream);
	p->file = file;
	p->name = utilStrDup(name);
	if(!cfgParseOperator(stream, '=')){
		EXC_THROW_MSG(excParseError, "\'=\' expected");
	}
	if(!cfgParseProperty(stream, file, (void **)&p->content, &p->contentType)){
		EXC_THROW_MSG(excParseError, "property value expected");
	}
	cfgParseOptionalOperator(stream, ';');
EXC_CATCH
	if(p != NULL){
		if(p->name != NULL)
			memFree(p->name);
		memFree(p);
	}
	EXC_RETHROW();
EXC_ENDTRY
	*entry = p;
	return 1;
}

static btbool	cfgParseListEntry(scanner_t *stream, btu8 *file, cfgProperty_t **entry)
{
cfgProperty_t	*p = NULL;
void			*content;
int				type;

	if(!cfgParseProperty(stream, file, &content, &type)){
		return 0;
	}
	p = memCalloc(sizeof(cfgProperty_t));
	p->line = scannerGetLineNumber(stream);
	p->file = file;
	p->content.list = content;	/* member of the union does not matter */
	p->contentType = type;
	cfgParseOptionalOperator(stream, ',');
	*entry = p;
	return 1;
}

static btbool	cfgParseProperty(scanner_t *stream, btu8 *file, void **contentp, int *type)
{
array_t			*arr;
hashtable_t		*dict;
btu8			*string;
cfgProperty_t	*entry;

	if(cfgParseOperator(stream, '(')){
		*type = CFG_PROPTYPE_LIST;
		arr = arrNew(&propertyCollection, 4);
		while(cfgParseListEntry(stream, file, &entry)){
			arrAddObject(arr, entry);
		}
		if(!cfgParseOperator(stream, ')')){
			arrFree(arr);
			EXC_THROW_MSG(excParseError, "\')\' expected");
		}
		*contentp = arr;
	}else if(cfgParseOperator(stream, '{')){
		*type = CFG_PROPTYPE_DICT;
		dict = htabNew(&propertyCollection, 16, 75);
		while(cfgParseDictEntry(stream, file, &entry)){
			if(htabFindObjectEqualTo(dict, entry) != NULL && LOG_MODE_IS_ACTIVE(LOGM_WARN)){
				logPrintf(LOGM_WARN, "cfgparse: duplicate defined property in line %d\n", scannerGetLineNumber(stream));
			}
			htabAddObject(dict, entry);
		}
		if(!cfgParseOperator(stream, '}')){
			htabFree(dict);
			EXC_THROW_MSG(excParseError, "\'}\' expected");
		}
		*contentp = dict;
	}else if(cfgParseName(stream, &string)){
		*type = CFG_PROPTYPE_STRING;
		*contentp = utilStrDup(string);
	}else{
		return 0;
	}
	return 1;
}

static int	cfgGetChar(FILE *fp)	/* need a function pointer, getc is a macro */
{
	return getc(fp);
}

static hashtable_t	*cfgParseFile(btu8 *volatile fileName)
{
hashtable_t		*volatile dict = NULL;
btu8			*value;
FILE			*volatile fp = NULL;
scanner_t		*volatile stream = NULL;
cfgProperty_t	*entry;

	fp = fopen(fileName, "r");
	if(fp == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "error opening config file %s: [%d] %s\n", fileName, errno, sysStrerror(errno));
		return NULL;
	}
EXC_TRY
	stream = scannerNew(cfgGetChar, fp);
	scannerSetAllowRem(stream, 1, 1, 1);	/* allow all kinds of remarks */
	cfgParseOptionalOperator(stream, '{');
	dict = htabNew(&propertyCollection, 16, 75);
	while(cfgParseDictEntry(stream, fileName, &entry)){
		if(htabFindObjectEqualTo(dict, entry) != NULL && LOG_MODE_IS_ACTIVE(LOGM_WARN)){
			logPrintf(LOGM_WARN, "cfgparse: duplicate defined section in line %d\n", scannerGetLineNumber(stream));
		}
		htabAddObject(dict, entry);
	}
	cfgParseOptionalOperator(stream, '}');
	if(scannerGetToken(stream, &value) != SCAN_TOKEN_EOF && LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		EXC_THROW_MSG(excParseError, "EOF expected");
	}
EXC_CATCH
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		if(EXC_NAME == excParseError)
			logPrintf(LOGM_ERROR, "parse error: %s in line %d of file %s\n", EXC_MSG, scannerGetLineNumber(stream), fileName);
		else
			logPrintf(LOGM_ERROR, "exception %s (%s) during parsing\n", EXC_NAME, EXC_MSG);
	}
	htabFree(dict);
	dict = NULL;
EXC_ENDTRY
	if(stream != NULL)
		scannerFree(stream);
	if(fp != NULL)
		fclose(fp);
	return dict;
}

/* ------------------------------------------------------------------------- */

btbool		cfgFileChanged(cfgFile_t *this)
{
struct stat	sbuf;

	if(stat(this->fileName, &sbuf) != 0){
		memset(&sbuf, 0, sizeof(sbuf));
	}
	return this->lastModify != (bti64)sbuf.st_mtime;
}

btbool		cfgFileParseIfChanged(cfgFile_t *this)
{
hashtable_t	*dict = NULL;
struct stat	sbuf;

	if(stat(this->fileName, &sbuf) != 0){
		memset(&sbuf, 0, sizeof(sbuf));
	}
	if(this->lastModify != (bti64)sbuf.st_mtime){
		dict = cfgParseFile(this->fileName);
		if(dict != NULL){	/* parsing was successful */
			cfgInvalidateCache();
			htabFree(this->plist->content.dict);
			this->plist->content.dict = dict;
		}else if(this->plist->content.dict != NULL && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "continuing with old configuration\n");
		this->lastModify = sbuf.st_mtime;
	}
	return dict != NULL;
}

cfgFile_t	*cfgFileNew(btu8 *fileName)
{
cfgFile_t	*this = NULL;

	this = memCalloc(sizeof(cfgFile_t));
	this->fileName = utilStrDup(fileName);
	this->lastModify = -1;
	this->plist = memCalloc(sizeof(plist_t));
	this->plist->contentType = CFG_PROPTYPE_DICT;
	this->plist->content.dict = htabNew(&propertyCollection, 16, 75);
	cfgFileParseIfChanged(this);
	return this;
}

void		cfgFileFree(cfgFile_t *this)
{
	if(this != NULL){
		if(this->fileName != NULL)
			memFree(this->fileName);
		if(this->plist != NULL){
			if(this->plist->content.dict != NULL)
				htabFree(this->plist->content.dict);
			memFree(this->plist);
		}
		memFree(this);
	}
}

plist_t		*cfgFilePlist(cfgFile_t *this)
{
/* don't call cfgFileParseIfChanged() here because it has to do a filesystem
 * operation. It's called periodically from the main loop anyway, if this
 * behaviour is required.
 */
	if(this->plist->content.dict != NULL)
		return this->plist;
	else
		return NULL;
}

/* ------------------------------------------------------------------------- */

cfgProperty_t	*cfgQueryWithName(btu8 *name, btint numLists, ...)
{
cfgProperty_t	*p = NULL, *list;
va_list			vlist;
btint			i;

	va_start(vlist, numLists);
	for(i=0;i<numLists;i++){
		list = va_arg(vlist, cfgProperty_t *);
		if(list == NULL)
			continue;	/* ignore */
		if(list->contentType != CFG_PROPTYPE_DICT){
			if(LOG_MODE_IS_ACTIVE(LOGM_WARN)){
				if(list->name != NULL)
					logPrintf(LOGM_WARN, "cfgparse: dictionary expected for property %s at line %d of file %s\n", list->name, list->line, list->file);
				else
					logPrintf(LOGM_WARN, "cfgparse: dictionary expected at line %d of file %s\n", list->line, list->file);
			}
		}else{
			p = htabFindObjectEqualTo(list->content.dict, &name);
			if(p != NULL)
				break;
		}
	}
	va_end(vlist);
	return p;
}

cfgProperty_t	*cfgQueryFromArray(btu8 *name, btint arrSize, plist_t **arr)
{
cfgProperty_t	*p = NULL, *list;
btint			i;

	for(i=0;i<arrSize;i++){
		list = arr[i];
		if(list == NULL)
			continue;	/* ignore */
		if(list->contentType != CFG_PROPTYPE_DICT){
			if(LOG_MODE_IS_ACTIVE(LOGM_WARN)){
				if(list->name != NULL)
					logPrintf(LOGM_WARN, "cfgparse: dictionary expected for property %s at line %d of file %s\n", list->name, list->line, list->file);
				else
					logPrintf(LOGM_WARN, "cfgparse: dictionary expected at line %d of file %s\n", list->line, list->file);
			}
		}else{
			p = htabFindObjectEqualTo(list->content.dict, &name);
			if(p != NULL)
				break;
		}
	}
	return p;
}


array_t	*cfgQueryList(plist_t *list)
{
array_t			*all, *rval;
btint			i, cnt;
cfgProperty_t	*p;

	rval = arrNew(&dummyInterface, 8);
	if(list != NULL && list->contentType == CFG_PROPTYPE_DICT){
		all = htabAllObjects(list->content.dict);
		arrReplaceInterface(all, &dummyInterface);	/* avoid freeing */
		cnt = arrObjectCount(all);
		for(i=0;i<cnt;i++){
			p = arrObjectAtIndex(all, i);
			arrAddObject(rval, p->name);
		}
		arrFree(all);	/* does not free referenced content */
	}
	return rval;
	
}

cfgProperty_t	*cfgQueryAtIndex(btsize index, plist_t *plist)
{
	if(plist == NULL)
		return NULL;
	if(plist->contentType == CFG_PROPTYPE_LIST){
		return arrObjectAtIndex(plist->content.list, index);
	}else if(index == 0 && plist->contentType == CFG_PROPTYPE_STRING){
		return plist;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_WARN))
		logPrintf(LOGM_WARN, "cfgparse: list expected at line %d of file %s\n", plist->line, plist->file);
	return NULL;
}

btsize			cfgQueryValueCount(plist_t *plist)
{
	if(plist == NULL)
		return 0;
	if(plist->contentType == CFG_PROPTYPE_STRING)
		return 1;
	if(plist->contentType == CFG_PROPTYPE_LIST)
		return arrObjectCount(plist->content.list);
	return 0;
}

/* ------------------------------------------------------------------------- */

btu32			cfgIntValue(cfgProperty_t *property, btu32 defaultValue)
{
char	*value;

	if(property == NULL)
		return defaultValue;
	if(property->contentType != CFG_PROPTYPE_STRING){
		if(LOG_MODE_IS_ACTIVE(LOGM_WARN)){
			if(property->name != NULL)
				logPrintf(LOGM_WARN, "cfgparse: int value expected for property %s at line %d of file %s\n", property->name, property->line, property->file);
			else
				logPrintf(LOGM_WARN, "cfgparse: int value expected for list element at line %d of file %s\n", property->line, property->file);
		}
		return defaultValue;
	}
	if(!isdigit(property->content.string[0]) && property->content.string[0] != '-'){
		if(LOG_MODE_IS_ACTIVE(LOGM_WARN))
			logPrintf(LOGM_WARN, "cfgparse: %s is not a number at line %d of file %s\n", property->content.string, property->line, property->file);
		return defaultValue;
	}
	value = property->content.string;
	return strtol(value, &value, 0);
}

btu8			*cfgStringValue(cfgProperty_t *property, btu8 *defaultValue)
{
	if(property == NULL)
		return defaultValue;
	if(property->contentType != CFG_PROPTYPE_STRING){
		if(LOG_MODE_IS_ACTIVE(LOGM_WARN)){
			if(property->name != NULL)
				logPrintf(LOGM_WARN, "cfgparse: value expected for property %s at line %d of file %s\n", property->name, property->line, property->file);
			else
				logPrintf(LOGM_WARN, "cfgparse: value expected for list element at line %d of file %s\n", property->line, property->file);
		}
		return defaultValue;
	}
	return property->content.string;
}

int				cfgEnumValue(cfgProperty_t *property, btu8 **enumNames, int defaultValue)
{
int		i;

	if(property == NULL)
		return defaultValue;
	if(property->contentType != CFG_PROPTYPE_STRING){
		if(LOG_MODE_IS_ACTIVE(LOGM_WARN)){
			if(property->name != NULL)
				logPrintf(LOGM_WARN, "cfgparse: enum value expected for property %s at line %d of file %s\n", property->name, property->line, property->file);
			else
				logPrintf(LOGM_WARN, "cfgparse: enum value expected for list element at line %d of file %s\n", property->line, property->file);
		}
		return defaultValue;
	}
	for(i=0; enumNames[i] != NULL; i++){
		if(utilStrCmp(property->content.string, enumNames[i], 0) == 0)
			return i;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_WARN))
		logPrintf(LOGM_WARN, "cfgparse: enum value %s not known at line %d of file %s\n", property->content.string, property->line, property->file);
	return defaultValue;
}


/* ------------------------------------------------------------------------- */

cfgProperty_t	*cfgBuildIntProperty(btu32 value)
{
cfgProperty_t	*p;
btu8			sval[32];

	p = memCalloc(sizeof(cfgProperty_t));
	/* keep the name unassigned, will be set when added to list */
	p->contentType = CFG_PROPTYPE_STRING;
	sprintf(sval, "%d", value);
	p->content.string = utilStrDup(sval);
	p->line = 0;
	p->file = "<internally created>";
	return p;
}

cfgProperty_t	*cfgBuildStringProperty(btu8 *value)
{
cfgProperty_t	*p;

	p = memCalloc(sizeof(cfgProperty_t));
	/* keep the name unassigned, will be set when added to list */
	p->contentType = CFG_PROPTYPE_STRING;
	p->content.string = utilStrDup(value);
	p->line = 0;
	p->file = "<internally created>";
	return p;
}

cfgProperty_t	*cfgBuildEnumProperty(btint value, btu8 **enumNames)
{
cfgProperty_t	*p;
btint			i;

	p = memCalloc(sizeof(cfgProperty_t));
	/* keep the name unassigned, will be set when added to list */
	p->contentType = CFG_PROPTYPE_STRING;
	for(i=0;enumNames[i]!=NULL;i++);	/* count names */
	if(value < 0 || value >= i){
		p->content.string = utilStrDup("<invalid>");
	}else{
		p->content.string = utilStrDup(enumNames[value]);
	}
	p->line = 0;
	p->file = "<internally created>";
	return p;
}

cfgProperty_t	*cfgBuildArrayProperty(void)
{
cfgProperty_t	*p;

	p = memCalloc(sizeof(cfgProperty_t));
	/* keep the name unassigned, will be set when added to list */
	p->contentType = CFG_PROPTYPE_LIST;
	p->content.list = arrNew(&propertyCollection, 4);
	p->line = 0;
	p->file = "<internally created>";
	return p;
}

cfgProperty_t	*cfgBuildDictProperty(void)
{
cfgProperty_t	*p;

	p = memCalloc(sizeof(cfgProperty_t));
	/* keep the name unassigned, will be set when added to list */
	p->contentType = CFG_PROPTYPE_DICT;
	p->content.dict = htabNew(&propertyCollection, 16, 75);
	p->line = 0;
	p->file = "<internally created>";
	return p;
}

void	cfgAppendToArrayProperty(cfgProperty_t *arr, cfgProperty_t *p)
{
	if(p == NULL)
		return;
	if(arr->contentType != CFG_PROPTYPE_LIST){
		cfgPropertyFree(p);
		return;
	}
	arrAddObject(arr->content.list, p);
	cfgInvalidateCache();	/* arr may have been part of config file */
}

void	cfgRemoveFromArrayProperty(cfgProperty_t *arr, btint index)
{
	if(arr->contentType != CFG_PROPTYPE_LIST)
		return;
	arrRemoveObjectAtIndex(arr->content.list, index);
	cfgInvalidateCache();	/* arr may have been part of config file */
}

void	cfgSetProperty(btu8 *name, plist_t *plist, cfgProperty_t *value)
{
cfgProperty_t	dummyProp;

	if(plist->contentType != CFG_PROPTYPE_DICT){
		EXC_THROW_MSG(excParseError, "trying to add named member to non-dictionary");
	}
	if(value == NULL){	/* delete property */
		dummyProp.name = name;
		htabRemoveObjectEqualTo(plist->content.dict, &dummyProp);
	}else{
		if(value->name != NULL){
			EXC_THROW_MSG(excParseError, "trying to make property member of two lists");
		}
		value->name = utilStrDup(name);
		htabAddObject(plist->content.dict, value);
	}
	cfgInvalidateCache();
}

/* ------------------------------------------------------------------------- */

static void	printIndent(FILE *fp, btint indentLevel)
{
btint	i;

	for(i=0;i<indentLevel;i++)
		fprintf(fp, "   ");
}

static void	unparseString(FILE *fp, btint indentLevel, btu8 *s)
{
	putc('\"', fp);
	while(*s){
		if(*s == '\\' || *s == '\"'){	/* escape it */
			putc('\\', fp);
			putc(*s, fp);
		}else if(*s >= 0x20 && *s < 128){	/* ordinary ASCII */
			putc(*s, fp);
		}else{	/* output as escape code */
			fprintf(fp, "\\0%03o", *s);
		}
		s++;
	}
	putc('\"', fp);
}

static void	unparseProperty(FILE *fp, btint indentLevel, cfgProperty_t *p);

static void	unparseDict(FILE *fp, btint indentLevel, hashtable_t *htab, btbool printBraces)
{
array_t			*all;
btint			i, cnt;
cfgProperty_t	*p;

	if(printBraces)
		fprintf(fp, "{\n");
	all = htabAllObjects(htab);
	cnt = arrObjectCount(all);
	for(i=0;i<cnt;i++){
		p = arrObjectAtIndex(all, i);
		printIndent(fp, indentLevel);
		unparseString(fp, indentLevel, p->name);
		fprintf(fp, " = ");
		unparseProperty(fp, indentLevel + 1, p);
		fprintf(fp, ";\n");
	}
	/* avoid freeing the entries: */
	arrReplaceInterface(all, &dummyInterface);
	arrFree(all);
	if(printBraces){
		printIndent(fp, indentLevel - 1);
		fprintf(fp, "}");
	}
}

static void	unparseArray(FILE *fp, btint indentLevel, array_t *arr)
{
btint			i, cnt;
cfgProperty_t	*p;

	fprintf(fp, "(\n");
	cnt = arrObjectCount(arr);
	for(i=0;i<cnt;i++){
		p = arrObjectAtIndex(arr, i);
		printIndent(fp, indentLevel);
		unparseProperty(fp, indentLevel + 1, p);
		if(i+1 < cnt)
			putc(',', fp);
		putc('\n', fp);
	}
	printIndent(fp, indentLevel - 1);
	fprintf(fp, ")");
}

static void	unparseProperty(FILE *fp, btint il, cfgProperty_t *p)
{
	switch(p->contentType){
	case CFG_PROPTYPE_STRING: unparseString(fp, il, p->content.string);	break;
	case CFG_PROPTYPE_LIST:	unparseArray(fp, il, p->content.list);		break;
	case CFG_PROPTYPE_DICT:	unparseDict(fp, il, p->content.dict, 2);	break;
	default: 				fprintf(fp, "\"\"");
	}
}

void	cfgUnparse(cfgFile_t *file)
{
FILE		*fp;
struct stat	sbuf;
btint		storederr;
time_t		now;

	fp = fopen(file->fileName, "w");
	storederr = errno;
	if(fp == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "cfgparse: error creating file %s: [%d] %s\n", file->fileName, storederr, sysStrerror(storederr));
		return;
	}
	now = time(NULL);
	fprintf(fp, "/* Automatically generated at: %s", ctime(&now));
	fprintf(fp, " * Filename: %s\n", file->fileName);
	fprintf(fp, " * If you edit by hand, you run the risk that Sharity will overwrite the file!\n");
	fprintf(fp, " */\n\n");
	unparseDict(fp, 0, file->plist->content.dict, 0);
	fclose(fp);
	if(stat(file->fileName, &sbuf) != 0){
		memset(&sbuf, 0, sizeof(sbuf));
	}
	file->lastModify = (bti64)sbuf.st_mtime;
}

/* ------------------------------------------------------------------------- */

#ifdef CFGPARSE_TEST

void	main(int argc, char **argv)
{
hashtable_t	*dict;

	LOG_SET_MODE(LOGM_ERROR);
	LOG_SET_MODE(LOGM_WARN);
	LOG_SET_MODE(LOGM_RESOURCE);
	if(argc < 2){
		fprintf(stderr, "usage: %s <filename>\n", argv[0]);
		exit(1);
	}
	dict = cfgParseFile(argv[1]);
	if(dict != NULL){
		htabFree(dict);
		fprintf(stderr, "parsing succeeded\n");
	}
	exit(0);
}

#endif

/* ------------------------------------------------------------------------- */

