/*
 * Name: exception.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-12
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: exception.h,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */

#ifndef __EXCEPTION_H_INCLUDED__
#define __EXCEPTION_H_INCLUDED__

/*
General Description:
The macros in this header implement an exception handling mechanism similar
to that in Java. An example for exception handling code would look like this:

EXC_DEFINE(myException)

void	function(void)
{
void	*p;

	EXC_TRY
		p = malloc(20);
		call_this_function(p);
		do_something_else();
	EXC_CATCH		// this part is optional
		if(EXC_ISX(myException)){
			printf("myException occured\n");
		}else{
			printf("an other exception occured\n");
		}
	EXC_FINALLY		// this part is optional
		free(p);
	EXC_ENDTRY		// this is mandatory, of course
}

An EXC_TRY/EXC_ENDTRY-block must have an EXC_CATCH or an EXC_FINALLY or both
statements.

An example of throwing an exception could look like this:

void	call_this_function(void *p)
{
	p = do_this(p);
	if(p == NULL){
		EXC_THROW(myException);
	}
	do_that();
}
*/


#include <stdio.h>
#include <setjmp.h>

/* ------------------------------------------------------------------------- */

typedef struct excDomain{
	struct excDomain	*parent;
	char				*exceptionName;
	char				*exceptionMessage;
	char				inHandler;
	char				isHandled;
	jmp_buf				jumpBuffer;
}excDomain_t;

extern excDomain_t	*excCurrentDomain;

/* ------------------------------------------------------------------------- */

/*
 * return values of setjmp():
 * 0	direct return
 * 1	retry
 * 2	exception occured
 */

#define	EXC_TRY	\
	{	excDomain_t __myDomain = {NULL, NULL, NULL};				\
		__myDomain.parent = excCurrentDomain;						\
		excCurrentDomain = &__myDomain;							\
		if(setjmp(__myDomain.jumpBuffer) < 2){						\
			__myDomain.inHandler = 0; __myDomain.isHandled = 1;

#define	EXC_CATCH	\
		}else if(!__myDomain.inHandler){							\
			__myDomain.inHandler = __myDomain.isHandled = 1;

/* FINALLY block is optional! */
#define	EXC_FINALLY	\
		}															\
		excCurrentDomain = __myDomain.parent;						\
		{			/* balance braces with and without FINALLY */
		/* exceptions in FINALLY block are handled by parent domain */

#define	EXC_ENDTRY	\
		}															\
		excCurrentDomain = __myDomain.parent;						\
		if(!__myDomain.isHandled){									\
			EXC_THROW_MSG(__myDomain.exceptionName, __myDomain.exceptionMessage);\
		}															\
	}

/* ------------------------------------------------------------------------- */

#define	EXC_RETRY()	\
		longjmp(__myDomain.jumpBuffer, 1)

/* use the following for debugging:
 *		*(int *)0 = 0;										\
 */
#if 0
/* debug version: */
#define	EXC_THROW_MSG(name, msg)	*(int *)0 = 0
#else
#define	EXC_THROW_MSG(name, msg)	\
	do{														\
		if(excCurrentDomain != NULL){						\
			excCurrentDomain->isHandled = 0;				\
			excCurrentDomain->exceptionName = (name);		\
			excCurrentDomain->exceptionMessage = (msg);	\
			longjmp(excCurrentDomain->jumpBuffer, 2);		\
		}													\
	}while(0)
#endif

#define	EXC_THROW(name)		EXC_THROW_MSG(name, "")

#define	EXC_RETHROW()		EXC_THROW_MSG(EXC_NAME, EXC_MSG)

/* ------------------------------------------------------------------------- */

#define EXC_DEFINE(name) char	*name = {#name}

#define EXC_DECLARE(name) extern char	*name

#define	EXC_NAME	(__myDomain.exceptionName)

#define	EXC_MSG	(__myDomain.exceptionMessage)

/* ------------------------------------------------------------------------- */

#define	EXC_ISX(exc)	(EXC_NAME == (exc))


EXC_DECLARE(excInternalError);

#endif	/* __EXCEPTION_H_INCLUDED__ */
