/*
 * Name: hashtable.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-24
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: hashtable.c,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */


#include "stdheaders.h"
#include "memalloc.h"
#include "exception.h"
#include "hashtable.h"
#include "logging.h"

/* ------------------------------------------------------------------------- */

static btint	dummyCompare(void *o1, void *o2)
{
btu8	*this = o1, *other = o2;

	return this - other;
}

static btu32	dummyHashVal(void *obj)
{
btint	x = (btint)obj;

	return x ^ (x >> 6) ^ (x >> 12);
}

static collectionObjectI_t	dummyInterface = {
	dummyCompare,
	dummyHashVal,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

hashtable_t	*htabNew(const collectionObjectI_t *contentObjectInterface, btu32 initialSize, btu32 maxFillPercentage)
{
hashtable_t	*volatile this = NULL;

EXC_TRY
    this = memCalloc(sizeof(hashtable_t));
    this->tableSize = initialSize;
    this->maxFillPercentage = maxFillPercentage;
    this->table = memCalloc(this->tableSize * sizeof(hashNode_t));
	if(contentObjectInterface != NULL)
    	this->objectInterface = *contentObjectInterface;
	else
    	this->objectInterface = dummyInterface;
EXC_CATCH
	htabFree(this);
	EXC_RETHROW();
EXC_ENDTRY
    return this;
}

/* ------------------------------------------------------------------------- */

void		htabFree(hashtable_t *this)
{
btu32		i;
hashNode_t	*p, *q;

	if(this != NULL){
		if(this->table != NULL){
			for(i=0;i<this->tableSize;i++){
				for(p=&this->table[i];p!=NULL;p=q){
					if(this->objectInterface.notifyRemovedFrom != NULL && p->obj != NULL){
						this->objectInterface.notifyRemovedFrom(p->obj);
					}
					q = p->next;
					if(p != &this->table[i])
						memFree(p);			/* free hashnodes */
				}
			}
			memFree(this->table);
		}
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

static inline void	internalAddObject(hashtable_t *this, void *obj)
{
btu32		hashIndex = this->objectInterface.hashValue(obj) % this->tableSize;
hashNode_t	*p;

    if(this->table[hashIndex].obj == NULL){	/* entry is still empty */
        this->table[hashIndex].obj = obj;
        this->objectCount++;
        return;
    }
    for(p=&this->table[hashIndex];p!=NULL;p=p->next){
        if(p->obj != NULL){
            if(this->objectInterface.compare(obj, p->obj) == 0){	/* replace existing */
                if(this->objectInterface.notifyRemovedFrom != NULL){
                    this->objectInterface.notifyRemovedFrom(p->obj);
                }
                p->obj = obj;
                return;
            }
        }
    }
	p = memAlloc(sizeof(hashNode_t));	/* insert new hashnode */
	p->obj = obj;
	p->next = this->table[hashIndex].next;
	this->table[hashIndex].next = p;
	this->objectCount++;
}

/* ------------------------------------------------------------------------- */

static void	resize(hashtable_t *this, btu32 newSize)
{
hashNode_t	*p, *q, *oldTable = this->table;
btu32		i, oldTableSize = this->tableSize;
void		*obj;

	this->table = memCalloc(newSize * sizeof(hashNode_t));
    this->objectCount = 0;
    this->tableSize = newSize;
    for(i=0;i<oldTableSize;i++){
        for(p=&oldTable[i];p!=NULL;p=q){
            obj = p->obj;
            q = p->next;
            if(p != &oldTable[i])
                memFree(p);			/* free hashnodes */
            if(obj != NULL)
                internalAddObject(this, obj);
        }
    }
    memFree(oldTable);
    if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE)){
        logPrintf(LOGM_RESOURCE, "hashtable: increasing to size %d (and rehashing)\n", newSize);
    }
}

/* ------------------------------------------------------------------------- */

void		htabAddObject(hashtable_t *this, void *obj)
{
    if(this->objectCount * 100 >= this->tableSize * this->maxFillPercentage){
        resize(this, 2 * this->tableSize);
    }
	/* First increment refCount, then add to list. Otherwise replacing an
	   object with itself frees it. */
    if(this->objectInterface.notifyAddedTo != NULL){
        this->objectInterface.notifyAddedTo(obj);
    }
    internalAddObject(this, obj);
}

/* ------------------------------------------------------------------------- */

btbool		htabRemoveObjectEqualTo(hashtable_t *this, void *obj)
{
btu32		hashIndex = this->objectInterface.hashValue(obj) % this->tableSize;
hashNode_t	*p, *q, *previous = NULL;
btbool		removed = 0;
void		*object;

    for(p=&this->table[hashIndex];p!=NULL;p=p->next){
        if(p->obj != NULL && this->objectInterface.compare(obj, p->obj) == 0){
			removed = 1;
			object = p->obj;
            if(previous != NULL){			/* not the first */
                previous->next = p->next;	/* take out of linked list */
                memFree(p);					/* free our hashnode structure */
            }else{							/* found in first */
                if(p->next != NULL){		/* there are other entries */
                    q = p->next;
                    *p = *q;				/* structure copy */
                    memFree(q);				/* this node has become unnecessary */
                }else{						/* was the only entry here */
                    p->obj = NULL;			/* mark as invalid */
                }
            }
            this->objectCount--;
			/* first remove and then notify to avoid recursion if the notified
			 * function tries to remove object from hashtable */
            if(this->objectInterface.notifyRemovedFrom != NULL)
                this->objectInterface.notifyRemovedFrom(object);
            break;
        }
        previous = p;
    }
	return removed;
}

/* ------------------------------------------------------------------------- */

void		*htabFindObjectEqualTo(hashtable_t *this, void *obj)
{
btu32		hashIndex = this->objectInterface.hashValue(obj) % this->tableSize;
hashNode_t	*p;

    for(p=&this->table[hashIndex];p!=NULL;p=p->next){
        if(p->obj != NULL){
            if(this->objectInterface.compare(obj, p->obj) == 0){
                return p->obj;
            }
        }
    }
	return NULL;
}

/* ------------------------------------------------------------------------- */

array_t		*htabAllObjects(hashtable_t *this)
{
array_t		*arr;
hashNode_t	*p;
btu32		i, cnt;

	cnt = this->objectCount > 0 ? this->objectCount : 1;
	arr = arrNew(&this->objectInterface, cnt);
    for(i=0;i<this->tableSize;i++){
        for(p=&this->table[i];p!=NULL;p=p->next){
            if(p->obj != NULL)
                arrAddObject(arr, p->obj);
        }
    }
    return arr;
}

/* ------------------------------------------------------------------------- */

