/*
 * Name: help.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-04-02
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: help.c,v 1.2 2001/01/08 15:53:50 cs Exp $
 */


#include "stdheaders.h"
#include "help.h"
#include "util.h"
#include "logging.h"
#include "system.h"
#include "ipc.h"
#include "hashtable.h"
#include "memalloc.h"

/* ------------------------------------------------------------------------- */

typedef struct helpEntry{
	btu8	*itemName;
	btu8	*helpText;
}helpEntry_t;


static hashtable_t	*helpEntries = NULL;

/* ------------------------------------------------------------------------- */

static void	helpEntryFree(void *obj)
{
helpEntry_t	*e = obj;

	if(e != NULL){
		if(e->itemName != NULL)
			memFree(e->itemName);
		if(e->helpText != NULL)
			memFree(e->helpText);
		memset(e, 0, sizeof(*e));
		memFree(e);
	}
}

static helpEntry_t	*helpEntryNew(btu8 *itemName, btu8 *helpText)
{
helpEntry_t	*volatile e = NULL;

EXC_TRY
	e = memCalloc(sizeof(*e));
	e->itemName = utilStrDup(itemName);
	e->helpText = utilStrDup(helpText);
EXC_CATCH
	helpEntryFree(e);
	EXC_RETHROW();
EXC_ENDTRY
	return e;
}

/* ------------------------------------------------------------------------- */

static btint	helpEntryCompare(void *o1, void *o2)
{
helpEntry_t	*this = o1, *other = o2;

	return utilStrCmp(this->itemName, other->itemName, 0);
}

static btu32	helpEntryHash(void *obj)
{
helpEntry_t	*e = obj;

	return utilStrHash(e->itemName, 0);
}

static collectionObjectI_t	helpEntryInterface = {
	helpEntryCompare,
	helpEntryHash,
	NULL,
	helpEntryFree,
};

/* ------------------------------------------------------------------------- */

static void	createItem(btu8 *itemName, ipcMsg_t *text)
{
helpEntry_t	*e;

	if(itemName[0] == 0)
		return;
	ipcMsgEnsureSize(text, text->size + 1);
	text->buffer[text->size] = 0;	/* 0-terminate string */
	e = helpEntryNew(itemName, text->buffer);
	htabAddObject(helpEntries, e);
}

void	helpParseFile(btu8 *helpFile)
{
btu8		*p, *start, currentItem[256], lineBuffer[32768];
ipcMsg_t	*text;
FILE		*fp;
btsize		len;
btint		storedErrno;

	if(helpEntries == NULL)
		helpEntries = htabNew(&helpEntryInterface, 64, 75);
	fp = fopen(helpFile, "r");
	storedErrno = errno;
	if(fp == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "help: can\'t open help-file \'%s\' for read: [%d] %s\n", helpFile, storedErrno, sysStrerror(storedErrno));
		return;
	}
	currentItem[0] = 0;
	text = ipcMsgNew();
	while(fgets(lineBuffer, sizeof(lineBuffer), fp) != NULL){
		if(memcmp(lineBuffer, "<!--", 4) == 0){
			createItem(currentItem, text);
			p = lineBuffer + 4;
			while(isspace(*p))
				p++;
			start = p;
			while(isalnum(*p))
				p++;
			*p = 0;
			if(strlen(start) > ELEMENTS(currentItem)-1)
				start[ELEMENTS(currentItem)-1] = 0;	/* emergency brake */
			strcpy(currentItem, start);
			text->size = 0;
		}else{
			len = strlen(lineBuffer);
			ipcMsgEnsureSize(text, text->size + len);
			memcpy(text->buffer + text->size, lineBuffer, len);
			text->size += len;
		}
	}
	createItem(currentItem, text);
	ipcMsgFree(text);
	fclose(fp);
}

/* ------------------------------------------------------------------------- */

btu8	*helpTextForItem(btu8 *item)
{
helpEntry_t	*e;

	if(item == NULL || item[0] == 0)
		return NULL;	/* no item, no help text */
	if(helpEntries == NULL){
		return item;	/* no help file loaded so far */
	}
	e = htabFindObjectEqualTo(helpEntries, &item);
	if(e == NULL){		/* item not defined in help files */
		return item;	/* let the user know the name of the item */
	}
	return e->helpText;
}

/* ------------------------------------------------------------------------- */
