/*
 * Name: lru.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-11-09
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: lru.h,v 1.1.1.1 1999/04/12 13:56:38 cs Exp $
 */

#ifndef __lru_H_INCLUDED__
#define __lru_H_INCLUDED__

#include "memalloc.h"

/*
General Description:
This simple class can be used to find the least recently used object from
a set of objects. All objects in an lru-container must be subclasses of
'lruObject_t', which can be achieved by making the first entry of the
object's structure a 'lruObject_t'. To tell the list that an object has
been used, call 'lruTouch()'. To get the oldest object from the list,
call 'lruOldestRelativeTo(<list>, NULL)', to get the next younger object
relative to a given object, call 'lruOldestRelativeTo(<list>, <given object>)'.
This class does not use the 'collectionObjectI_t' like 'array_t' and
'hashtable_t', it only references the objects it contains.
The rest should be self-expaining.
*/

/* ------------------------------------------------------------------------- */

typedef struct lruObject{
    struct lruObject	*next;
    struct lruObject	**prevnext;
}lruObject_t;

typedef struct lru{
    lruObject_t			*head;
    lruObject_t			**tail;
    btu32				objectCount;
    bti32				pointerOffset;
}lru_t;

/* ------------------------------------------------------------------------- */

static inline lru_t	*lruNew(bti32 pointerOffset)
{
lru_t	*this = memCalloc(sizeof(lru_t));

    this->tail = &this->head;
    this->pointerOffset = pointerOffset;
    return this;
}

/* ------------------------------------------------------------------------- */

static inline void	lruFree(lru_t *this)
{
    memFree(this);
}

/* ------------------------------------------------------------------------- */

static inline void	lruAddObject(lru_t *this, void *obj)
{
lruObject_t	*element = (lruObject_t *)((bti8 *)obj + this->pointerOffset);
    
	if(element->prevnext != NULL){	/* element already in an LRU */
		DEBUG_TRAP();
	}
	element->prevnext = this->tail;
	*this->tail = element;
	this->tail = &element->next;
	element->next = NULL;
    this->objectCount++;
}

/* ------------------------------------------------------------------------- */

static inline void	lruTouch(lru_t *this, void *obj)
{
lruObject_t	*element = (lruObject_t *)((bti8 *)obj + this->pointerOffset);

	if(element->next == NULL)			/* element is already at the tail! */
		return;
    if(element->prevnext == NULL){		/* element not in LRU list */
        DEBUG_TRAP();
	}
	*element->prevnext = element->next;	/* take element out of the list */
	element->next->prevnext = element->prevnext;
	element->prevnext = this->tail;		/* put element into tail position */
	*this->tail = element;
	this->tail = &element->next;
	element->next = NULL;
}

/* ------------------------------------------------------------------------- */

static inline void	lruRemoveObject(lru_t *this, void *obj)
{
lruObject_t	*element = (lruObject_t *)((bti8 *)obj + this->pointerOffset);

	if(element->prevnext == NULL)
		return;		/* object not in list, no need to remove */
	*element->prevnext = element->next;
	if(this->tail == &element->next){	/* if we are the last element */
		this->tail = element->prevnext;	/* previous is last one now */
	}
	if(element->next != NULL){
		element->next->prevnext = element->prevnext;
	}
	element->prevnext = NULL;
	element->next = NULL;
	this->objectCount--;
}

/* ------------------------------------------------------------------------- */

static inline void	*lruOldestRelativeTo(lru_t *this, void *obj)
{
lruObject_t	*element = (lruObject_t *)((bti8 *)obj + this->pointerOffset);
bti8		*rval;

	rval = (obj == NULL) ? (bti8 *)this->head : (bti8 *)element->next;
	return rval == NULL ? NULL : rval - this->pointerOffset;
}

/* ------------------------------------------------------------------------- */

static inline btu32	lruObjectCount(lru_t *this)
{
    return this->objectCount;
}

/* ------------------------------------------------------------------------- */

static inline btbool	lruStructureIsOK(lru_t *this)
{
btint		count = 0;
lruObject_t	*e;

	for(e = this->head;e != NULL; e=e->next)
		count++;
	return count == this->objectCount;
}

/* ------------------------------------------------------------------------- */

#endif	/* __lru_H_INCLUDED__ */
