/*
 * Name: unicode.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-22
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: unicode.c,v 1.6 2001/05/31 09:38:08 cs Exp $
 */


#include "stdheaders.h"
#include "memalloc.h"
#include "unicode.h"
#include "logging.h"
#include "util.h"

/* ------------------------------------------------------------------------- */

typedef struct	uniTable{
	btuni			codeBase;
	btsize			codeEnd;	/* first char that is not part of table */
	uniProperty_t	*tableData;
}uniTable_t;

typedef struct asciiToUniTable{
	btu8					myIndex;
	btu16					htabSize;
	btuni					*uniRep;	/* [256], allocated on demand */
	struct asciiToUniTable	**next;		/* hashtable, allocated on demand */
}asciiToUniTable_t;


/* ------------------------------------------------------------------------- */

static btuni	codepage437[128] = {	/* this is the default codepage for the server */
	0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
	0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
	0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
	0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
	0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
	0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
	0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
	0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
	0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x255e, 0x255f,
	0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256c, 0x2567,
	0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256b,
	0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
	0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x00b5, 0x03c4,
	0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
	0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
	0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0,
};

/* ------------------------------------------------------------------------- */
/* ------------------------ unicode initialisation ------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct encodingInfo{
	btuni				errorChar;
	btu8				*asciiErrorChar;
	btint				maxAsciiLen;
	asciiToUniTable_t	decoder;
}encodingInfo_t;

static encodingInfo_t	encodingInfo[UNI_NUM_ENCODINGS];
static uniTable_t		*uniTables;		/* tables representing unicode entities */
static btsize			uniTableCount;	/* number of tables for representation */

/* ------------------------------------------------------------------------- */

static inline btbool	isTouched(uniProperty_t *uniTable, btsize index)
{
btint	i;

	for(i=0;i<UNI_NUM_ENCODINGS;i++)
		if(uniTable[index].encodingRep[i].ptr != NULL)
			return 1;
	return uniTable[index].info.upcaseDiff != 0 || uniTable[index].info.lowcaseDiff != 0;
}

struct gap{
	btsize	base;
	btsize	count;
};

static int gapCompare(const void *e1, const void *e2)
{
const struct gap	*g1 = e1, *g2 = e2;

	if(g1->base < g2->base)
		return -1;
	if(g1->base > g2->base)
		return 1;
	return 0;
}

static void	compressTable(uniProperty_t *uniTable, btsize maxBlocks, btsize *memUsage)
{
btsize		base = 0, count = 0, i, j, k;
struct gap	*gaps;

	if(uniTables != NULL){
		for(i=0;i<uniTableCount;i++){
			if(uniTables[i].tableData != NULL)
				memFree(uniTables[i].tableData);
		}
		memFree(uniTables);
	}
	uniTables = memCalloc(maxBlocks * sizeof(uniTable_t));
	gaps = memCalloc((maxBlocks + 1) * sizeof(*gaps));
	for(i=1;i<maxBlocks+1;i++)	/* leave one null-gap at the beginning */
		gaps[i].base = 65536;	/* dummy-gaps that will be sorted to the end */
	for(i=0;i<65536;i++){
		if(isTouched(uniTable, i)){
			if(count > 0){
				for(j=0;j<maxBlocks;j++){
					/* Insert before next smaller gap, throwing out the smallest gaps
					 * at the end of the list. Only replace null-gap at j==0 (the
					 * one with base=0, count=0) with a larger base=0 gap.
					 */
					if(count > gaps[j].count && (j != 0 || base == 0)){
						for(k=maxBlocks-1;k>j;k--){
							gaps[k] = gaps[k-1];
						}
						gaps[j].count = count;
						gaps[j].base = base;
						break;
					}
				}
			}
			base = i + 1;
			count = 0;
		}else{
			count++;
		}
	}
	gaps[maxBlocks].base = base;	/* last gap starts one behind the last touched char */
	gaps[maxBlocks].count = 65536 - base;
	for(i=0;i<=maxBlocks;i++){		/* sort 0-sized gaps to the end in qsort */
		if(gaps[i].count <= 0)
			gaps[i].base = 65536;
	}
	qsort(gaps, maxBlocks + 1, sizeof(*gaps), gapCompare);
	i=0;
	for(j=1;j<maxBlocks+1 && i<maxBlocks;j++){
		uniTables[i].codeBase = gaps[j-1].base + gaps[j-1].count;
		uniTables[i].codeEnd = gaps[j].base;
		if(uniTables[i].codeEnd - uniTables[i].codeBase > 0)
			i++;
		if(gaps[j].base >= 65536)
			break;
	}
	uniTableCount = i;
	memFree(gaps);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "unicode: using %d tables for forward mapping:\n", uniTableCount);
	for(i=0;i<uniTableCount;i++){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "table %d: codes %d to (excluding) %d\n", i, uniTables[i].codeBase, uniTables[i].codeEnd);
		count = (uniTables[i].codeEnd - uniTables[i].codeBase) * sizeof(uniProperty_t);
		if(memUsage != NULL)
			*memUsage += count;
		uniTables[i].tableData = memAlloc(count);
		memcpy(uniTables[i].tableData, &uniTable[uniTables[i].codeBase], count);
	}
}

/* ------------------------------------------------------------------------- */

static void	asciiToUniTableRehash(asciiToUniTable_t *table, btsize *memUsage)
{
btint				i, newSize = table->htabSize * 2;
asciiToUniTable_t	**newTab;

	if(newSize < 1)
		newSize = 1;
	newTab = memCalloc(newSize * sizeof(asciiToUniTable_t *));
	for(i=0;i<table->htabSize;i++){
		if(table->next[i] != NULL){
			newTab[table->next[i]->myIndex % newSize] = table->next[i];
		}
	}
	if(memUsage != NULL)
		*memUsage += (newSize - table->htabSize) * sizeof(asciiToUniTable_t *);
	if(table->next != NULL)
		memFree(table->next);
	table->next = newTab;
	table->htabSize = newSize;
}

static void	asciiToUniTableInsert(asciiToUniTable_t *table, asciiToUniTable_t *successor, btsize *memUsage)
{
	while(table->next == NULL || table->next[successor->myIndex % table->htabSize] != NULL)
		asciiToUniTableRehash(table, memUsage);
	table->next[successor->myIndex % table->htabSize] = successor;
}

static inline void	reverseTableInit(asciiToUniTable_t *table, btu8 index)
{
	memset(table, 0, sizeof(*table));
	table->myIndex = index;
}

static void	freeReverseTable(asciiToUniTable_t *table, btbool freeRoot)
{
int		i;

	if(table->next != NULL){
		for(i=0;i<table->htabSize;i++){
			if(table->next[i] != NULL){
				freeReverseTable(table->next[i], 1);
				table->next[i] = NULL;
			}
		}
		memFree(table->next);
		table->next = NULL;
	}
	if(table->uniRep != NULL){
		memFree(table->uniRep);
		table->uniRep = NULL;
	}
	if(freeRoot)
		memFree(table);
}

static void	calcReverseTables(btint encoding, encodingInfo_t *info, btsize *memUsage)
{
btsize				i, len;
btu8				buffer[512], *p, *q;
asciiToUniTable_t	*table, *next;
btbool				mapOK;

	freeReverseTable(&info->decoder, 0);
	reverseTableInit(&info->decoder, 0);
	for(i=65535;i>=0;i--){	/* use reverse order to give precedence to lower codes in case of a conflict */
		q = buffer;
		mapOK = convCharUniToAscii(i, &q, encoding);
		len = q - buffer;
		if(len > info->maxAsciiLen)
			info->maxAsciiLen = len;
		if(mapOK && len > 0){	/* if character is mapped */
			table = &info->decoder;
			for(p=buffer;p<q-1;p++){
				next = NULL;
				if(table->next != NULL)
					next = table->next[*p % table->htabSize];
				if(next == NULL || next->myIndex != *p){
					next = memAlloc(sizeof(asciiToUniTable_t));
					reverseTableInit(next, *p);
					if(memUsage != NULL)
						*memUsage += sizeof(asciiToUniTable_t);
					asciiToUniTableInsert(table, next, memUsage);
				}
				table = next;
			}
			if(table->uniRep != NULL && table->uniRep[*p] != 0 && LOG_MODE_IS_ACTIVE(LOGM_WARN)){
				logPrintf(LOGM_WARN, "unicode: unicode chars 0x%04x and 0x%04x have same mapping! (enc=%d)\n", table->uniRep[*p], i, encoding);
			}
			if(table->uniRep == NULL){
				table->uniRep = memCalloc(sizeof(btuni) * 256);
				if(memUsage != NULL)
					*memUsage += sizeof(btuni) * 256;
			}
			table->uniRep[*p] = i;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	trivialSetup(uniProperty_t *uniTable, btsize size, btbool withCp437)
{
btsize		i, j;
btuni		c;

	if(withCp437){
		for(i=0;i<128;i++){	/* server is initialized to codepage 437 */
			c = codepage437[i];
			if(c < size){
				if((uniTable[c].info.flags & (1<<UNI_ENCODING_SERVER)) != 0){
					memFree(uniTable[c].encodingRep[UNI_ENCODING_SERVER].ptr);
					uniTable[c].info.flags &= ~(1<<UNI_ENCODING_SERVER);
				}
				uniTable[c].encodingRep[UNI_ENCODING_SERVER].ptr = NULL;
				uniTable[c].encodingRep[UNI_ENCODING_SERVER].ascii[0] = i + 128;
			}
		}
	}
	for(i=0;i<size;i++){
		if((i >= 0x61 && i <= 0x7a) || (i >= 0xe0 && i <= 0xfe && i != 0xf7)){
			uniTable[i].info.upcaseDiffNegative = 1;
			uniTable[i].info.upcaseDiff = 0x20;
		}else{
			uniTable[i].info.upcaseDiffNegative = 0;
			uniTable[i].info.upcaseDiff = 0;
		}
		uniTable[i].info.lowcaseDiffNegative = 0;
		if((i >= 0x41 && i <= 0x5a) || (i >= 0xc0 && i <= 0xde && i != 0xd7))
			uniTable[i].info.lowcaseDiff = 0x20;
		else
			uniTable[i].info.lowcaseDiff = 0;
		for(j=0;j<UNI_NUM_ENCODINGS;j++){
			if(i < (j == UNI_ENCODING_SERVER && withCp437 ? 128 : 256)){
				if((uniTable[i].info.flags & (1<<j)) != 0){
					memFree(uniTable[i].encodingRep[j].ptr);
					uniTable[i].info.flags &= ~(1<<j);
				}
				uniTable[i].encodingRep[j].ptr = NULL;
				uniTable[i].encodingRep[j].ascii[0] = i;
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

void	uniTrivialInit(void)
{
btsize	i;
btsize	tableBytes = 0;

	if(uniTables != NULL){
		for(i=0;i<uniTableCount;i++){
			if(uniTables[i].tableData != NULL){
				memFree(uniTables[i].tableData);
				uniTables[i].tableData = NULL;
			}
		}
	}else{
		uniTables = memCalloc(sizeof(uniTable_t));
	}
	uniTableCount = 1;
	/* set up some reasonable defaults */
	uniTables->codeBase = 0;
	uniTables->codeEnd = 256;
	uniTables->tableData = memCalloc(256 * sizeof(uniProperty_t));
	trivialSetup(uniTables->tableData, 256, 1);
	for(i=0;i<UNI_NUM_ENCODINGS;i++){
		encodingInfo[i].errorChar = '?';
		encodingInfo[i].asciiErrorChar = "?";
	}
	for(i=0;i<UNI_NUM_ENCODINGS;i++){
		calcReverseTables(i, &encodingInfo[i], &tableBytes);
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
		logPrintf(LOGM_INFO, "unicode: trivialInit: using %d bytes for forward and %d bytes for reverse mapping tables\n", 256 * sizeof(uniProperty_t), tableBytes);
}

/* ------------------------------------------------------------------------- */

void	uniInit(btu8 *caseConvertPath, btu8 *serverMapPath, btu8 *clientMapPath, btsize maxBlocks)
{
uniProperty_t	*uniTable = NULL;
btint			i;
btu8			*mapFile;
btsize			forwardBytes = 0, reverseBytes = 0;

	uniTable = memCalloc(65536 * sizeof(uniProperty_t));
	trivialSetup(uniTable, 65536, 0);
	if(caseConvertPath != NULL && !uniParseCaseConvertFile(caseConvertPath, uniTable))
		goto errorOccured;
	for(i=0;i<UNI_NUM_ENCODINGS;i++){
		mapFile = i == UNI_ENCODING_SERVER ? serverMapPath : clientMapPath;
		if(mapFile != NULL && !uniParseMappingFile(mapFile, i, uniTable, &encodingInfo[i].errorChar, &encodingInfo[i].asciiErrorChar))
			goto errorOccured;
	}
	compressTable(uniTable, maxBlocks, &forwardBytes);
	memFree(uniTable);
	for(i=0;i<UNI_NUM_ENCODINGS;i++){
		calcReverseTables(i, &encodingInfo[i], &reverseBytes);
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
		logPrintf(LOGM_INFO, "unicode: init: using %d bytes for forward and %d bytes for reverse mapping tables\n", forwardBytes, reverseBytes);
	return;
errorOccured:
	memFree(uniTable);
	if(LOG_MODE_IS_ACTIVE(LOGM_WARN))
		logPrintf(LOGM_WARN, "unicode: error parsing tables, using defaults\n");
	uniTrivialInit();
}

/* ------------------------------------------------------------------------- */
/* -------------------------- unicode operations --------------------------- */
/* ------------------------------------------------------------------------- */

btuni	*uniStrDup(btuni *s)
{
btsize	len;
btuni	*new;

	if(s == NULL)
		return NULL;
	len = uniStrLen(s) + 1;
    new = memAlloc(len * sizeof(btuni));
    memcpy(new, s, len * sizeof(btuni));
    return new;
}

void	uniStrCpy(btuni *dest, btuni *src, btsize arrSize)
{
int		i;

	arrSize--;
	for(i=0; src[i] != 0 && i < arrSize; i++){
		dest[i] = src[i];
	}
	dest[i] = 0;
}

void	asciiStrCpy(btu8 *dest, btu8 *src, btsize arrSize)
{
int		i;

	arrSize--;
	for(i=0; src[i] != 0 && i < arrSize; i++){
		dest[i] = src[i];
	}
	dest[i] = 0;
}

void	uniStrAddHash(btu32 *crcp, btuni *s, btbool isCaseSensitive)
{
btu32	crc = *crcp;
btuni	c;

	if(isCaseSensitive){
		while(*s){
			c = *s;
			crc = utilCrc32table[(crc ^ c) & 0xff] ^ (crc >> 8);
			s++;
		}
	}else{
		while(*s){
			c = uniCharToupper(*s);
			crc = utilCrc32table[(crc ^ c) & 0xff] ^ (crc >> 8);
			s++;
		}
	}
	*crcp = crc;
}

static inline uniProperty_t	*uniLookup(btuni uniChar)
{
btsize	i;

	for(i=0;i<uniTableCount;i++){
		if(uniChar >= uniTables[i].codeBase && uniChar < uniTables[i].codeEnd){
			return &uniTables[i].tableData[uniChar - uniTables[i].codeBase];
		}
	}
	return NULL;
}

btuni	uniCharToupper(btuni c)
{
uniProperty_t	*entry;

	if((entry = uniLookup(c)) == NULL)
		return c;
	if(entry->info.upcaseDiffNegative)
		return c - entry->info.upcaseDiff;
	else
		return c + entry->info.upcaseDiff;
}

btuni	uniCharTolower(btuni c)
{
uniProperty_t	*entry;

	if((entry = uniLookup(c)) == NULL)
		return c;
	if(entry->info.lowcaseDiffNegative)
		return c - entry->info.lowcaseDiff;
	else
		return c + entry->info.lowcaseDiff;
}

btu8	asciiCharToupper(btu8 c)
{
	return uniCharToupper(c);	/* this is not strictly correct... */
}

btu8	asciiCharTolower(btu8 c)
{
	return uniCharTolower(c);	/* this is not strictly correct... */
}

/* ------------------------------------------------------------------------- */
/* --------------------------- basic conversion ---------------------------- */
/* ------------------------------------------------------------------------- */

btuni	convCharAsciiToUni(btu8 **string, btint encoding)
{
asciiToUniTable_t	*table = &encodingInfo[encoding].decoder;
btu8				c, *p, *consumed;
btuni				u = encodingInfo[encoding].errorChar;

	p = *string;
	if((c = *p++) == 0)
		return 0;
	consumed = p;	/* we consume at least one char, even if it becomes an error char */
	for(;;){
		if(table->uniRep != NULL && table->uniRep[c] != 0){
			u = table->uniRep[c];
			consumed = p;
		}
		if(table->next == NULL)
			break;
		/* continue if we can because we search for longest match */
		table = table->next[c % table->htabSize];
		if(table == NULL || table->myIndex != c)
			break;
		if((c = *p++) == 0)
			break;
	}
	*string = consumed;
	return u;
}

btbool	convCharUniToAscii(btuni c, btu8 **string, btint encoding)
{
uniProperty_t	*entry;
btsize			i;
btu8			*p;

	if((entry = uniLookup(c)) == NULL || entry->encodingRep[encoding].ptr == NULL){
		i = strlen(encodingInfo[encoding].asciiErrorChar);
		memcpy(*string, encodingInfo[encoding].asciiErrorChar, i);
		*string += i;
		return 0;	/* mapping had an error */
	}else{
		if((entry->info.flags & (1<<encoding)) == 0){
			for(i=0;i<sizeof(btu8 *);i++){
				if(entry->encodingRep[encoding].ascii[i] == 0)
					break;
				*(*string)++ = entry->encodingRep[encoding].ascii[i];
			}
		}else{
			for(p=entry->encodingRep[encoding].ptr; *p!=0; p++){
				*(*string)++ = *p;
			}
		}
		return 1;	/* mapping was OK */
	}
}

/* ------------------------------------------------------------------------- */
/* -------------------------- derived conversion --------------------------- */
/* ------------------------------------------------------------------------- */

void	convAsciiToUni(btu8 *ascii, btuni *uni, btsize arrSize, btint encoding)
{
btint	i;

	arrSize--;	/* reserve one char for terminating null */
	for(i=0; *ascii != 0 && i < arrSize; i++){
		uni[i] = convCharAsciiToUni(&ascii, encoding);
	}
	uni[i] = 0;
}

void	convUniToAscii(btuni *uni, btu8 *ascii, btsize arrSize, btint encoding)
{
btint	i;

	arrSize -= encodingInfo[encoding].maxAsciiLen;
	for(i=0; uni[i] != 0 && i < arrSize; i++){
		convCharUniToAscii(uni[i], &ascii, encoding);
	}
	*ascii = 0;
}

/* ------------------------------------------------------------------------- */

btu8	*uniToCstring(btuni *uni)
{
static btu8	rval[512];	/* this should be sufficient for debug prints */

	convUniToAscii(uni, rval, sizeof(rval), UNI_ENCODING_CLIENT);
	return rval;
}

/* ------------------------------------------------------------------------- */
/* ----------------------- unicode wildcard matching ----------------------- */
/* ------------------------------------------------------------------------- */

/* The following is derived from Rich Salz' wildmat.c which is also included
 * in this source tree. The size of characters has been changed to btuni.
 */

#define MATCH_SUCCESS			1
#define MATCH_FAILED			0
#define MATCH_ABORT				-1

#define UNI_CHAR_NEGATE_CLASS	((btuni)'^')
#define	UNI_CHAR_OPENBRACKET	((btuni)'[')
#define	UNI_CHAR_CLOSEBRACKET	((btuni)']')
#define	UNI_CHAR_MINUS			((btuni)'-')

/* ------------------------------------------------------------------------- */

/*
 *	Match text and p, return MATCH_SUCCESS, MATCH_FAILED, or MATCH_ABORT.
 */
static int	uniDoMatch(btuni *text, btuni *p)
{
register int	last;
register int	matched;
register int	reverse;

	for ( ; *p; text++, p++) {
		if (*text == 0 && *p != UNI_CHAR_ASTERISK)
			return MATCH_ABORT;
		switch (*p) {
		case UNI_CHAR_BACKSLASH:
			/* Literal match with following character. */
			p++;
			/* FALLTHROUGH */
		default:
			if (*text != *p)
				return MATCH_FAILED;
			continue;
		case UNI_CHAR_QUESTION:
			/* Match anything. */
			continue;
		case UNI_CHAR_ASTERISK:
			while (*++p == UNI_CHAR_ASTERISK)
				/* Consecutive stars act just like one. */
				continue;
			if (*p == 0)
				/* Trailing star matches everything. */
				return MATCH_SUCCESS;
			while (*text)
				if ((matched = uniDoMatch(text++, p)) != MATCH_FAILED)
					return matched;
			return MATCH_ABORT;
		case UNI_CHAR_OPENBRACKET:
			reverse = p[1] == UNI_CHAR_NEGATE_CLASS ? MATCH_SUCCESS : MATCH_FAILED;
			if (reverse)
				/* Inverted character class. */
				p++;
			matched = MATCH_FAILED;
			if (p[1] == UNI_CHAR_CLOSEBRACKET || p[1] == UNI_CHAR_MINUS)
				if (*++p == *text)
					matched = MATCH_SUCCESS;
			for (last = *p; *++p && *p != UNI_CHAR_CLOSEBRACKET; last = *p)
				/* This next line requires a good C compiler. */
				if (*p == UNI_CHAR_MINUS && p[1] != UNI_CHAR_CLOSEBRACKET
					? *text <= *++p && *text >= last : *text == *p)
					matched = MATCH_SUCCESS;
			if (matched == reverse)
				return MATCH_FAILED;
			continue;
		}
	}
	return *text == 0;
}

/* ------------------------------------------------------------------------- */

btbool	uniWildcardMatch(btuni *text, btuni *pattern)
{
	return uniDoMatch(text, pattern) == MATCH_SUCCESS;
}


/* ------------------------------------------------------------------------- */
