/*
 * Name: nfsfileadapter.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-11-08
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfsfileadapter.c,v 1.10 2002/03/22 17:48:02 cs Exp $
 */

#include "stdheaders.h"
#include "nfs_prot.h"
#include "nfsfileadapter.h"
#include "memalloc.h"
#include "process.h"
#include "blockpool.h"
#include "errcodes.h"
#include "logging.h"
#include "unicode.h"
#include "nfslookup.h"
#include "module.h"
#include "backend.h"
#include "hashtable.h"
#include "nfs2const.h"
#include "cfgparse.h"

#define	NAME_ARRAY_SIZE	(NFS_MAXNAMLEN + 1)
#define	PATH_ARRAY_SIZE	(NFS_MAXPATHLEN + 1)

/* ------------------------------------------------------------------------- */

typedef struct backendImp{
	void	(*vfsRetain)(void *obj);
	void	(*vfsRelease)(void *obj);
	void	(*vfsStatfs)(void *obj, void *user, void *caller);
	void	(*vfsConnect)(void *obj, void *user, void *caller);

	void	(*fileRetain)(void *obj);
	void	(*fileRelease)(void *obj);
	void	(*fileLookup)(void *parent, btuni *name, void *user, void *caller);
	void	(*fileGetAttr)(void *obj, void *user, void *caller);
	void	(*fileSetAttr)(void *obj, void *user, beFileSetAttr_t *attr, void *caller);
	void	(*fileReaddir)(void *obj, void *user, btint startIndex, btbool (*resultCallback)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir), void *caller);
	void	(*fileCreate)(void *parent, btuni *name, btbool isDir, void *user, beFileSetAttr_t *attr, void *caller);
	void	(*fileRemove)(void *parent, btuni *name, btbool isDir, void *user, void *caller);
	void	(*fileRename)(void *fromDir, btuni *fromName, void *toDir, btuni *toName, void *user, void *caller);
	void	(*fileRW)(void *obj, btint operation, btbool writeThrough, btu8 *data, btoffset offset, btsize len, void *user, void *caller);
	void	(*fileHardLink)(void *from, void *toDir, btuni *name, void *user, void *caller);
	void	(*fileSymlink)(void *inDir, btuni *name, btuni *linkPath, void *user, beFileSetAttr_t *nfsAttr, void *caller);
	void	(*fileReadlink)(void *obj, void *user, void *caller);
	void	*(*fileParent)(void *obj);
}backendImp_t;

struct nfsFileAdapter{
	void				*vfs;
	backendImp_t		backend;
	btint				refCount;
	btint				dirLinks;	/* number of links given to directories if not supplied by backend */
	btint				dirSize;	/* faked size for directories */
	struct nfsLookup	*lookupStrategy;
	btbool				changeMicrosecondFileTimes;
};

/* ------------------------------------------------------------------------- */

typedef struct nfsCtx{
    PROC_VARSET;
    nfsFileAdapter_t	*a;
	void				*user;
	void				*file;
	void				*otherFile;
    btbool				isDir;
	btbool				doFileLookup;
    struct entry		*entryHead;
    struct entry		**entryTail;
    btsize				restLen;
    btu8				*data;
	btint				functionIndex;
    union {
        nfs_fh		getattr;
        sattrargs	setattr;
        diropargs	lookup;
        nfs_fh		readlink;
        readargs	read;
        writeargs	write;
        createargs	create;
        diropargs	remove;
        renameargs	rename;
        linkargs	link;
        symlinkargs	symlink;
        createargs	mkdir;
        diropargs	rmdir;
        readdirargs	readdir;
        nfs_fh		statfs;
    }				*arg;
	beFileSetAttr_t		setAttr;
    btuni				uniname[NAME_ARRAY_SIZE];
    btuni				unipath[PATH_ARRAY_SIZE];
}nfsCtx_t;

static blockpool_t	*nfsPool = NULL;

static inline void	freeEntries(struct entry **headp);

static void	nfsCtxFree(void *block)
{
nfsCtx_t	*this = block;

	if(this->file != NULL){
		this->a->backend.fileRelease(this->file);
		this->file = NULL;
	}
	if(this->otherFile != NULL){
		this->a->backend.fileRelease(this->otherFile);
		this->otherFile = NULL;
	}
	freeEntries(&this->entryHead);
    if(this->data != NULL)
        memFree(this->data);
    poolReturnBlock(nfsPool, block);
}

/* ------------------------------------------------------------------------- */

static btint	nfsFaCompare(void *obj, void *otherObj)
{
nfsFileAdapter_t	*this = obj, *other = otherObj;

	return (btu8 *)this->vfs - (btu8 *)other->vfs;
}

static btu32	nfsFaHashVal(void *obj)
{
nfsFileAdapter_t	*this = obj;

	return (btint)this->vfs ^ ((btint)this->vfs >> 8);
}

static collectionObjectI_t	faInterace = {
	nfsFaCompare,
	nfsFaHashVal,
	NULL,
	NULL
};

static hashtable_t	*allFa = NULL;

/* ------------------------------------------------------------------------- */

static void	nfsFaFree(nfsFileAdapter_t *fa)
{
	if(fa != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "nfsFileAdapter: freeing object at 0x%08x\n", (int)fa);
		if(fa->vfs != NULL){
			htabRemoveObjectEqualTo(allFa, fa);
			fa->backend.vfsRelease(fa->vfs);
			fa->vfs = NULL;
		}
		if(fa->lookupStrategy != NULL){
			luRelease(fa->lookupStrategy);
			fa->lookupStrategy = NULL;
		}
		memFree(fa);
	}
}

/* ------------------------------------------------------------------------- */

static nfsFileAdapter_t	*nfsFaNew(struct module *backend, void *vfs, struct cfgProperty **cfgPath, btint numEntries)
{
nfsFileAdapter_t	*volatile fa;
nfsCtx_t			*context = NULL;	/* not really needed... */

    if(nfsPool == NULL)
        nfsPool = poolNew(sizeof(nfsCtx_t), (btu8 *)&context->setAttr - (btu8 *)context);
EXC_TRY
	fa = memCalloc(sizeof(nfsFileAdapter_t));
	fa->lookupStrategy = luNew(vfs, backend, cfgPath, numEntries);
	fa->backend.vfsRetain = backend->implementationOf(BE_VFS_RETAIN);
	fa->backend.vfsRelease = backend->implementationOf(BE_VFS_RELEASE);
	fa->backend.vfsStatfs = backend->implementationOf(BE_VFS_STATFS);
	fa->backend.vfsConnect = backend->implementationOf(BE_VFS_CONNECT);
	
	fa->backend.fileRetain = backend->implementationOf(BE_FILE_RETAIN);
	fa->backend.fileRelease = backend->implementationOf(BE_FILE_RELEASE);
	fa->backend.fileLookup = backend->implementationOf(BE_FILE_LOOKUP);
	fa->backend.fileGetAttr = backend->implementationOf(BE_FILE_GETATTR);
	fa->backend.fileReaddir = backend->implementationOf(BE_FILE_READDIR);
	fa->backend.fileSetAttr = backend->implementationOf(BE_FILE_SETATTR);
	fa->backend.fileCreate = backend->implementationOf(BE_FILE_CREATE);
	fa->backend.fileRemove = backend->implementationOf(BE_FILE_REMOVE);
	fa->backend.fileRename = backend->implementationOf(BE_FILE_RENAME);
	fa->backend.fileRW = backend->implementationOf(BE_FILE_RW);
	fa->backend.fileHardLink = backend->implementationOf(BE_FILE_HARDLINK);
	fa->backend.fileSymlink = backend->implementationOf(BE_FILE_SYMLINK);
	fa->backend.fileReadlink = backend->implementationOf(BE_FILE_READLINK);
	fa->backend.fileParent = backend->implementationOf(BE_FILE_PARENT);

	fa->vfs = vfs;
	fa->backend.vfsRetain(fa->vfs);
	fa->dirLinks = cfgIntValue(cfgQueryFromArray("fakedDirLinks", numEntries, cfgPath), 2);
	fa->dirSize = cfgIntValue(cfgQueryFromArray("fakedDirSize", numEntries, cfgPath), 1024 * 128);
	fa->changeMicrosecondFileTimes = cfgEnumValue(cfgQueryFromArray("changeMicrosecondFileTimes", numEntries, cfgPath), cfgYesNoEnum, 0);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "nfsFileAdapter: allocated new object at 0x%08x\n", (int)fa);
EXC_CATCH
	nfsFaFree(fa);
	EXC_RETHROW();
EXC_ENDTRY
	return fa;
}

nfsFileAdapter_t	*nfsFaForBackend(struct module *backend, void *vfs, struct cfgProperty **cfgPath, btint numEntries)
{
nfsFileAdapter_t	*fa;

	if(allFa == NULL){
		allFa = htabNew(&faInterace, 8, 75);
	}
	fa = htabFindObjectEqualTo(allFa, &vfs);
	if(fa == NULL){
		fa = nfsFaNew(backend, vfs, cfgPath, numEntries);
		htabAddObject(allFa, fa);
	}
	nfsFaRetain(fa);
	return fa;
}

void	nfsFaRetain(nfsFileAdapter_t *fa)
{
	fa->refCount++;
}

void	nfsFaRelease(nfsFileAdapter_t *fa)
{
	if(--fa->refCount == 0){
		nfsFaFree(fa);
	}else if(fa->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

btinum	nfsFaRootHandle(nfsFileAdapter_t *fa, void *destHandle, btu32 mountId)
{
btinum	inum;

	luRootHandle(fa->lookupStrategy, destHandle, &inum, mountId);
	return inum;
}

/* ------------------------------------------------------------------------- */

#define	NFSERR_INVALID	14	/* this is an invalid nfs error which should be mapped by the client */

static int	translateError(btError *err)
{
    if(err->class != 0 && LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE)){
        logPrintf(LOGM_NFSTRACE, "nfs: cifs error class=%d code=%d\n", err->class, err->code);
    }
    switch(err->class){
    case ERRCLS_OK:						return NFS_OK;
    case ERRCLS_CIFS_DOS:
        switch(err->code){
		case ERR_CIFS_DOSbadfunc:		return NFSERR_INVALID;
		case ERR_CIFS_DOSbadfile:		return NFSERR_NOENT;
		case ERR_CIFS_DOSbadpath:		return NFSERR_NOENT;
		case ERR_CIFS_DOSnofids:		return NFSERR_IO;
		case ERR_CIFS_DOSnoaccess:		return NFSERR_ACCES;
		case ERR_CIFS_DOSbadfid:		return NFSERR_IO;
		case ERR_CIFS_DOSbadmcb:		return NFSERR_IO;
		case ERR_CIFS_DOSnomem:			return NFSERR_IO;
		case ERR_CIFS_DOSbadmem:		return NFSERR_IO;
		case ERR_CIFS_DOSbadenv:		return NFSERR_IO;
		case ERR_CIFS_DOSbadformat:		return NFSERR_IO;
		case ERR_CIFS_DOSbadaccess:		return NFSERR_ACCES;
		case ERR_CIFS_DOSbaddata:		return NFSERR_FBIG;
		case ERR_CIFS_DOSbaddrive:		return NFSERR_NXIO;
		case ERR_CIFS_DOSremcd:			return NFSERR_IO;
		case ERR_CIFS_DOSdiffdevice:	return NFSERR_NODEV;
		case ERR_CIFS_DOSnofiles:		return NFSERR_NOENT;
		case ERR_CIFS_DOSbadshare:		return NFSERR_ACCES;
		case ERR_CIFS_DOSlock:			return NFSERR_IO;
		case ERR_CIFS_DOSnotshared:		return NFSERR_ACCES;
		case ERR_CIFS_DOSfilexists:		return NFSERR_EXIST;
		case ERR_CIFS_DOSnotempty:		return NFSERR_NOTEMPTY;
            /* the following are heuristic values derived from smbfs */
		case 87:						return NFSERR_IO;
		case ERR_CIFS_DOSdestexists:	return NFSERR_EXIST;
        default:						return NFSERR_IO;
        }
        break;
    case ERRCLS_CIFS_SRV:
        switch(err->code){
		case ERR_CIFS_SRVerror:			return NFSERR_IO;
		case ERR_CIFS_SRVbadpw:			return NFSERR_ACCES;
		case ERR_CIFS_SRVaccess:		return NFSERR_ACCES;
		case ERR_CIFS_SRVinvnid:		return NFSERR_IO;
		case ERR_CIFS_SRVinvnetname:	return NFSERR_NXIO;
		case ERR_CIFS_SRVinvdevice:		return NFSERR_NXIO;
		case ERR_CIFS_SRVqfull:			return NFSERR_NOSPC;
		case ERR_CIFS_SRVqtoobig:		return NFSERR_NAMETOOLONG;
		case ERR_CIFS_SRVqeof:			return NFSERR_IO;
		case ERR_CIFS_SRVinvpfid:		return NFSERR_IO;
		case ERR_CIFS_SRVsmbcmd:		return NFSERR_IO;
		case ERR_CIFS_SRVsrverror:		return NFSERR_IO;
		case ERR_CIFS_SRVfilespecs:		return NFSERR_IO;
		case ERR_CIFS_SRVbadpermits:	return NFSERR_ACCES;
		case ERR_CIFS_SRVsetattrmode:	return NFSERR_ACCES;
		case ERR_CIFS_SRVpaused:		return NFSERR_ACCES;
		case ERR_CIFS_SRVmsgoff:		return NFSERR_ACCES;
		case ERR_CIFS_SRVnoroom:		return NFSERR_NOSPC;
		case ERR_CIFS_SRVrmuns:			return NFSERR_IO;
		case ERR_CIFS_SRVtimeout:		return NFSERR_IO;
		case ERR_CIFS_SRVnoresource:	return NFSERR_NOSPC;
		case ERR_CIFS_SRVtoomanyuids:	return NFSERR_ACCES;
		case ERR_CIFS_SRVbaduid:		return NFSERR_ACCES;
		case ERR_CIFS_SRVusempx:		return NFSERR_IO;
		case ERR_CIFS_SRVusestd:		return NFSERR_IO;
		case ERR_CIFS_SRVcontmpx:		return NFSERR_IO;
		case ERR_CIFS_SRVnosupport:		return NFSERR_IO;
        default:						return NFSERR_IO;
        }
        break;
    case ERRCLS_CIFS_HARD:
        switch(err->code){
		case ERR_CIFS_HRDnowrite:		return NFSERR_ROFS;
		case ERR_CIFS_HRDbadunit:		return NFSERR_NXIO;
		case ERR_CIFS_HRDnotready:		return NFSERR_IO;
		case ERR_CIFS_HRDbadcmd:		return NFSERR_IO;
		case ERR_CIFS_HRDdata:			return NFSERR_IO;
		case ERR_CIFS_HRDbadreq:		return NFSERR_IO;
		case ERR_CIFS_HRDseek:			return NFSERR_IO;
		case ERR_CIFS_HRDbadmedia:		return NFSERR_IO;
		case ERR_CIFS_HRDbadsector:		return NFSERR_IO;
		case ERR_CIFS_HRDnopaper:		return NFSERR_IO;
		case ERR_CIFS_HRDwrite:			return NFSERR_IO;
		case ERR_CIFS_HRDread:			return NFSERR_IO;
		case ERR_CIFS_HRDgeneral:		return NFSERR_IO;
		case ERR_CIFS_HRDbadshare:		return NFSERR_IO;
		case ERR_CIFS_HRDlock:			return NFSERR_IO;
		case ERR_CIFS_HRDwrongdisk:		return NFSERR_IO;
		case ERR_CIFS_HRDFCBUnavail:	return NFSERR_IO;
		case ERR_CIFS_HRDsharebufexc:	return NFSERR_IO;
        default:						return NFSERR_IO;
        }
        break;
    case ERRCLS_FRAMEWORK:
        switch(err->code){
		case ERR_FRWmodulenotfound:		return NFSERR_IO;
		case ERR_FRWmoduletypeerr:		return NFSERR_IO;
		case ERR_FRWalreadymounted:		return NFSERR_EXIST;
		case ERR_FRWipcparameters:		return NFSERR_IO;
		case ERR_FRWipcversion:			return NFSERR_IO;
		case ERR_FRWnomem:				return NFSERR_IO;
		case ERR_FRWnosuchobj:			return NFSERR_IO;
		case ERR_FRWnosuchmethod:		return NFSERR_IO;
		case ERR_FRWumountperm:			return NFSERR_PERM;
		case ERR_FRWclientcomm:			return NFSERR_IO;
		case ERR_FRWnosuchmnt:			return NFSERR_NOENT;
		case ERR_FRWmpaccessperm:		return NFSERR_ACCES;
		case ERR_FRWautomounted:		return NFSERR_ACCES;
		case ERR_FRWipcnotimpl:			return NFSERR_IO;
		case ERR_FRWusrabort:			return NFSERR_ACCES;
        default:						return NFSERR_IO;
        }
        break;
    case ERRCLS_CIFS_LOCAL:
        switch(err->code){
		case ERR_CIFS_LOCconnlost:		return -1;
		case ERR_CIFS_LOCtimeout:		return NFSERR_IO;
		case ERR_CIFS_LOCbogusdata:		return NFSERR_IO;
		case ERR_CIFS_LOCstale:			return NFSERR_STALE;
		case ERR_CIFS_LOCpwencrypt:		return NFSERR_ACCES;
		case ERR_CIFS_LOClookup:		return NFSERR_NOENT;
		case ERR_CIFS_LOClicclients:	return NFSERR_PERM;
		case ERR_CIFS_LOClicservers:	return NFSERR_PERM;
		case ERR_CIFS_LOClicense:		return NFSERR_IO;
		case ERR_CIFS_LOCsrvconn:		return NFSERR_IO;
		case ERR_CIFS_LOCwrongphrase:	return NFSERR_ACCES;
		case ERR_CIFS_LOCneedpassphrase: return NFSERR_ACCES;
		case ERR_CIFS_LOCnetbiosrefused: return NFSERR_ACCES;
		case ERR_CIFS_LOChavenossl:		return NFSERR_ACCES;
		case ERR_CIFS_LOCnosslserver:	return NFSERR_ACCES;
		case ERR_CIFS_LOCsslerror:		return NFSERR_IO;
		case ERR_CIFS_LOCnomountdir:	return NFSERR_IO;
		case ERR_CIFS_LOCnobrowser:		return NFSERR_ACCES;
		case ERR_CIFS_LOCotherlogin:	return NFSERR_ACCES;
        default:						return NFSERR_IO;
        }
        break;
    case ERRCLS_UNIX:
        switch(err->code){
		case 0:							return NFS_OK;
		case EPERM:						return NFSERR_PERM;
		case ENOENT:					return NFSERR_NOENT;
		case ENXIO:						return NFSERR_NXIO;
		case ETXTBSY:
		case EWOULDBLOCK:
		case EACCES:					return NFSERR_ACCES;
#if EEXIST != ENOTEMPTY
		case EEXIST:					return NFSERR_EXIST;
#endif
		case ENODEV:					return NFSERR_NODEV;
		case ENOTDIR:					return NFSERR_NOTDIR;
		case EISDIR:					return NFSERR_ISDIR;
		case E2BIG:
		case EFBIG:						return NFSERR_FBIG;
		case ENOSPC:					return NFSERR_NOSPC;
		case EROFS:						return NFSERR_ROFS;
		case ENAMETOOLONG:				return NFSERR_NAMETOOLONG;
		case ENOTEMPTY:					return NFSERR_NOTEMPTY;
#ifdef EDQUOT
		case EDQUOT:					return NFSERR_DQUOT;
#endif
#ifdef ETIMEDOUT
		case ETIMEDOUT:					return NFSERR_IO;
#endif
		default:						return NFSERR_ACCES;
        }
        break;
    default:					return NFSERR_ACCES;
    }
}

/* ------------------------------------------------------------------------- */

static void	nfsNotImplemented(void *context, void *arg)
{
nfsCtx_t	*this = context;
nfsstat		res = NFSERR_IO;

    PROC_BEGIN
    PROC_RETURN(&res);
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	translateReadAttributes(nfsFileAdapter_t *fa, struct fattr *nfsAttr, beFileAttr_t *fileAttr, btinum inum)
{
static btu32	microsecondDeviation = 0;

	nfsAttr->type = fileAttr->type;
	nfsAttr->mode = fileAttr->unixMode & 07777;
	switch(fileAttr->type){
		case BE_FILETYPE_REG:		nfsAttr->mode |= NFSMODE_REG;	break;
		case BE_FILETYPE_DIR:		nfsAttr->mode |= NFSMODE_DIR;	break;
		case BE_FILETYPE_BLK:		nfsAttr->mode |= NFSMODE_BLK;	break;
		case BE_FILETYPE_CHR:		nfsAttr->mode |= NFSMODE_CHR;	break;
		case BE_FILETYPE_LNK:		nfsAttr->mode |= NFSMODE_LNK;	break;
		case BE_FILETYPE_SOCK:		nfsAttr->mode |= NFSMODE_SOCK;	break;
		case BE_FILETYPE_FIFO:		nfsAttr->mode |= NFSMODE_FIFO;	break;
	}
	if(fileAttr->numLinks != 0){
		nfsAttr->nlink = fileAttr->numLinks;
	}else if(fileAttr->type == BE_FILETYPE_DIR){
		nfsAttr->nlink = fa->dirLinks;
	}else{
		nfsAttr->nlink = 1;
	}
	nfsAttr->uid = fileAttr->uid;
	nfsAttr->gid = fileAttr->gid;
	if(fileAttr->type == BE_FILETYPE_DIR){
		nfsAttr->size = fileAttr->size == 0 ? fa->dirSize : fileAttr->size;
	}else{
		nfsAttr->size = fileAttr->size;
	}
	nfsAttr->blocksize = NFS2CONST_BLOCKSIZE;
	if(fileAttr->type == BE_FILETYPE_BLK || fileAttr->type == BE_FILETYPE_CHR){
		nfsAttr->rdev = (fileAttr->devMajor << 8) | (fileAttr->devMinor & 0xff);
	}else{
		nfsAttr->rdev = NFS_FIFO_DEV;
	}
	/* alloc size for directories should be at least 1... */
	if(fileAttr->type == BE_FILETYPE_DIR){
		nfsAttr->blocks = fileAttr->allocatedSize == 0 ? 1 : (fileAttr->allocatedSize + NFS2CONST_BLOCKSIZE - 1) / NFS2CONST_BLOCKSIZE;
	}else{
		nfsAttr->blocks = (fileAttr->allocatedSize + NFS2CONST_BLOCKSIZE - 1) / NFS2CONST_BLOCKSIZE;
	}
	nfsAttr->fsid = 0;
	nfsAttr->fileid = inum;
	nfsAttr->atime.seconds = fileAttr->atime.seconds;
	nfsAttr->atime.useconds = fileAttr->atime.nseconds / 1000;
	nfsAttr->mtime.seconds = fileAttr->mtime.seconds;
	nfsAttr->mtime.useconds = fileAttr->mtime.nseconds / 1000;
	if(fa->changeMicrosecondFileTimes){
		nfsAttr->mtime.useconds += microsecondDeviation;
		microsecondDeviation = (microsecondDeviation + 1) % 10000;
	}
	nfsAttr->ctime.seconds = fileAttr->ctime.seconds;
	nfsAttr->ctime.useconds = fileAttr->ctime.nseconds / 1000;

#if defined SOLARIS2	/* Solaris 7 is not year 2038 safe.... */
	nfsAttr->atime.seconds &= 0x7fffffff;
	nfsAttr->mtime.seconds &= 0x7fffffff;
	nfsAttr->ctime.seconds &= 0x7fffffff;
#endif
}

static void	translateWriteAttributes(beFileSetAttr_t *fileAttr, struct sattr *nfsAttr)
{
	fileAttr->setUnixMode = (bti32)nfsAttr->mode != -1;
	fileAttr->setUid = (bti32)nfsAttr->uid != -1;
	fileAttr->setGid = (bti32)nfsAttr->gid != -1;
	fileAttr->setSize = (bti32)nfsAttr->size != -1;
	fileAttr->setAtime = (bti32)nfsAttr->atime.seconds != -1 || (bti32)nfsAttr->atime.useconds != -1;
	fileAttr->setMtime = (bti32)nfsAttr->mtime.seconds != -1 || (bti32)nfsAttr->mtime.useconds != -1;

	fileAttr->unixMode = nfsAttr->mode;
	fileAttr->uid = nfsAttr->uid;
	fileAttr->gid = nfsAttr->gid;
	fileAttr->size = nfsAttr->size;
	fileAttr->atime.seconds = nfsAttr->atime.seconds;
	fileAttr->atime.nseconds = nfsAttr->atime.useconds * 1000;
	fileAttr->mtime.seconds = nfsAttr->mtime.seconds;
	fileAttr->mtime.nseconds = nfsAttr->mtime.useconds * 1000;
}

/* ------------------------------------------------------------------------- */

static void	nfsLookup(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
btinum		inum;
diropres	res;
void		*parentFile;

    PROC_BEGIN
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->lookup.dir, this->user, 0, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->file = rval->vfsFile;
	this->doFileLookup = 1;
	if(this->arg->lookup.name[0] == '.'){	/* special treatment for '.' and '..' */
		if(this->arg->lookup.name[1] == 0){
			this->doFileLookup = 0;	/* it's just the current file */
		}else if(this->arg->lookup.name[1] == '.' && this->arg->lookup.name[2] == 0){
			if((parentFile = this->a->backend.fileParent(this->file)) != NULL){
				this->a->backend.fileRetain(parentFile);
				this->a->backend.fileRelease(this->file);
				this->file = parentFile;
			}
			this->doFileLookup = 0;
		}
	}
	convLocalAsciiToUni(this->arg->lookup.name, this->uniname, NAME_ARRAY_SIZE);
	if(this->doFileLookup){
		PROC_WAITFOR_X(2, this->a->backend.fileLookup(this->file, this->uniname, this->user, this));
		if((res.status = translateError(&frval->err)) != NFS_OK){
			PROC_RETURN(&res);
		}
		luCacheFile(this->a->lookupStrategy, frval->file);
		this->a->backend.fileRelease(this->file);
		this->file = frval->file;
	}
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
   if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	luHandleOfFile(this->a->lookupStrategy, &this->arg->lookup.dir, this->uniname, &res.diropres_u.diropres.file, &inum);
	translateReadAttributes(this->a, &res.diropres_u.diropres.attributes, frval->attr, inum);
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static blockpool_t	*readdirEntries = NULL;

static btbool	readdirIterationCallback(void *context, btuni *fileName, btu32 continuationIndex, btbool isDir)
{
nfsCtx_t		*this = context;
char			name[NAME_ARRAY_SIZE];
btsize			slen, nlen, len;
struct entry	*e;
btinum			inum;

    convUniToLocalAscii(fileName, name, NAME_ARRAY_SIZE);
    if(name[0] == '.' && (name[1] == 0 || (name[1] == '.' && name[2] == 0))){
        return 1;	/* skip '.' and '..' entries because we have already faked them */
    }
	slen = strlen(name);
    nlen = (slen/4 + 1) * 4;	/* actual size in XDR */
    len = nlen + sizeof(struct entry);	/* length of this entry in memory */
    if(len > this->restLen){
        return 0;						/* don't accept this entry, break */
    }
    this->restLen -= len;
    e = poolGetBlock(readdirEntries);
    *(this->entryTail) = e;
	luHandleOfFile(this->a->lookupStrategy, &this->arg->readdir.dir, fileName, NULL, &inum);
    e->fileid = inum;
    e->name = memAlloc(nlen);
    memcpy(e->name, name, slen + 1);
    *(btu32 *)&(e->cookie) = htonl(continuationIndex + 2);	/* add an offset of 2 for faked '.' and '..' entries */
    this->entryTail = &(e->nextentry);
    return 1;
}

static void	nfsReaddir(void *context, void *arg)
{
nfsCtx_t		*this = context;
readdirres		res;
luRes_t			*rval = arg;
beFileRes_t		*frval = arg;
btu32			startIndex;
struct entry	*e;
btinum			inum;
btuni			fname[3];

	PROC_BEGIN
    if(readdirEntries == NULL)
        readdirEntries = poolNew(sizeof(struct entry), sizeof(struct entry));
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->lookup.dir, this->user, 0, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
    startIndex = ntohl(*(btu32 *)&(this->arg->readdir.cookie));
    this->restLen = this->arg->readdir.count;
    this->entryTail = &(this->entryHead);
/* fake '.' and '..' entries because they _must_ be correct and we can't rely on the server */
    if(startIndex == 0){	/* fake '.' entry */
        e = poolGetBlock(readdirEntries);
        *(this->entryTail) = e;
		fname[0] = '.';
		fname[1] = 0;
		luHandleOfFile(this->a->lookupStrategy, &this->arg->readdir.dir, fname, NULL, &inum);
        e->fileid = inum;
        e->name = memAlloc(2);
        strcpy(e->name, ".");
        *(btu32 *)&(e->cookie) = htonl(1);
        this->entryTail = &(e->nextentry);
        startIndex++;
        this->restLen -= sizeof(struct entry) + 4;	/* 4 is the rounded string length */
    }
    if(startIndex == 1){	/* fake '..' entry */
        e = poolGetBlock(readdirEntries);
        *(this->entryTail) = e;
		fname[0] = fname[1] = '.';
		fname[2] = 0;
		luHandleOfFile(this->a->lookupStrategy, &this->arg->readdir.dir, fname, NULL, &inum);
        e->fileid = inum;
        e->name = memAlloc(3);
        strcpy(e->name, "..");
        *(btu32 *)&(e->cookie) = htonl(2);
        this->entryTail = &(e->nextentry);
        startIndex++;
        this->restLen -= sizeof(struct entry) + 4;	/* 4 is the rounded string length */
    }
    startIndex -= 2;
    PROC_WAITFOR_X(2, this->a->backend.fileReaddir(this->file, this->user, startIndex, readdirIterationCallback, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    res.readdirres_u.reply.entries = this->entryHead;
    res.readdirres_u.reply.eof = frval->eof;
    PROC_RETURN(&res);
    PROC_END
}

static inline void	freeEntries(struct entry **headp)
{
struct entry	*p, *next;
    
    if(*headp != NULL){
        for(p=*headp;p!=NULL;p=next){
			next = p->nextentry;
            memFree(p->name);
            poolReturnBlock(readdirEntries, p);
        }
        *headp = NULL;
    }
}

/* ------------------------------------------------------------------------- */

static void	nfsStatfs(void *context, void *arg)
{
nfsCtx_t	*this = context;
beVfsRes_t	*rval = arg;
statfsres	res;

    PROC_BEGIN
    PROC_WAITFOR_X(1, this->a->backend.vfsStatfs(this->a->vfs, this->user, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    res.statfsres_u.reply.tsize = 8192;	/* optimum transfer size */
    res.statfsres_u.reply.bsize = rval->bytesPerAU;
    res.statfsres_u.reply.blocks = rval->totalAU;
    res.statfsres_u.reply.bfree = rval->availableAU;
    res.statfsres_u.reply.bavail = res.statfsres_u.reply.bfree;
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsGetattr(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
attrstat	res;
btinum		inum;
btuni		fname[2];

    PROC_BEGIN
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->getattr, this->user, 0, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
	PROC_WAITFOR_X(2, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	fname[0] = '.';
	fname[1] = 0;
	luHandleOfFile(this->a->lookupStrategy, &this->arg->getattr, fname, NULL, &inum);
	translateReadAttributes(this->a, &res.attrstat_u.attributes, frval->attr, inum);
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsSetattr(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
attrstat	res;
btinum		inum;
btuni		fname[2];

	PROC_BEGIN
	if(this->a->backend.fileSetAttr == NULL){
		res.status = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->setattr.file, this->user, 0, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
	translateWriteAttributes(&this->setAttr, &this->arg->setattr.attributes);
    PROC_WAITFOR_X(2, this->a->backend.fileSetAttr(this->file, this->user, &this->setAttr, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	fname[0] = '.';
	fname[1] = 0;
	luHandleOfFile(this->a->lookupStrategy, &this->arg->setattr.file, fname, NULL, &inum);
	translateReadAttributes(this->a, &res.attrstat_u.attributes, frval->attr, inum);
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsCreateAny(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
diropres	res;
btinum		inum;

	PROC_BEGIN
	if(this->a->backend.fileCreate == NULL){
		res.status = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->create.where.dir, this->user, 1, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->create.where.name, this->uniname, NAME_ARRAY_SIZE);
	translateWriteAttributes(&this->setAttr, &this->arg->create.attributes);
    PROC_WAITFOR_X(2, this->a->backend.fileCreate(this->file, this->uniname, this->isDir, this->user, &this->setAttr, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->a->backend.fileRelease(this->file);
    this->file = frval->file;
	luCacheFile(this->a->lookupStrategy, frval->file);
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	luHandleOfFile(this->a->lookupStrategy, &this->arg->create.where.dir, this->uniname, &res.diropres_u.diropres.file, &inum);
	translateReadAttributes(this->a, &res.diropres_u.diropres.attributes, frval->attr, inum);
    PROC_RETURN(&res);
    PROC_END
}
/* ------------------------------------------------------------------------- */

static void	nfsCreate(void *context, void *arg)
{
nfsCtx_t	*this = context;

    this->isDir = 0;
    this->proc.process = nfsCreateAny;	/* overload process */
    nfsCreateAny(this, arg);
}

/* ------------------------------------------------------------------------- */

static void	nfsMkdir(void *context, void *arg)
{
nfsCtx_t	*this = context;

    this->isDir = 1;
    this->proc.process = nfsCreateAny;	/* overload process */
    nfsCreateAny(this, arg);
}

/* ------------------------------------------------------------------------- */

static void	nfsRemoveAny(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
nfsstat		res;

	PROC_BEGIN
	if(this->a->backend.fileRemove == NULL){
		res = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->remove.dir, this->user, 0, this));
    if((res = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->remove.name, this->uniname, NAME_ARRAY_SIZE);
    PROC_WAITFOR_X(2, this->a->backend.fileRemove(this->file, this->uniname, this->isDir, this->user, this));
	if((res = translateError(&frval->err)) == NFS_OK){
		luMarkStale(this->a->lookupStrategy, &this->arg->remove.dir, this->uniname);
	}
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsRemove(void *context, void *arg)
{
nfsCtx_t	*this = context;

    this->isDir = 0;
    this->proc.process = nfsRemoveAny;	/* overload process */
    nfsRemoveAny(this, arg);
}

/* ------------------------------------------------------------------------- */

static void	nfsRmdir(void *context, void *arg)
{
nfsCtx_t	*this = context;

    this->isDir = 1;
    this->proc.process = nfsRemoveAny;	/* overload process */
    nfsRemoveAny(this, arg);
}

/* ------------------------------------------------------------------------- */

static void	nfsRename(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
nfsstat		res;

	PROC_BEGIN
	if(this->a->backend.fileRename == NULL){
		res = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->rename.from.dir, this->user, 0, this));
    if((res = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
	PROC_WAITFOR_X(2, luLookupFile(this->a->lookupStrategy, &this->arg->rename.to.dir, this->user, 0, this));
    if((res = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->otherFile = rval->vfsFile;
	convLocalAsciiToUni(this->arg->rename.from.name, this->uniname, NAME_ARRAY_SIZE);
    convLocalAsciiToUni(this->arg->rename.to.name, this->unipath, PATH_ARRAY_SIZE);
    PROC_WAITFOR_X(3, this->a->backend.fileRename(this->file, this->uniname, this->otherFile, this->unipath, this->user, this));
	if((res = translateError(&frval->err)) == NFS_OK){
		luRename(this->a->lookupStrategy, &this->arg->rename.from.dir, this->uniname, &this->arg->rename.to.dir, this->unipath);
	}
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsRead(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
btuni		fname[2];
readres		res;
btinum		inum;

	PROC_BEGIN
	if(this->a->backend.fileRW == NULL){
		res.status = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->read.file, this->user, 1, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->file = rval->vfsFile;
    this->data = memAlloc(this->arg->read.count);	/* will be freed in context free */
    PROC_WAITFOR_X(2, this->a->backend.fileRW(this->file, BE_FILE_OP_READ, 1, this->data, this->arg->read.offset, this->arg->read.count, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->restLen = frval->dataLen;
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    res.readres_u.reply.data.data_len = this->restLen;
    res.readres_u.reply.data.data_val = this->data;
	fname[0] = '.';
	fname[1] = 0;
	luHandleOfFile(this->a->lookupStrategy, &this->arg->read.file, fname, NULL, &inum);
	translateReadAttributes(this->a, &res.readres_u.reply.attributes, frval->attr, inum);
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsWrite(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
btuni		fname[2];
attrstat	res;
btinum		inum;

    PROC_BEGIN
	if(this->a->backend.fileRW == NULL){
		res.status = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->write.file, this->user, 1, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->file = rval->vfsFile;
    PROC_WAITFOR_X(2, this->a->backend.fileRW(this->file, BE_FILE_OP_WRITE, 1, this->arg->write.data.data_val, this->arg->write.offset, this->arg->write.data.data_len, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    PROC_WAITFOR_X(3, this->a->backend.fileGetAttr(this->file, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	fname[0] = '.';
	fname[1] = 0;
	luHandleOfFile(this->a->lookupStrategy, &this->arg->read.file, fname, NULL, &inum);
 	translateReadAttributes(this->a, &res.attrstat_u.attributes, frval->attr, inum);
	PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsLink(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
nfsstat		res;

	PROC_BEGIN
	if(this->a->backend.fileHardLink == NULL){
		res = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->link.from, this->user, 0, this));
    if((res = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->file = rval->vfsFile;
	PROC_WAITFOR_X(2, luLookupFile(this->a->lookupStrategy, &this->arg->link.to.dir, this->user, 0, this));
    if((res = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
	this->otherFile = rval->vfsFile;
    convLocalAsciiToUni(this->arg->link.to.name, this->uniname, NAME_ARRAY_SIZE);
    PROC_WAITFOR_X(3, this->a->backend.fileHardLink(this->file, this->otherFile, this->uniname, this->user, this));
    if((res = translateError(&frval->err)) == NFS_OK){
		this->a->backend.fileRelease(frval->file);
	}
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsSymlink(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
nfsstat		res;

	PROC_BEGIN
	if(this->a->backend.fileSymlink == NULL){
		res = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->symlink.from.dir, this->user, 0, this));
    if((res = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
    convLocalAsciiToUni(this->arg->symlink.from.name, this->uniname, NAME_ARRAY_SIZE);
    convLocalAsciiToUni(this->arg->symlink.to, this->unipath, PATH_ARRAY_SIZE);
	translateWriteAttributes(&this->setAttr, &this->arg->symlink.attributes);
    PROC_WAITFOR_X(2, this->a->backend.fileSymlink(this->file, this->uniname, this->unipath, this->user, &this->setAttr, this));
    if((res = translateError(&frval->err)) == NFS_OK){
		luCacheFile(this->a->lookupStrategy, frval->file);
		this->a->backend.fileRelease(frval->file);
	}
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	nfsReadlink(void *context, void *arg)
{
nfsCtx_t	*this = context;
luRes_t		*rval = arg;
beFileRes_t	*frval = arg;
char		path[PATH_ARRAY_SIZE];
readlinkres	res;

	PROC_BEGIN
	if(this->a->backend.fileReadlink == NULL){
		res.status = NFSERR_ACCES;
		PROC_RETURN(&res);
	}
	PROC_WAITFOR_X(1, luLookupFile(this->a->lookupStrategy, &this->arg->readlink, this->user, 0, this));
    if((res.status = translateError(&rval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    this->file = rval->vfsFile;
    PROC_WAITFOR_X(2, this->a->backend.fileReadlink(this->file, this->user, this));
    if((res.status = translateError(&frval->err)) != NFS_OK){
        PROC_RETURN(&res);
    }
    convUniToLocalAscii(frval->path, path, PATH_ARRAY_SIZE);
    res.readlinkres_u.data = path;
    PROC_RETURN(&res);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static btcallback	nfsFuncs[] = {
	nfsNotImplemented,	/* NULL */
	nfsGetattr,
	nfsSetattr,
	nfsNotImplemented,	/* ROOT */
	nfsLookup,
	nfsReadlink,
	nfsRead,
	nfsNotImplemented,	/* WRITECACHE */
	nfsWrite,
	nfsCreate,
	nfsRemove,
	nfsRename,
	nfsLink,
	nfsSymlink,
	nfsMkdir,
	nfsRmdir,
	nfsReaddir,
	nfsStatfs
};

/* ------------------------------------------------------------------------- */

static void	nfsFaExecuteRequestProc(void *ctx, void *arg)
{
nfsCtx_t	*this = ctx;
beVfsRes_t	*rval = arg;
nfsstat		res;

	PROC_BEGIN;
	PROC_WAITFOR_X(1, this->a->backend.vfsConnect(this->a->vfs, this->user, this));
	if((res = translateError(&rval->err)) != NFS_OK){
		PROC_RETURN(&res);
	}
    this->proc.currentState = 0;	/* overload current process */
	this->proc.process = nfsFuncs[this->functionIndex];
	PROC_START(this, NULL);
	return;	/* as a consequence of the overload, terminate without callback */
    PROC_END
}


void	nfsFaExecuteRequest(nfsFileAdapter_t *fa, btint functionIndex, void *arg, void *user, void *caller)
{
nfsCtx_t	*ctx;

    PROC_INIT(ctx, nfsFaExecuteRequestProc, caller, nfsPool, nfsCtxFree);
	ctx->a = fa;
	ctx->arg = arg;
	ctx->user = user;
	ctx->functionIndex = functionIndex;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

