/*
 * Name: nfs3xdr.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 2001-10-25
 * Tabsize: 4
 * Copyright: (c) 2001 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfs3xdr.c,v 1.1 2001/11/06 09:50:36 cs Exp $
 */

#include "stdheaders.h"
#include "exception.h"
#include "logging.h"
#include "memalloc.h"
#include "sunrpc.h"
#include "nfs3xdr.h"


/* ------------------------------------------------------------------------- */
/* ------------------------- NFS3 underlying types ------------------------- */
/* ------------------------------------------------------------------------- */

static void	nfs3EncodeInt64(struct srpcMsg *msg, btu64 val)
{
	srpcEncodeInt(msg, (btu32)(val>>32));
	srpcEncodeInt(msg, (btu32)val);
}

static void	nfs3DecodeInt64(struct srpcMsg *msg, btu64 *val)
{
btu32	msw;

	msw = srpcDecodeInt(msg);
	*val = srpcDecodeInt(msg);
	*val |= (btu64)msw << 32;
}

static void	nfs3EncodeSpecdata(struct srpcMsg *msg, struct specdata3 *val)
{
	srpcEncodeInt(msg, val->specdata1);
	srpcEncodeInt(msg, val->specdata2);
}

static void	nfs3DecodeSpecdata(struct srpcMsg *msg, struct specdata3 *val)
{
	val->specdata1 = srpcDecodeInt(msg);
	val->specdata2 = srpcDecodeInt(msg);
}

static void	nfs3EncodeNfstime(struct srpcMsg *msg, struct nfstime3 *val)
{
	srpcEncodeInt(msg, val->seconds);
	srpcEncodeInt(msg, val->nseconds);
}

static void	nfs3DecodeNfstime(struct srpcMsg *msg, struct nfstime3 *val)
{
	val->seconds = srpcDecodeInt(msg);
	val->nseconds = srpcDecodeInt(msg);
}

static void	nfs3EncodeFattr(struct srpcMsg *msg, struct fattr3 *val)
{
	srpcEncodeInt(msg, val->type);
	srpcEncodeInt(msg, val->mode);
	srpcEncodeInt(msg, val->nlink);
	srpcEncodeInt(msg, val->uid);
	srpcEncodeInt(msg, val->gid);
	nfs3EncodeInt64(msg, val->size);
	nfs3EncodeInt64(msg, val->used);
	nfs3EncodeSpecdata(msg, &val->rdev);
	nfs3EncodeInt64(msg, val->fsid);
	nfs3EncodeInt64(msg, val->fileid);
	nfs3EncodeNfstime(msg, &val->atime);
	nfs3EncodeNfstime(msg, &val->mtime);
	nfs3EncodeNfstime(msg, &val->ctime);
}

static void	nfs3DecodeSetBtu32(struct srpcMsg *msg, struct set_btu32 *val)
{
	val->set = srpcDecodeInt(msg);
	if(val->set)
		val->val = srpcDecodeInt(msg);
}

static void	nfs3DecodeSetBtu64(struct srpcMsg *msg, struct set_btu64 *val)
{
	val->set = srpcDecodeInt(msg);
	if(val->set)
		nfs3DecodeInt64(msg, &val->val);
}

static void	nfs3DecodeSetNfstime(struct srpcMsg *msg, struct set_nfstime3 *val)
{
	val->mode = srpcDecodeInt(msg);
	if(val->mode == NFS3_TIMESET_SET_TO_CLIENT_TIME){
		nfs3DecodeNfstime(msg, &val->time);
	}
}

static void	nfs3DecodeSattr(struct srpcMsg *msg, struct sattr3 *val)
{
	nfs3DecodeSetBtu32(msg, &val->mode);
	nfs3DecodeSetBtu32(msg, &val->uid);
	nfs3DecodeSetBtu32(msg, &val->gid);
	nfs3DecodeSetBtu64(msg, &val->size);
	nfs3DecodeSetNfstime(msg, &val->atime);
	nfs3DecodeSetNfstime(msg, &val->mtime);
}

static void	nfs3DecodeSetattrguard(struct srpcMsg *msg, struct sattrguard3 *val)
{
	val->check = srpcDecodeInt(msg);
	if(val->check)
		nfs3DecodeNfstime(msg, &val->obj_ctime);
}

static void	nfs3EncodeWccAttr(struct srpcMsg *msg, struct wcc_attr *val)
{
	nfs3EncodeInt64(msg, val->size);
	nfs3EncodeNfstime(msg, &val->mtime);
	nfs3EncodeNfstime(msg, &val->ctime);
}

static void	nfs3EncodePreOpAttr(struct srpcMsg *msg, struct pre_op_attr *val)
{
	srpcEncodeInt(msg, val->attributes_follow);
	if(val->attributes_follow)
		nfs3EncodeWccAttr(msg, &val->attributes);
}

static void	nfs3EncodePostOpAttr(struct srpcMsg *msg, struct post_op_attr *val)
{
	srpcEncodeInt(msg, val->attributes_follow);
	if(val->attributes_follow)
		nfs3EncodeFattr(msg, &val->attributes);
}

static void	nfs3EncodeWccData(struct srpcMsg *msg, struct wcc_data *val)
{
	nfs3EncodePreOpAttr(msg, &val->before);
	nfs3EncodePostOpAttr(msg, &val->after);
}

static void	nfs3EncodePostOpFh(struct srpcMsg *msg, struct post_op_fh3 *val)
{
	srpcEncodeInt(msg, val->handle_follows);
	if(val->handle_follows)
		nfs3EncodeFh(msg, &val->handle);
}

static void	nfs3EncodeEntries(struct srpcMsg *msg, struct entry3 *val)
{
	while(val != NULL){
		srpcEncodeInt(msg, 1);	/* pointer is not null */
		nfs3EncodeInt64(msg, val->fileid);
		srpcEncodeString(msg, val->name);
		nfs3EncodeInt64(msg, val->cookie);
		val = val->nextentry;
	}
	srpcEncodeInt(msg, 0);	/* pointer is NULL now */
}

static void	nfs3EncodeEntriesplus(struct srpcMsg *msg, struct entryplus3 *val)
{
	while(val != NULL){
		srpcEncodeInt(msg, 1);	/* pointer is not null */
		nfs3EncodeInt64(msg, val->fileid);
		srpcEncodeString(msg, val->name);
		nfs3EncodeInt64(msg, val->cookie);
		nfs3EncodePostOpAttr(msg, &val->name_attributes);
		nfs3EncodePostOpFh(msg, &val->name_handle);
		val = val->nextentry;
	}
	srpcEncodeInt(msg, 0);	/* pointer is NULL now */
}

/* ------------------------------------------------------------------------- */
/* ----------------------------- NFS3 exports ------------------------------ */
/* ------------------------------------------------------------------------- */

void	nfs3DecodeFh(struct srpcMsg *msg, struct nfs_fh3 *val)
{
	val->data = srpcDecodeVarBlock(msg, (btsize *)&val->len);
}

void	nfs3EncodeFh(struct srpcMsg *msg, struct nfs_fh3 *val)
{
	srpcEncodeVarBlock(msg, val->data, val->len);
}

void	nfs3DecodeDiropargs(struct srpcMsg *msg, struct diropargs3 *val)
{
	nfs3DecodeFh(msg, &val->dir);
	val->name = srpcDecodeString(msg);
}

void	nfs3EncodeGetattrRes(struct srpcMsg *msg, struct GETATTR3res *val)
{
	srpcEncodeInt(msg, val->status);
	if(val->status == NFS3_OK){
		nfs3EncodeFattr(msg, &val->attributes);
	}
}

void	nfs3DecodeSetattrArgs(struct srpcMsg *msg, struct SETATTR3args *val)
{
	nfs3DecodeFh(msg, &val->object);
	nfs3DecodeSattr(msg, &val->new_attributes);
	nfs3DecodeSetattrguard(msg, &val->guard);
}

void	nfs3EncodeSetattrRes(struct srpcMsg *msg, struct SETATTR3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodeWccData(msg, &val->obj_wcc);
}

void	nfs3EncodeLookupRes(struct srpcMsg *msg, struct LOOKUP3res *val)
{
	srpcEncodeInt(msg, val->status);
	if(val->status == NFS3_OK){
		nfs3EncodeFh(msg, &val->object);
		nfs3EncodePostOpAttr(msg, &val->obj_attributes);
		nfs3EncodePostOpAttr(msg, &val->dir_attributes);
	}else{
		nfs3EncodePostOpAttr(msg, &val->dir_attributes);
	}
}

void	nfs3DecodeAccessArgs(struct srpcMsg *msg, struct ACCESS3args *val)
{
	nfs3DecodeFh(msg, &val->object);
	val->access = srpcDecodeInt(msg);
}

void	nfs3EncodeAccessRes(struct srpcMsg *msg, struct ACCESS3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->obj_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeInt(msg, val->access);
	}
}

void	nfs3EncodeReadlinkRes(struct srpcMsg *msg, struct READLINK3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->symlink_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeString(msg, val->data);
	}
}

void	nfs3DecodeReadArgs(struct srpcMsg *msg, struct READ3args *val)
{
	nfs3DecodeFh(msg, &val->file);
	nfs3DecodeInt64(msg, &val->offset);
	val->count = srpcDecodeInt(msg);
}

void	nfs3EncodeReadRes(struct srpcMsg *msg,struct READ3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->file_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeInt(msg, val->count);
		srpcEncodeInt(msg, val->eof);
		srpcEncodeVarBlock(msg, val->data.data, val->data.len);
	}
}

void	nfs3DecodeWriteArgs(struct srpcMsg *msg, struct WRITE3args *val)
{
	nfs3DecodeFh(msg, &val->file);
	nfs3DecodeInt64(msg, &val->offset);
	val->count = srpcDecodeInt(msg);
	val->stable = srpcDecodeInt(msg);
	val->data.data = srpcDecodeVarBlock(msg, (btsize *)&val->data.len);
}

void	nfs3EncodeWriteRes(struct srpcMsg *msg, struct WRITE3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodeWccData(msg, &val->file_wcc);
	if(val->status == NFS3_OK){
		srpcEncodeInt(msg, val->count);
		srpcEncodeInt(msg, val->committed);
		srpcEncodeFixedBlock(msg, val->verf, NFS3_WRITEVERFSIZE);
	}
}

void	nfs3DecodeCreateArgs(struct srpcMsg *msg, struct CREATE3args *val)
{
	nfs3DecodeDiropargs(msg, &val->where);
	val->mode = srpcDecodeInt(msg);
	if(val->mode == NFS3_CREATE_GUARDED || val->mode == NFS3_CREATE_UNCHECKED){
		nfs3DecodeSattr(msg, &val->u.guarded_or_unchecked.obj_attributes);
	}else if(val->mode == NFS3_CREATE_EXCLUSIVE){
		memcpy(val->u.exclusive.verf, srpcDecodeFixedBlock(msg, NFS3_CREATEVERFSIZE), NFS3_CREATEVERFSIZE);
	}
}

void	nfs3EncodeCreateRes(struct srpcMsg *msg, struct CREATE3res *val)
{
	srpcEncodeInt(msg, val->status);
	if(val->status == NFS3_OK){
		nfs3EncodePostOpFh(msg, &val->obj);
		nfs3EncodePostOpAttr(msg, &val->obj_attributes);
		nfs3EncodeWccData(msg, &val->dir_wcc);
	}else{
		nfs3EncodeWccData(msg, &val->dir_wcc);
	}
}

void	nfs3DecodeMkdirArgs(struct srpcMsg *msg, struct MKDIR3args *val)
{
	nfs3DecodeDiropargs(msg, &val->where);
	nfs3DecodeSattr(msg, &val->attributes);
}

void	nfs3DecodeSymlinkArgs(struct srpcMsg *msg, struct SYMLINK3args *val)
{
	nfs3DecodeDiropargs(msg, &val->where);
	nfs3DecodeSattr(msg, &val->symlink_attributes);
	val->symlink_data = srpcDecodeString(msg);
}

void	nfs3DecodeMknodArgs(struct srpcMsg *msg, struct MKNOD3args *val)
{
	nfs3DecodeDiropargs(msg, &val->where);
	val->type = srpcDecodeInt(msg);
	if(val->type == NFS3_TYPE_CHR || val->type == NFS3_TYPE_BLK){
		nfs3DecodeSattr(msg, &val->u.char_or_blk.dev_attributes);
		nfs3DecodeSpecdata(msg, &val->u.char_or_blk.spec);
	}else if(val->type == NFS3_TYPE_SOCK || val->type == NFS3_TYPE_FIFO){
		nfs3DecodeSattr(msg, &val->u.sock_or_fifo.pipe_attributes);
	}
}

void	nfs3EncodeRemoveRes(struct srpcMsg *msg, struct REMOVE3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodeWccData(msg, &val->dir_wcc);
}

void	nfs3DecodeRenameArgs(struct srpcMsg *msg, struct RENAME3args *val)
{
	nfs3DecodeDiropargs(msg, &val->from);
	nfs3DecodeDiropargs(msg, &val->to);
}

void	nfs3EncodeRenameRes(struct srpcMsg *msg, struct RENAME3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodeWccData(msg, &val->fromdir_wcc);
	nfs3EncodeWccData(msg, &val->todir_wcc);
}

void	nfs3DecodeLinkArgs(struct srpcMsg *msg, struct LINK3args *val)
{
	nfs3DecodeFh(msg, &val->file);
	nfs3DecodeDiropargs(msg, &val->link);
}

void	nfs3EncodeLinkRes(struct srpcMsg *msg, struct LINK3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->file_attributes);
	nfs3EncodeWccData(msg, &val->linkdir_wcc);
}

void	nfs3DecodeReaddirArgs(struct srpcMsg *msg, struct READDIR3args *val)
{
	nfs3DecodeFh(msg, &val->dir);
	nfs3DecodeInt64(msg, &val->cookie);
	memcpy(val->cookieverf, srpcDecodeFixedBlock(msg, NFS3_COOKIEVERFSIZE), NFS3_COOKIEVERFSIZE);
	val->count = srpcDecodeInt(msg);
}

void	nfs3EncodeReaddirRes(struct srpcMsg *msg, struct READDIR3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->dir_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeFixedBlock(msg, val->cookieverf, NFS3_COOKIEVERFSIZE);
		nfs3EncodeEntries(msg, val->entries);
		srpcEncodeInt(msg, val->eof);
	}
}

void	nfs3DecodeReaddirplusArgs(struct srpcMsg *msg, struct READDIRPLUS3args *val)
{
	nfs3DecodeFh(msg, &val->dir);
	nfs3DecodeInt64(msg, &val->cookie);
	memcpy(val->cookieverf, srpcDecodeFixedBlock(msg, NFS3_COOKIEVERFSIZE), NFS3_COOKIEVERFSIZE);
	val->dircount = srpcDecodeInt(msg);
	val->maxcount = srpcDecodeInt(msg);
}

void	nfs3EncodeReaddirplusRes(struct srpcMsg *msg, struct READDIRPLUS3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->dir_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeFixedBlock(msg, val->cookieverf, NFS3_COOKIEVERFSIZE);
		nfs3EncodeEntriesplus(msg, val->entries);
		srpcEncodeInt(msg, val->eof);
	}
}

void	nfs3EncodeFsstatRes(struct srpcMsg *msg, struct FSSTAT3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->obj_attributes);
	if(val->status == NFS3_OK){
		nfs3EncodeInt64(msg, val->tbytes);
		nfs3EncodeInt64(msg, val->fbytes);
		nfs3EncodeInt64(msg, val->abytes);
		nfs3EncodeInt64(msg, val->tfiles);
		nfs3EncodeInt64(msg, val->ffiles);
		nfs3EncodeInt64(msg, val->afiles);
		srpcEncodeInt(msg, val->invarsec);
	}
}

void	nfs3EncodeFsinfoRes(struct srpcMsg *msg, struct FSINFO3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->obj_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeInt(msg, val->rtmax);
		srpcEncodeInt(msg, val->rtpref);
		srpcEncodeInt(msg, val->rtmult);
		srpcEncodeInt(msg, val->wtmax);
		srpcEncodeInt(msg, val->wtpref);
		srpcEncodeInt(msg, val->wtmult);
		srpcEncodeInt(msg, val->dtpref);
		nfs3EncodeInt64(msg, val->maxfilesize);
		nfs3EncodeNfstime(msg, &val->time_delta);
		srpcEncodeInt(msg, val->properties);
	}
}

void	nfs3EncodePathconfRes(struct srpcMsg *msg, struct PATHCONF3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodePostOpAttr(msg, &val->obj_attributes);
	if(val->status == NFS3_OK){
		srpcEncodeInt(msg, val->linkmax);
		srpcEncodeInt(msg, val->name_max);
		srpcEncodeInt(msg, val->no_trunc);
		srpcEncodeInt(msg, val->chown_restricted);
		srpcEncodeInt(msg, val->case_insensitive);
		srpcEncodeInt(msg, val->case_preserving);
	}
}

void	nfs3DecodeCommitArgs(struct srpcMsg *msg, struct COMMIT3args *val)
{
	nfs3DecodeFh(msg, &val->file);
	nfs3DecodeInt64(msg, &val->offset);
	val->count = srpcDecodeInt(msg);
}

void	nfs3EncodeCommitRes(struct srpcMsg *msg, struct COMMIT3res *val)
{
	srpcEncodeInt(msg, val->status);
	nfs3EncodeWccData(msg, &val->file_wcc);
	if(val->status == NFS3_OK){
		srpcEncodeFixedBlock(msg, val->verf, NFS3_WRITEVERFSIZE);
	}
}

/* ------------------------------------------------------------------------- */
