/*
 * Name: blockalloc.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-29
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: blockalloc.c,v 1.2 1999/09/10 23:04:12 cs Exp $
 */


#include "stdheaders.h"
#include "blockalloc.h"
#include "exception.h"
#include "memalloc.h"
#include "system.h"
#include "array.h"
#include "logging.h"
#include "cfgparse.h"
#include "mainmodule.h"

/* ------------------------------------------------------------------------- */

typedef struct smartArray{
	pagePool_t	*pagePool;
	btint		numIndirections;
	void		**basePage;
}smartArray_t;

typedef struct bitMap{
	struct bitMap	*nextLevel;	/* [256] */
	btu32			bits[8];
}bitMap_t;

typedef struct bitArray{
	btint		numLevels;	/* 1 levels = bits in root are leafs */
	bitMap_t	root;
}bitArray_t;


/* A pagelist is a list of pages which are all divided into blocks of the
 * same size (elementSize).
 */
typedef struct pageList{
	bti32			blockSize;
	bti32			blocksPerPage;
	smartArray_t	pages;			/* actual storage area */
	bitArray_t		freeBlocks;		/* which blocks are free */
}pageList_t;

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	dummyInterface = {
	NULL, NULL, NULL, NULL
};

pagePool_t	*pagePoolNew(void)
{
pagePool_t		*volatile pp = NULL;
btint			numEntries;
cfgProperty_t	*path[8];
static btint	numPages;
static btu32	numPagesChg = 0;

EXC_TRY
    pp = memCalloc(sizeof(pagePool_t));
	pp->usedBases = arrNew(&dummyInterface, 16);
	if(CFG_DID_CHANGE(numPagesChg)){
		numEntries = mainModCfgPath(path, ELEMENTS(path));
		numPages = cfgIntValue(cfgQueryFromArray("pagePoolInitialSize", numEntries, path), 128);
	}
	pp->numPages = numPages;
	pp->nextPage = pp->numPages;	/* currently no pages available */
	pp->memBase = NULL;	/* alloc on demand */
EXC_CATCH
	pagePoolFree(pp);
	EXC_RETHROW();
EXC_ENDTRY
	return pp;
}

/* ------------------------------------------------------------------------- */

void	pagePoolFree(pagePool_t *pp)
{
btint	i, l;

	if(pp != NULL){
		if(pp->memBase != NULL){
			sysVfree(pp->memBase);
			pp->memBase = NULL;
		}
		if(pp->usedBases != NULL){
			l = arrObjectCount(pp->usedBases);
			for(i=0;i<l;i++){
				sysVfree(arrObjectAtIndex(pp->usedBases, i));
			}
			arrFree(pp->usedBases);
			pp->usedBases = NULL;
		}
		memFree(pp);
	}
}

/* ------------------------------------------------------------------------- */

inline void	*pagePoolGetPages(pagePool_t *pp, bti32 numPages)
{
void	*pageBase, *newBase;

	if(pp->nextPage + numPages > pp->numPages){	/* current base used up */
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "blockalloc: pagePoolGetPages: allocating new vm pool due to %d page request (lost=%d)\n", numPages, pp->numPages - pp->nextPage);
		newBase = sysValloc(pp->numPages * BLOCKALLOC_PAGESIZE);
		if(newBase == NULL){	/* virtual memory exhausted! */
			EXC_THROW_MSG(memExcNomem, "Out of VIRTUAL memory");
		}
		if(pp->memBase != NULL)
			arrAddObject(pp->usedBases, pp->memBase);
		pp->memBase = newBase;
		pp->nextPage = 0;
	}
	pageBase = (btu8 *)pp->memBase + pp->nextPage * BLOCKALLOC_PAGESIZE;
	pp->nextPage += numPages;
	memset(pageBase, 0, BLOCKALLOC_PAGESIZE * numPages);
	return pageBase;
}

/* ------------------------------------------------------------------------- */

static btbool	barrSetBitInLevel(bitMap_t *bm, btu32 index, btint level)
{
btu32	mask;
btint	i, j;

	if(level != 0 && bm->nextLevel == NULL)
		bm->nextLevel = memCalloc(256 * sizeof(bitMap_t));
	i = (index >> (level * 8)) & 0xff;
	if(level == 0 || barrSetBitInLevel(bm->nextLevel + i, index, level - 1)){
		mask = 1 << (i % 32);
		if((bm->bits[i/32] & mask) == 0){
			bm->bits[i/32] |= mask;
			if(bm->bits[i/32] == 0xffffffffUL){	/* there's a chance the array is full */
				for(j=0;j<8;j++){
					if(bm->bits[j] != 0xffffffffUL)
						return 0;	/* not full now */
				}
				return 1;	/* is full now */
			}
		}
	}
	return 0;	/* has not changed */
}

static void	barrSetBit(bitArray_t *ba, btu32 index)
{
bitMap_t	*p;
btint		i;

	while((index >> (ba->numLevels * 8)) != 0){	/* I'm afraid we have to rearrange things... */
		p = memCalloc(256 * sizeof(bitMap_t));
		p[0] = ba->root;
		ba->root.nextLevel = p;
		memset(ba->root.bits, 0, sizeof(ba->root.bits));
		ba->root.bits[0] = 1;	/* suppose the old root was full */
		for(i=0;i<8;i++){
			if(p->bits[i] != 0xffffffffUL){	/* at least one is not full */
				ba->root.bits[0] = 0;
				break;
			}
		}
		ba->numLevels++;	/* we just inserted a level... */
	}
	barrSetBitInLevel(&ba->root, index, ba->numLevels - 1);
}

/* ------------------------------------------------------------------------- */

static void	barrClearBit(bitArray_t *ba, btu32 index)
{
bitMap_t	*p;
btint		i, l;

	p = &ba->root;
	if((index >> (ba->numLevels * 8)) != 0)
		return;	/* bit is outside of our range */
	for(l=ba->numLevels;l!=0;){
		l--;
		i = (index >> (l * 8)) & 0xff;
		p->bits[i/32] &= ~(1 << (i % 32));
		if(p->nextLevel == NULL)
			return;	/* nothing to clear out there.... */
		p = p->nextLevel + i;
	}
}

/* ------------------------------------------------------------------------- */

static btu32	barrFirstClearBit(bitArray_t *ba)
{
btu32		index = 0, subindex;
bitMap_t	*p;
btint		i, j, l;

	p = &ba->root;
	for(l=0;l<ba->numLevels;l++){
		for(i=0;i<8;i++){
			if(p->bits[i] != 0xffffffffUL){	/* there's something free */
				for(j=0;;j++){
					if((p->bits[i] & (1 << j)) == 0){	/* found the clear bit */
						subindex = i * 32 + j;
						index = (index << 8) | subindex;
						if(p->nextLevel == NULL)
 							return index << (8 * (ba->numLevels - ++l));
						p = p->nextLevel + subindex;
						goto outerBreak;
					}
				}
			}
		}
/* If we get here, there's either an inconsistency, or the array is full
 * up to the amount possible with the current value of numLevels. Since the
 * former should be impossible, we return the next index outside our range.
 */
 		return 1 << (ba->numLevels * 8);
outerBreak:;
	}
	return index;
}

/* ------------------------------------------------------------------------- */

static void	barrFreeMap(bitMap_t *bm)
{
btint	i;

	if(bm->nextLevel != NULL){
		for(i=0;i<256;i++)
			barrFreeMap(bm->nextLevel + i);
		memFree(bm->nextLevel);
	}
}

static void	barrExit(bitArray_t *ba)
{
	barrFreeMap(&ba->root);
}

/* ------------------------------------------------------------------------- */

static void	*sarrPageAtIndex(smartArray_t *sa, btu32 index, btbool createOnDemand)
{
void	***b, *p;
btint	i;
btu32	splitIndex[8];	/* 8 is sufficient even for 64 bit pointers */

	for(i=0;index!=0;i++){
		splitIndex[i] = index & (BLOCKALLOC_PAGESIZE / sizeof(void *) - 1);
		index /= BLOCKALLOC_PAGESIZE / sizeof(void *);
	}
	if(i > sa->numIndirections){	/* index too large */
		if(!createOnDemand)
			return NULL;
		b = &sa->basePage;
		while(sa->numIndirections < i){
			p = *b;
			*b = pagePoolGetPages(sa->pagePool, 1);
			(*b)[0] = p;
			sa->numIndirections++;
		}
	}
	for(;i<sa->numIndirections;i++)
		splitIndex[i] = 0;
	b = &sa->basePage;
	for(i=sa->numIndirections;i!=0;){
		i--;
		if(*b == NULL){
			if(!createOnDemand)
				return NULL;
			*b = pagePoolGetPages(sa->pagePool, 1);
		}
		b = (void ***)&(*b)[splitIndex[i]];
	}
	if(*b == NULL && createOnDemand){
		*b = pagePoolGetPages(sa->pagePool, 1);
	}
	return *b;
}

/* ------------------------------------------------------------------------- */

static void	pageListExit(pageList_t *pl)
{
	barrExit(&pl->freeBlocks);
}

static void	pageListInit(pageList_t *pl, blockAllocator_t *ba, bti32 blockSize)
{
	if(blockSize < sizeof(btu32)){
		EXC_THROW_MSG(excInternalError, "blocksize too small");
	}
	pl->blockSize = blockSize;
	pl->blocksPerPage = BLOCKALLOC_PAGESIZE / blockSize;
	if(pl->blocksPerPage == 0){
		EXC_THROW_MSG(excInternalError, "blocksize exceeds page size");
	}
	pl->pages.pagePool = ba->pagePool;
	pl->freeBlocks.numLevels = 1;
}

/* ------------------------------------------------------------------------- */

blockAllocator_t	*ballocNew(pagePool_t *pp, btu32 *blockSizes, btint indexBits)
{
blockAllocator_t	*volatile ba = NULL;
btint				i, markerBits;

EXC_TRY
    ba = memCalloc(sizeof(blockAllocator_t));
	ba->pagePool = pp;
	for(i=0; blockSizes[i] != 0; i++);	/* search to end of list */
	ba->pageListCount = i;
	for(markerBits=1;markerBits<10;markerBits++){
		if((1<<markerBits) >= ba->pageListCount)
			break;
	}
	if(markerBits >= 10){
		EXC_THROW_MSG(excInternalError, "page list count out of range");
	}
	ba->markerShift = indexBits - markerBits;
	ba->indexMask = (1 << ba->markerShift) - 1;
	ba->pageList = memCalloc(ba->pageListCount * sizeof(pageList_t));
	for(i=0;i<ba->pageListCount;i++)
		pageListInit(&ba->pageList[i], ba, blockSizes[i]);
EXC_CATCH
	ballocFree(ba);
	EXC_RETHROW();
EXC_ENDTRY
	return ba;
}

/* ------------------------------------------------------------------------- */

void	ballocFree(blockAllocator_t *ba)
{
btint		i;

	if(ba != NULL){
		ba->pagePool = NULL;
		if(ba->pageList != NULL){
			for(i=0;i<ba->pageListCount;i++)
				pageListExit(&ba->pageList[i]);
			memFree(ba->pageList);
			ba->pageList = NULL;
		}
		memFree(ba);
	}
}

/* ------------------------------------------------------------------------- */

void	*ballocBlockAddress(blockAllocator_t *ba, btu32 blockIndex)
{
btu32			plindex, pagenum;
pageList_t		*pl;
btu8			*p;

	blockIndex--;
	plindex = blockIndex >> ba->markerShift;
	if(plindex >= ba->pageListCount){
		return NULL;	/* not allocated */
	}
	pl = &ba->pageList[plindex];
	blockIndex &= ba->indexMask;
	pagenum = blockIndex / pl->blocksPerPage;
	blockIndex %= pl->blocksPerPage;	/* make relative within page */
	if((p = sarrPageAtIndex(&pl->pages, pagenum, 0)) == NULL){
		return NULL;
	}
	return p + blockIndex * pl->blockSize;
}

/* ------------------------------------------------------------------------- */

btsize	ballocBlockSize(blockAllocator_t *ba, btu32 blockIndex)
{
btu32			plindex;

	blockIndex--;
	plindex = blockIndex >> ba->markerShift;
	if(plindex >= ba->pageListCount){
		EXC_THROW_MSG(excInternalError, "ballocSize: bogus index");
	}
	return ba->pageList[plindex].blockSize;
}

/* ------------------------------------------------------------------------- */

btu32	ballocAllocBlock(blockAllocator_t *volatile ba, btsize bytes, btsize *realSize)
{
btint			i, pagenum;
volatile btint	plindex;
pageList_t		*pl;
btu32			lindex;
volatile btu32	index = 0;
btu8			*base;

	for(i=0;i<ba->pageListCount;i++){
		if(ba->pageList[i].blockSize >= bytes)
			break;
	}
	if(i >= ba->pageListCount){	/* blocks of this size can not be handled */
		EXC_THROW_MSG(excInternalError, "ballocAllocBlock: block too large");
	}
	plindex = i;
EXC_TRY
	pl = &ba->pageList[plindex];
	*realSize = pl->blockSize;
	index = barrFirstClearBit(&pl->freeBlocks);	/* next free block */
	pagenum = index / pl->blocksPerPage;
	lindex = index % pl->blocksPerPage; 	/* index localized to page */
	base = sarrPageAtIndex(&pl->pages, pagenum, 1);
	base += lindex * pl->blockSize;
	barrSetBit(&pl->freeBlocks, index);
EXC_CATCH
	if(EXC_NAME != memExcNomem || ++plindex >= ba->pageListCount){
		EXC_RETHROW();
	}
	EXC_RETRY();
EXC_ENDTRY
	index |= plindex << ba->markerShift;	/* merge in pagelist index */
	return index + 1;	/* avoid index == 0 */
}

/* ------------------------------------------------------------------------- */

void	ballocFreeBlock(blockAllocator_t *ba, btu32 blockIndex)
{
btu32			plindex;
pageList_t		*pl;

	blockIndex--;
	plindex = blockIndex >> ba->markerShift;
	if(plindex >= ba->pageListCount){
		EXC_THROW_MSG(excInternalError, "ballocFreeBlock: bogus index");
	}
	pl = &ba->pageList[plindex];
	blockIndex &= ba->indexMask;
	barrClearBit(&pl->freeBlocks, blockIndex);
}


/* ------------------------------------------------------------------------- */
/* ------------------------- builtin test routine -------------------------- */
/* ------------------------------------------------------------------------- */

#ifdef TEST_BLOCKALLOC


void	ballocCheckConsistency(blockAllocator_t *ba)
{
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

#include <math.h>
#include "test.h"


#define	ZONE_SIZE		1024*4096
#define	TEST_MINSIZE	1
#define	TEST_MAXSIZE	4096

#define	HIST_GRAN		10000

typedef struct{
	int		size;
	btu32	index;
	void	*checkBase;
}memblock_t;

static blockAllocator_t	*allocator;

static memblock_t	memory[ZONE_SIZE/TEST_MINSIZE];	/* should be sufficient */
static int			index_limit;			/* search limit in 'memory' */
static int			block_count;			/* number of allocated blocks */
static int			total_allocated;		/* amount of allocated memory */

static int			fullHistogram[(ZONE_SIZE + HIST_GRAN/2)/HIST_GRAN+1];

/* ------------------------------------------------------------------------- */

btu32	myAlloc(blockAllocator_t *ba, btsize size)
{
volatile btu32	index;
btsize			realsize;

EXC_TRY
	index = ballocAllocBlock(ba, size, &realsize);
EXC_CATCH
	if(EXC_NAME != memExcNomem){
		EXC_RETHROW();
	}
	index = 0xffffffff;
EXC_ENDTRY
	return index;
}

void	myFree(blockAllocator_t *ba, btu32 index)
{
	ballocFreeBlock(ba, index);
}

static memblock_t	*getFreeBlock(int size)
{
int		i;

	for(i=0;i<index_limit;i++){
		if(memory[i].checkBase == NULL)
			break;
	}
	if(i == index_limit)
		index_limit++;
	memory[i].checkBase = malloc(size);
	memory[i].size = size;
	return &memory[i];
}

static void	fillBlock(memblock_t *block)
{
int		i, len = block->size;
char	*p = (char *)(ballocBlockAddress(allocator, block->index));

	for(i=0;i<len;i++)
		if(i == 0)
			p[i] = random_int(255) + 1;
		else
			p[i] = random_int(256);
	memcpy((char *)block->checkBase, p , len);
}

static void	verifyBlocks(void)
{
int		i;
void	*base;

	for(i=0;i<index_limit;i++){
		if(memory[i].checkBase != NULL){
			base = ballocBlockAddress(allocator, memory[i].index);
			if(memcmp(base, memory[i].checkBase, memory[i].size) != 0){
				printf("*** content error in block index %d of length %d\n",
						memory[i].index, memory[i].size);
				abort();
			}
		}
	}
}

static int	allocateBlock(void)
{
int			size;
memblock_t	*block;

	size = TEST_MINSIZE + random_int(TEST_MAXSIZE - TEST_MINSIZE + 1);
	block = getFreeBlock(size);
	block->index = myAlloc(allocator, size);
	if(block->index == 0xffffffff){
		free(block->checkBase);
		block->checkBase = NULL;
		return 0;
	}
	block_count++;
	total_allocated += size;
	fillBlock(block);
	return 1;
}

static void	freeBlock(memblock_t *block)
{
	myFree(allocator, block->index);
	free(block->checkBase);
	block->checkBase = NULL;
	block_count--;
	total_allocated -= block->size;
	if(total_allocated < 0){
		printf("fatal in freeBlock()\n");
		abort();
	}
}

void main(void)
{
int			count, i, j, freethis;
btsize		blockSizes[] = {4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, TEST_MAXSIZE, 0};

	random_init(12, 34, 56, 78);
EXC_TRY
	allocator = ballocNew(pagePoolNew(), blockSizes, 32);
	printf("%d pages (%d bytes) administrative info\n", allocator->pagePool->nextPage, allocator->pagePool->nextPage * BLOCKALLOC_PAGESIZE);
	for(count=1;;count++){
		for(i=1;;i++){			/* fill memory in a loop */
			if(!allocateBlock()){
				break;
			}
		}
		verifyBlocks();
		fullHistogram[(total_allocated + HIST_GRAN/2) / HIST_GRAN]++;
		if((count % 1000) == 0){
			printf("count = %d\n", count);
			printf("%8s %8s\n", "total", "count");
			printf("---------------------------------------\n");
			for(i=0;i<(ZONE_SIZE+HIST_GRAN/2)/HIST_GRAN+1;i++){
				if(fullHistogram[i] != 0){
					printf("%8d %8d\n", i * HIST_GRAN, fullHistogram[i]);
				}
			}
			printf("---------------------------------------\n");
			ballocCheckConsistency(allocator);
			printf("---------------------------------------\n");
		}
		freethis = random_int(block_count);
		for(j=i=0;i<index_limit;i++){
			if(memory[i].checkBase == NULL)
				continue;
			if(j >= freethis){
				freeBlock(&memory[i]);
				j = 0;
			}else{
				j++;
			}
		}
	}
EXC_CATCH
	fprintf(stderr, "*** Exception occured: %s (%s)\n", EXC_NAME, EXC_MSG);
	abort();
EXC_ENDTRY
}

#endif /* TEST_MALLOC */
