/*
 * Name: blockalloc.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-29
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: blockalloc.h,v 1.1.1.1 1999/04/12 13:56:43 cs Exp $
 */

#ifndef __blockalloc_H_INCLUDED__
#define __blockalloc_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This module contains two classes for managing virtual memory:
1. pagePool: This class allocates an amount of virtual memory and gives away
   slices of it on demand.
2. blockAllocator: This class gives functionality comparable to that of
   malloc(), only it can be optimized for frequently used block sizes. It
   uses block indices instead of addresses. Block indices can not be zero.
*/

/* ------------------------------------------------------------------------- */

#define	BLOCKALLOC_PAGESIZE		8192
/* BLOCKALLOC_PAGESIZE should be in the order of magnitude of the system's
 * virtual memory page size and should differ from it by an integer factor.
 * For performance reasons, it should not be smaller than the page size.
 * The size of 8192 has been carefully chosen. Please don't change.
 */

typedef struct pagePool{
	bti32				numPages;
	bti32				nextPage;
	void				*memBase;
	struct array		*usedBases;		/* old bases for later freeing */
}pagePool_t;

typedef struct blockAllocator{
	pagePool_t			*pagePool;		/* that's where pages are taken from */
	bti32				indexMask;
	bti32				markerShift;
	bti32				pageListCount;	/* number of pagelists administered */
	struct pageList		*pageList;		/* an array of pagelists */
}blockAllocator_t;

/* ------------------------------------------------------------------------- */

pagePool_t	*pagePoolNew(void);
/* Creates a new pagePool.
 */
void	pagePoolFree(pagePool_t *pp);
/* Frees all memory allocated in pagePoolNew().
 */
void	*pagePoolGetPages(pagePool_t *pp, bti32 numPages);
/* Slices off 'numPages' pages from the pool of memory. One page is
 * BLOCKALLOC_PAGESIZE bytes in size.
 */

/* ------------------------------------------------------------------------- */

blockAllocator_t	*ballocNew(pagePool_t *pp, btu32 *blockSizes, btint indexBits);
/* Creates a new blockAllocator. The blockAllocator builds on a pagePool
 * object, which must be passed in 'pp' on creation. The 'blockSizes' array
 * is an array of integers holding frequently used block sizes. This array
 * _must_ be ordered for increasing sizes and must be terminated with a zero
 * size entry. Blocks are addressed by index. Since the user may wish to use
 * less than 32 bits to store indices, 'indexBits' specifies the number of
 * bits required for indices.
 * Throws: excInternalError, memExcNomem
 */
void	ballocFree(blockAllocator_t *ba);
/* Frees all memory allocated with ballocNew().
 */
void	*ballocBlockAddress(blockAllocator_t *ba, btu32 index);
/* Finds out the physical memory address of a block allocated with
 * ballocAllocBlock(). Returns NULL if the block is not allocated.
 */
btsize	ballocBlockSize(blockAllocator_t *ba, btu32 index);
/* Returns the number of bytes that may be used in the block at 'index'.
 * Throws: excInternalError
 */
btu32	ballocAllocBlock(blockAllocator_t *ba, btsize bytes, btsize *realSize);
/* Allocates a new block of memory with at least 'bytes' bytes. The block will
 * actually have room for one of the sizes passed in ballocNew(). The real size
 * of the block (which can be used by the allocator) is returned in 'realSize'.
 * The blocks allocated with 'ballocAllocBlock()' must be used, that means the
 * first four bytes of the block must not be zero because they are used as
 * indication for unallocated blocks.
 * Returns: the index of the block allocated.
 * Throws: excInternalError, memExcNomem
 */
void	ballocFreeBlock(blockAllocator_t *ba, btu32 blockIndex);
/* Frees a memory block allocated with ballocAllocBlock().
 * Throws: excInternalError, memExcNomem
 */

/* ------------------------------------------------------------------------- */

#endif /* __blockalloc_H_INCLUDED__ */
