/*
 * Name: guimanager.h
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-03-01
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: guimanager.h,v 1.5 2000/01/21 12:26:47 cs Exp $
 */

#ifndef __guimanager_H_INCLUDED__
#define __guimanager_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This module contains some helper functions for working with GUI objects.
*/

/* ------------------------------------------------------------------------- */

struct array;
struct ipcMsg;
struct guiClass;
struct ipcSrvClient;

typedef struct guimanRes{
	btError	err;
	btbool	didSet;
	btint	button;
	btu8	*file;
}guimanRes_t;

#define	GUIMAN_DEFAULT_ENTRY	".default"	/* name of default hierarchy element */

/* ------------------------------------------------------------------------- */
void	guimanGetValues(btu8 ***cfgPath, btu8 ***hierarchicalPath, struct array *guiEntries, struct ipcMsg *rq, struct ipcMsg *reply);
/* This function is a generic implementation for the data source method
 * MAIN_GUI_METHOD_GET for objects of type MAIN_GUI_CLASS_PLAINCFG or
 * MAIN_GUI_CLASS_HIERARCHYCFG.
 * Parameters: 'cfgPath' is a pointer to an array of string array pointers
 *    (nice thing, isn't it :-)) describing where in the configuration and
 *    database files to look up values. It's best to look at usage examples
 *    to understand it. 'hierarchicalPath' is the same thing as 'cfgPath',
 *    additionally used in hierarchical configurations. If 'hierarchicalPath'
 *    is NULL, the object is of type MAIN_GUI_CLASS_PLAINCFG. The components
 *    of 'hierarchicalPath' are postfixed with the hierarchy element before
 *    the lookup in the configuration/database is done. 'guiEntries' is an
 *    array holding the guiEntry_t descriptions of the objects, 'rq' is the
 *    message received from the GUI application holding the parameters of the
 *    request and 'reply' is the reply message that will be sent to the GUI
 *    application.
 */
void	guimanSetValues(btbool isHierarchical, btu8 **cfgPath, struct array *guiEntries, struct ipcMsg *rq, struct ipcMsg *reply, btint uid);
/* This function is a generic implementation of the data source method
 * MAIN_GUI_METHOD_SET for objects of type MAIN_GUI_CLASS_PLAINCFG or
 * MAIN_GUI_CLASS_HIERARCHYCFG.
 * Parameters: 'isHierarchical' distringuishes between MAIN_GUI_CLASS_PLAINCFG
 *    and MAIN_GUI_CLASS_HIERARCHYCFG objects, 'cfgPath' is the destination
 *    in the config/database file where to write the values, 'guiEntries'
 *    are the guiEntry_t entries describing the GUI object, 'rq' is the
 *    request message from the GUI application holding the parameters,
 *    'reply' is the reply message that will be sent back to the GUI
 *    application and 'uid' is the requesting user's Unix user ID for access
 *    permission checks.
 */
btbool	guimanAllowAccess(btint permFlags, btint uid);
/* This function tells you whether a user (with given Unix uid) can change
 * a setting with the given permission flags.
 * Parameters: 'permFlags' are the permission flags of the setting as they
 *    are defined in uiLib/maininterface.h defines MAIN_GUI_PERM_*, 'uid'
 *    is the Unix user ID of the user.
 * Returns: 1, if the user may change the setting, 0 otherwise.
 */
void	guimanRunDialog(struct guiClass *description, struct ipcSrvClient *client, void *dataBase, btbool runModal, btbool ontopOfFileselector, void *caller);
/* This function runs an entire dialog on the GUI application and returns by
 * calling the callback only after the user finished the dialog. Dialogs are
 * described with GUI classes of type MAIN_GUI_CLASS_PLAINCFG.
 * Parameters: 'description' is a guiClass_t structure of type
 *    MAIN_GUI_CLASS_PLAINCFG containing the entries and buttons of the dialog.
 *    the field 'extraInfo' in each entry gives the offset from 'dataBase'
 *    where the result of the dialog will be written (and where the defaults
 *    presented to the user come from). 'client' is the representation of the
 *    GUI application where the dialog will run. 'dataBase' is a structure
 *    representing the contents of the dialog. Strings are character pointers
 *    pointing to an allocated space. The returned strings must be freed by
 *    the caller. 'runModal' is a flag which tells whether the dialog should
 *    be run modally by the GUI. This flag is advisory only, it's up to the
 *    GUI how it is implemented. 'ontopOfFileselector' is a flag which tells
 *    the GUI to display the dialog on top of fileselection dialogs if
 *    possible. This flag is only advisory, too. 'caller' is the caller's
 *    context for the callback.
 * Callback: guimanRes_t with 'err' set to any error that might have occured
 *    (e.g. application quitted, etc.). 'didSet' tells the caller whether
 *    the user aborted with the cancel button (didSet = 0) or whether the
 *    values were actually set (didSet = 1).
 */
void	guimanRunFileselector(struct ipcSrvClient *client, btu8 *title, btu8 *startAtPath, void *caller);
/* This function runs a file selection dialog at the given GUI client.
 * Parameters: 'client' specifies the GUI client where to display the
 *    fileselection dialog, 'title' is the title line of the file selection,
 *    if the GUI toolkit allows it, 'startAtPath' is the preselected path
 *    presented to the user and 'caller' is the caller's context.
 * Callback: guimanRes_t with 'err' set to communication errors that might
 *    have occured, 'didSet' set to '1' if the user clicked 'OK' and 'file'
 *    set to the selected file path.
 */
void	guimanRunAlert(struct ipcSrvClient *client, btu8 *title, btu8 *text, btu8 *button1, btu8 *button2, btu8 *button3, void *caller);
/* This function runs an alert panel at the given GUI client.
 * Parameters: 'client' is the GUI client where the alert will be displayed,
 *    'title' is the title of the alert, which will be printed in a bigger
 *    font than the alert text, 'text' is the alert text which is automatically
 *    wrapped to fit the panel, 'button1' to 'button3' are the texts displayed
 *    in the buttons (if a text is NULL, there is no associated button). Button
 *    3 is the default button, if it exists. 'caller' is the caller's context.
 * Callback: guimanRes_t with 'err' set to communication errors, 'button' set
 *    to the index of the button clicked (1 for button1, 3 for button3, 0 if
 *    the alert has been closed without clicking a button).
 */
void	guimanRunSelection(struct ipcSrvClient *client, btu8 *title, btu8 *ok, btu8 *cancel, struct array *texts, void *caller);
/* This function runs a selection box at the given GUI client.
 * Parameters: 'client' is the GUI client where the box will be dislplayed,
 *    'title' is the title displayed in the title bar of the box, 'ok' is the
 *    text displayed in the OK button, 'cancel' is the text displayed in the
 *    cancel button, 'texts' is an array of (btu8 *) containing the texts among
 *    which the user shall choose and 'caller' is the caller's context.
 * Callback: guimanRes_t with 'err' set to any communication error that might
 *    have occured, 'didSet' set to '1' if the user clicked 'OK' or double-
 *    clicked an entry and 'file' set to the entry selected or holding the
 *    text entered by the user.
 */
void	guimanDisplayError(struct ipcSrvClient *client, btError *err, void *caller);
/* This function displays an error at the given GUI client.
 * Parameters: 'client' is the GUI client where the error should be displayed,
 *    'err' is the representation of the error and 'caller' is the caller's
 *    context.
 * Callback: guimanRes_t with 'err' set.
 */
void	guimanAskData(struct ipcSrvClient *client, btu8 *description, btbool withEcho, void *caller);
/* This function asks the user on the commandline interface for a string.
 * It first prints out the string 'description', which should tell the user
 * what to type and then takes the user's input. If 'withEcho' is non-zero,
 * the characters will be echoed as they are typed, otherwise they are treated
 * as a password and not echoed.
 * Parameters: 'client' is the UI client where the dialog should take place,
 *    'description' is the prompt presented to the user, 'withEcho' tells
 *    the user interface whether the typed string is secret and 'caller' is
 *    the caller's context.
 * Callback: guimanRes_t with 'err' and 'file' set. 'file' points to the
 *    string the user typed in.
 */
/* ------------------------------------------------------------------------- */

#endif /* __guimanager_H_INCLUDED__ */
