/*
 * Name: ipcclient.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-12
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: ipcclient.h,v 1.7 2001/03/31 13:31:43 cs Exp $
 */

#ifndef __ipcclient_H_INCLUDED__
#define __ipcclient_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This module contains RPC functions doing the marshalling and unmarshalling
called by the user interface commands. There's an RPC interface to the
daemon for user interface commands, if you didn't know....
*/

/* ------------------------------------------------------------------------- */

struct ipcMsg;

typedef struct ipccRequestData{
	btu32			sid;
	bti32			procId;
	struct ipcMsg	*msg;
	btint			protocolVersion;
}ipccRequestData_t;

typedef struct ipccRes{
	btError	err;
	btu8	*errmsg;
}ipccRes_t;

typedef struct ipccModuleInfo{
	btu8	*name;
	btint	type;
	btint	fnBase;
	btint	fnCount;
}ipccModuleInfo_t;
/* This structure is a container for what's known about a module */

/* ------------------------------------------------------------------------- */

extern btbool	ipcclientDebug;
/* If this variable is set to a non-zero value, the ipcclient module logs
 * all communication with the sharity daemon to standard error.
 */

void	ipccSetRequestHandler(btcallback handler, void *ctx);
/* The request handler is a function which is called if a client library
 * function receives a request from the daemon instead of an expected reply.
 * The handler is expected to process the request and send the reply back to
 * the daemon. This function sets the request handler function.
 * Parameters: 'handler' is a function pointer to the handler function and
 *    'ctx' is passed as the first parameter to the handler function.
 * Call parameters for handler: The first parameter is the context pointer
 * passed to this function and the second parameter is a pointer to a
 * ipccRequestData_t structure. If the daemon crahsed, the second parameter
 * is set to NULL to notify the client about the daemon death.
 */

void	ipccSetReadMessageHandler(btcallback handler);
/* The readMessage handler is called to read a message from the socket to
 * the daemon. A client may require a different handler if it has to read
 * messages from its own eventloop.
 * Parameters: 'handler': function pointer to a function which has the
 *     following two parameters: 'ipcMsg_t *msg' (a pointer to the message
 *     into which the data should be read, setting the message's index to
 *     the first content-byte of the message) and 'btbool *succeeded'
 *     (return 0 in the boolean if no message was available, 1 otherwise).
 * Note: If the handler function returns 0 in 'succeeded' indicating that
 * no message is available, it will be called again immediately if the
 * expected message has not been received by other means.
 */

btint	ipccNumOutstandingReplies(void);
/* This function returns the number of replies expected from the daemon by
 * the RPC mechanism. This number may be needed by a readMessage handler
 * to determine whether a received reply should be evaluated or not.
 */

void	ipccSetProtocolVersion(btint fnBase, btint fnCount, btint version);
/* This function tells the client library which protocol version is used for
 * which module in the inter-process communication to the Sharity daemon.
 * The module is specified in terms of an IPC function number range.
 * Parameters: 'fnBase' is the first IPC function number of the module,
 *    fnCount is the number of functions defined for the module and 'version'
 *    is the IPC protocol version implemented by the client. This number is
 *    used internally for version checks.
 */
btint	ipccProtocolVersion(btint fnNumber);
/* This function queries the protocol version (set with
 * ipccSetProtocolVersion()) for a given IPC function number.
 */

btint	ipccConnect(void);
void	ipccWriteMsg(struct ipcMsg *msg);
btint	ipccReadMsg(struct ipcMsg *msg);
/* This function reads a message from the pipe. It positions the read pointer
 * after the message length and protocol version entries.
 * Returns: The protocol version. May return -1 if no message could be found
 * (you may call it again in this case).
 */

/* creating/reading IPC messages by hand: */
struct ipcMsg	*ipccCurrentRxMessage(void);
void	ipccStartProc(struct ipcMsg *msg, int msgId);
void	ipccRpc(struct ipcMsg *msg);
/* This function executes the remote procedure call. The message in 'msg' is
 * sent to the daemon and the reply is received in the same message object.
 * If a request from the daemon is received instead of the reply, the function
 * passed in 'ipccSetRequestHandler()' (if any) is called. ipccRpc() must only
 * be called with messages initilized with 'ipccStartProc()' because
 * 'ipccStartProc()' passes information to 'ipccRpc()' in a static variable.
 */

/* added by karl for MAC OS X */
void	ipccQueueReply(struct ipcMsg *msg);


/* some handy utility functions: */
void	ipccPrintErr(ipccRes_t *res);

btint	ipccUserId(void);
/* This function returns the user-ID of the user who executes the client
 * utility. This function is used instead of getuid() because we might
 * want to use the uidsecure module where uidsecUserId() must be used.
 * This gives us a single point to change.
 */

/* ------------------------------------------------------------------------- */

ipccModuleInfo_t	*ipccGetModuleInfo(ipccRes_t *res, btu8 *moduleName);

void	ipccRegisterForCallback(ipccRes_t *res, btu8 *modName);
void	ipccUnregisterForCallback(ipccRes_t *res, btu8 *modName);
void	ipccWillTerminate(ipccRes_t *res, btint termReason, btbool *doTerm);
void	ipccListModules(ipccRes_t *res);
void	ipccSlashChars(ipccRes_t *res, btu8 *slash, btu8 *backsl, btsize bufsize);
void	ipccMount(ipccRes_t *res, btu8 *feModule, btu8 *fePath, btu8 *beModule, btu8 *bePath, btbool readOnly, btbool fakeMode);
void	ipccUmount(ipccRes_t *res, btu8 *mountpoint, btbool fakeMode);
void	ipccGetObjectDescription(ipccRes_t *res, btint objectId);
void	ipccConfigObjects(ipccRes_t *res);
void	ipccGetCfgValue(ipccRes_t *res, btu8 *cfgPath, btbool dbOnly, btu8 **value);
void	ipccSetCfgValue(ipccRes_t *res, btu8 *cfgPath, btu8 *value);
void	ipccDelCfgValue(ipccRes_t *res, btu8 *cfgPath);

void	ipccStoremnt(ipccRes_t *res, btu8 *mountpoint, btbool deleteStored);
void	ipccLookupList(ipccRes_t *res, btu8 *listName, btint *listId);
void	ipccHelpForKey(ipccRes_t *res, btu8 *helpItem, btu8 **helpText);

void	ipccCifsLogin(ipccRes_t *res, btu8 *serverName, btu8 *shareName);
void	ipccCifsLogout(ipccRes_t *res, btu8 *serverName);
void	ipccCifsGetLicense(ipccRes_t *res, btu8 *lstring, btu8 **licenseType, bti32 *maxClients, bti32 *maxServers, btbool *srvCountLocal, btu8 **expiry);
void	ipccCifsCurrentLicenseLoad(ipccRes_t *res, btint *servers, btint *clients);
void	ipccCifsCurrentServers(ipccRes_t *res);
void	ipccCifsSetLicense(ipccRes_t *res, btu8 ldata[12]);
void	ipccCifsStore(ipccRes_t *res, btu8 *serverName, btbool doDelete);


/* ------------------------------------------------------------------------- */

#endif	/* __ipcclient_H_INCLUDED__ */
